//-*-c++-*-

// ====================================
//  NAME    : FWUKL1 Library
//
//  VERSION : 1
//
//  REVISION: 0
//
//  DATE    : 2006/10/12
//
//  RELEASE : 2007/04/04
//
//  PLATFORM: PVSS II
//
//  AUTHOR  : Gareth Rogers
//
//  EMAIL   : rogers@hep.phy.cam.ac.uk
// ====================================

/* \mainpage UKL1 Library Documentation
 *
 * This library is used to manipulate the state of the L1 board for the RICH detector.
 * Currently there are four possible states the board can exist in:
 *   \li Unconfigured This is the initial state of the board after turning on, a reset or if any register is
 *         not correctly configured. Its behaviour in this state is undefined, however it is may act vaguley
 *         sensibly depending on what caused this state to be reached.
 *   \li Ready All the registers on the L1 board have been configured and the board is ready to receive data.
 *         All input channels, front end (FE) FPGAs and gigabit ethernet (GBE) ports are disabled,no data
 *         can there for be sent or received. This state can be reached from any of the other defined states.
 *   \li Start This state can only be reached from the ready state. It configures the enable settings for the
 *         input channels, FE FPGAs and GBE ports. Once this has been called data can be receieved, processed
 *         and transmitted. If the start state is reached and ready has not been called at least once then the
 *         state will be unconfigured.
 *   \li Stop This state is effectively the same as ready, as it disables the input channels, FE FPGAs and GBE
 *         ports. It can only be reached from either the ready or start states otherwise the board state will
 *         be unconfigured.
 * \n
 * The library has been designed such that all setting of hardware registers can be done through the function
 * 'ukl1SetCcpcState', using the states defined in the constants and global variables section
 * (e.g. UKL1_START_STATE). It will call all the necessary functions to set the hardware to the appropriate state,
 * logging any errors that occur and returning a status number to indicate whether the state change was successful.
 * This is the only function that a panel will need to call to configure the board.
 * \n
 * Functionality is provided to create a Ukl1Board datapoint will sensible default settings for the each datapoint
 * element value. The datapoint will be called RichL1 followed by the name of the RICH L1 board as it appears in the
 * DIM name server, allowing a unique datapoint to be created for each of RICH L1 boards found in a system. The
 * state machine will access the values stored in the appropriate datapoint by setting the global variable
 * ukl1CcpcName. This allows both the appropriate datapoint settings to be accessed and the appropriate L1 board
 * to be written to. The state of each of these boards is maintained through the global mapping ukl1CcpcList, which
 * maintains a knowledge of each L1 board state, associating it with its name. It is up to the User to ensure that
 * this always contains a valid set of L1 boards.
 * \n
 * A second datapoint, RichL1Log, contains the status, warning and error messages that are generated by the
 * library function calls and this datapoint should be monitored to ensure that all library messages are seen.
 *
 */

// =========================
//  CONSTANTs & GLOBAL VARs
// =========================

/** @defgroup SectionConstants Constants
 *  List of constants that can be used by the user.
 *  @{
 */

// /*! Defines the numeric value of the message level . */
// const unsigned FWUKL1_ALWAYS_MESSAGE_LEVEL = 0;

// /*! Defines the numeric value of the message level . */
// const unsigned FWUKL1_INFO_MESSAGE_LEVEL = 1;

// /*! Defines the numeric value of the message level . */
// const unsigned FWUKL1_DEBUG_MESSAGE_LEVEL = 2;

// /*! Defines the numeric value of the message level . */
// const unsigned FWUKL1_VERBOSE_MESSAGE_LEVEL = 3;

/*! Defines the name of the top level control unit (CU) that is used in the L1 FSM tree. */
const string FWUKL1_CU = "L1";
/*! Defines the name of the data point type of the UKL1 hardware type. All UKL1 board data points are created of this type. */
const string FWUKL1_HW_TYPE = "UKL1";

/*! Broadcast FE FPGA number. Value used to write to all the FE FPGAs. */
const unsigned FWUKL1_BROADCAST_FPGA = 5;
/*! Broadcast FE FPGA channel number. Value used to write to all of a FE FPGA's channels. */
const unsigned FWUKL1_BROADCAST_CHANNEL = 10;
/*! Number of Front end (ingress) FPGAs on the RICH L1 board. */
const unsigned FWUKL1_FRONT_FPGAS = 4;
/*! Number of channels that are on each front end FPGA. */
const unsigned FWUKL1_FPGA_CHANNELS = 9;
/*! Number of channels that are on the RICH L1 board. */
const unsigned FWUKL1_BOARD_CHANNELS = 36;
/*! Number of ports on the gigabit ethernet card. */
const unsigned FWUKL1_GBE_PORTS = 4;
/*! Number of temperature sensors present on the board. */
const unsigned FWUKL1_TEMPERATURE_SENSORS = 3;

/*! Defines a string, which represents the unconfigured state. */
const string FWUKL1_NOT_READY_STATE = "NOT_READY";
/*! Defines a string, which represents the ready state. */
const unsigned FWUKL1_CONFIGURING_STATE = 1;
/*! Defines a string, which represents the stop state. */
const unsigned FWUKL1_STOP_STATE = 2;
/*! Defines a string, which represents the start state. */
const unsigned FWUKL1_START_STATE = 3;

/*! This numeric constant is used to identify a production UKL1 board. */
const unsigned FWUKL1_PRODUCTION = 2;
/*! This numeric constant is used to identify a prototype UKL1 board. */
const unsigned FWUKL1_PROTOTYPE = 1;

/*! Constant that defines the size of a 32-bit data word. */
const unsigned FWUKL1_32BIT_WORD = 32;
/*! Constant that defines the size of a 16-bit data word. */
const unsigned FWUKL1_16BIT_WORD = 16;
/*! Constant that defines the size of an 8-bit data word. */
const unsigned FWUKL1_8BIT_WORD = 8;

/*! Constant that defines the prefix to all the datapoints created from the DPT HwTypeCCPCUKL1. */
//const string FWUKL1_HW_DPNAME_PREFIX = "HWCCPCUKL1/";
/*! Name of the data point element that stores the list of names of found L1 boards.*/
const string FWUKL1_L1_LIST_DPE = "Ukl1BoardDimNamesList.DimServerNameList";

// /*! A level that indicates what status messages should be written to the buffer. */
// global unsigned gufwUkl1_StatusMessageLevel = 1;//FWUKL1_INFO_MESSAGE_LEVEL;
// /*! A level that indicates what warning messages should be written to the buffer. */
// global unsigned gufwUkl1_WarningMessageLevel = 1;//FWUKL1_INFO_MESSAGE_LEVEL;
// /*! A level that indicates what error messages should be written to the buffer. */
// global unsigned gufwUkl1_ErrorMessageLevel = 1;//FWUKL1_INFO_MESSAGE_LEVEL;

/*! A boolean to control whether or not the hardware writes are verified by reading back from that register. */
global bool gbfwUkl1_verifyWrite = FALSE;

// global bool bfwUkl1_SendMessageToDataPoint = TRUE;


//@}


// ==================
//  UKL1 BOARD SETUP
// ==================

/** @defgroup SectionSetup
 *    The functions in this section are responsible for setting up the software object that represents a UKL1 board.
 *    It provides a function that is capable of dynamically adding or removing the UKL1 boards from the system. By
 *    placing in an infinite loop it can be used to add/remove UKL1 boards from the system automatically. This should
 *    be used with some caution as it will regenerate the FSM tree everytime a board is added or removed from the system.\n
 *    The ability to manually add/delete UKL1 boards from the system is also provided, by the User must take care that
 *    they are adding/removing a valid UKL1 board to or from the system.\n
 *    UKL1 boards can be identified via their ID number and a function is provided that is capable of identifying UKL1
 *    production and prototype boards. It is via this function that the fwUkl1_findAndAddUkl1Boards() identifies UKL1
 *    boards from all the CCPCs present in the DIM server.
 *  @{
 */

/*!
 * Gets the names off all the CCPCs found in the DIM server and then identifies the UKL1 boards based on their ID number.
 * It will then add boards that are found to the system. It is also capable of removing boards that are found in the system
 * but no longer present in DIM.
 *
 * \param  refresh If TRUE the list existing list of UKL1 boards will be updated with any new boards found and those
 *           no longer present will be removed. FALSE all existing boards will be deleted along with their associated
 *           settings and the whole list created from new.
 * \param  deleteExisting If the datapoint is found to exist it will be deleted and then recreated with the relevant
 *           settings.
 * \param  forceReconfigure At present this only affects the DP used to interact with the hardware and not the FSM tree.
 *           If TRUE it will recreate the default settings from those hardcoded into the file and reconfigures the
 *           harware data point from these settings. If FALSE the settings are just loaded from the defaults.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * There are several modes of operation of this function depending on the values of refresh, deleteExisting and forceReconfigure.
 * Full details of how the HW DP and the FSM tree are affected by these can be found in the documentation for _fwUkl1_createHwTypeCCPCUKL1
 * and _fwUkl1_createUkl1FsmTreeDU. A summary of the most useful modes are described.
 * \li refresh=TRUE, deleteExisting=FALSE, forcereconfigure=FALSE - The most efficient mode, this finds those UKL1 boards that are
 *       not in the system and creates them loading from defaults, it deletes UKL1 boards no longer in the DNS node and will
 *       perform no updates on the boards that already fully exist in the system. This is the recommend mode of operation for standard use.
 * \li refresh=FALSE, deleteExisting=FALSE, forceReconfigure=FALSE - This first clears all the boards from the system before adding all
 *       all those identified in the DNS node from defaults. Useful if the settings for existing UKL1 boards are believed to be corrupt.
 * \li Any mode with deleteExisting=TRUE will cause the exising software objects to be deleted and before recreated, much like refresh=TRUE.
 * \li forceReconfigure=TRUE is meant to be used only by experts as it is will reload from settings hardcoded into this library.
 *       Only required when an expert makes a change to the datapoint structure and then the library must be updated to match.
 *       Can also help if the defaults are believed to be corrupt or when first installed in a system.
 */
void fwUkl1_findAndAddUkl1Boards(bool refresh, bool deleteExisting, bool forceReconfigure, dyn_string& exceptionInfo) {
  //Holds the list of CCPC names retrieved.
  dyn_string ccpcList;
  //First get a list of presently registered CCPCs in the DIM server.
  fwCcpc_get(ccpcList);
  //Save the total number found for later use.
  const int numCcpc = dynlen(ccpcList);
  //Filter the CCPC names to determine, which of these are actually UKL1 boards.
  //Holds the names of the found UKL1 boards.
  dyn_string toBeAddedUkl1List;
  //Holds the version number of the found UKL1 boards.
  dyn_uint versionList;
  //Keep track of the number of UKL1 boards added to the list, start from 1 due to PVSS array silliness.
  unsigned numUkl1Added = 1;
  //Loop over all the found CCPCs.
  for (unsigned ccpc = 1; ccpc <= numCcpc; ++ccpc) {
    //Name of the CCPC currently being checked.
    const string ccpcName = ccpcList[ccpc];
    //Check to see if it is a production or prototype.
    //Ignore the return as it will probably just indicate a prototype anyway...
    int isUkl1;
    string serialNumber;
    fwUkl1_isUkl1Board(ccpcName, isUkl1, serialNumber, exceptionInfo);
    //At the moment we can ignore any errors as if they do occur it means that we probably have the prototype board.
    //Eventually this will need to be checked properly.
    if ( 0 != isUkl1 ) {
      //Add the name of the board to the list of UKL1 names.
      toBeAddedUkl1List[numUkl1Added] = ccpcName;
      //Add constant to declare whether we are dealing with a production or prototype board.
      versionList[numUkl1Added] = isUkl1;
      //Increment the number added.
      ++numUkl1Added;
    }//if(0!=isUkl1) Don't do anything for else as we are just not interested in this board.
  }//CCPC for

  //Produce a warning message if we failed to find any UKL1s or even CCPCs.
  if ( 0 == numCcpc )
    fwException_raise(exceptionInfo, "WARNING", "fwUkl1_findAndAddBoards(): No CCPCs found in the DIM server.", "0");
  else if ( 0 == dynlen(toBeAddedUkl1List) )
    fwException_raise(exceptionInfo, "WARNING", "fwUkl1_findAndAddBoards(): No UKL1 boards found, but found " + numCcpc + " CCPCs in the DIM server.", "0");
  //If we didn't find anything then the rest of this section is doesn't really need to be done, but it won't do anything anyway.

  //Now we have a list of UKL1 boards in the system lets deal with them.
  //The variable toBeAddedList holds the names of the UKL1 boards that need to be added to the library.
  //This always need to contain all the found boards as we always need to check the datapoints for consistency
  //and ensure the relevant DIM and FSM functionality is established.
  //The function fwUkl1_AddUkl1Board will ensure that the minimum possible is done depending on calling arguments.

  //Holds the names of the UKL1 boards that need to be removed from the library. It will either be those that are
  //no longer present after a refresh or all of them in the case of a reload.
  dyn_string toBeDeletedUkl1List;
  //Just find the new boards in the system and those that have been lost.
  if (refresh) {
    //Reset the number added.
    numUkl1Added = 1;
    //Look through the list of existing names and check for those that need to be deleted.
    //Get the list.
    const dyn_string currentList = fwUkl1_getUkl1NamesList(exceptionInfo);
    const int curListSize = dynlen(currentList);
    for (int board = 1; board <= curListSize; ++board) {
      //Check all the names in the library and check if they are still present in the new server list.
      if ( 0 == dynContains(toBeAddedUkl1List, currentList[board]) ) {
	//If they are not present they need to be deleted.
	toBeDeletedUkl1List[numUkl1Added] = currentList[board];
      }
    }
  }// if (refresh).
  else {
    //In this case we are starting afresh and can just remove any existing settings from 
    toBeDeletedUkl1List = fwUkl1_getUkl1NamesList(exceptionInfo);
    //Will return an empty list in the even of an error. We will just have to see why later.
  }// else (refresh)

  //We will also delete any incomplete boards that are found in the system. They can always go.
  //Give all the incomplete names to the delete FSM DU and HW DP functions as they can handle a name
  //not exisiting.
  //Just return emtpy arrays in the event of failure so don't worry about errors.
  dynAppend(toBeDeletedUkl1List, fwUkl1_getUkl1HwDpNamesOnlyList(exceptionInfo));
  dynAppend(toBeDeletedUkl1List, fwUkl1_getUkl1FsmTreeNamesOnlyList(exceptionInfo));

  //Delete those boards that need to be deleted.
  //Must be done before creation as we need to ensure for the reload case the list is empty.
  if ( 0 != dynlen(toBeDeletedUkl1List) ) {
    //Only call delete if we have to.
    DebugTN("fwUkl1_findAndAddUkl1Boards(): Deleting the following UKL1 boards from the system:\n" + toBeDeletedUkl1List);
    //Exception info won't be checked after this call as we will continue anyway, just let the panel deal with the exceptions.
    fwUkl1_deleteUkl1Boards(toBeDeletedUkl1List, exceptionInfo);
  }

  //Add those boards that need to be added.
  if ( 0 != dynlen(toBeAddedUkl1List) ) {
    DebugTN("fwUkl1_findAndAddUkl1Boards(): Adding the following UKL1 boards from the system:\n" + toBeAddedUkl1List);
    //Exception info won't be checked after this call as we will continue anyway, just let the panel deal with the exceptions.
    fwUkl1_addUkl1Boards(toBeAddedUkl1List, versionList, deleteExisting, forceReconfigure, exceptionInfo);
  }

  //Found and added all the boards!
  return;
}

/*!
 * Checks the given CCPC to see if it attached to UKL1 board. It identifies the board by the serial number accessible via
 * the I2C bus.
 *
 * \param  ccpcName Name of the CCPC as found in the DIM server.
 * \param  isUkl1 Returned by reference, there are three possible values:
 *   \li FWUKL1_PRODUCTION The board is a UKL1 production board.
 *   \li FWUKL1_PROTOTYPE The board is a UKL1 prototype board.
 *   \li 0 The board is not a UKL1.
 * \param  serialNumber Returned by reference, the serial number of the found board. Returns:
 *   \li Prototype in the event that it is a prototype without a working PROM.
 *   \li Null if the board is not a UKL1.
 *   \li Fail if an error occurred.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * There exists a UKL1 prototype that did not have a working PROM and as such no serial number can be read. In this case the
 * an exception is returned, but isUkl1 will indicate that it is a prototype and will return `PROTOTYPE' for the serial number.
 * This should be treated with suspicion especially if it is known that there are no prototype boards in the system. It will
 * eventually be `retired' and this function will need to be updated.
 */
void fwUkl1_isUkl1Board(string ccpcName, unsigned& isUkl1, string& serialNumber, dyn_string& exceptionInfo) {
  //Attempt to read the serial number from the given CCPC. All boards should have this identifier in the same place.
  //This cannot be done through the fwUkl1_getUkl1SerialNumber function as that assumes the existence of the relevant
  //hardware data point. We can't assume it exists as we want to check if the board is a UKL1 before creating it!
  //Use the fwCcpc library I2C read function. It should be on bus 0 at address 0x50, sub-address 0. It is 32-bits big (4-bytes)
  //and we will assume an infinite page size as it doesn't really matter.
  //The bus is not constant as when dealing with prototypes it could be 1!
  int bus=0;
  const int address=0x50, subAddress=0, size=4, pageSize=-1;
  //We wish the negative acknowledge on address to be enabled.
  const bool nack = TRUE;
  //Holds the read data.
  dyn_char ddcData;
  int status1 = fwCcpc_I2CReadSub(ccpcName, bus, address, 0, size, pageSize, FWCCPC_I2C_COMBINED, nack, ddcData);
  if ( 0 == status1 ) {
    //Successfully read the serial number from the CCPC. Is it an UKL1?
    const unsigned ccpcSerNum = fwCcpc_convertHexToDec( fwUkl1_convertByteToHexLittleEndian(ddcData[1]), exceptionInfo );
    //If the word is byte swapped then the revision will occupy the 3rd nibble.
    //This we don't care about when identifying the board, hence mask it.
    if ( (0x601 == (ccpcSerNum>>20)) || (0x1060 == (ccpcSerNum&0xF0FF)) ) {
      //Check whether the serial number indicates that it is a production or prototype board.
      if ( (0x00200000 == (ccpcSerNum&0x00F00000)) || (0x200 == (ccpcSerNum&0xF00)) ) {
	//Production.
	isUkl1 = FWUKL1_PRODUCTION;
      } else {
	//Prototype
	isUkl1 = FWUKL1_PROTOTYPE;
      }
      //Convert the serial number to a string.
      serialNumber = fwCcpc_convertDecToHex(ccpcSerNum);
    } else {
      //The CCPC is not attached to a UKL1, so we must tell everyone.
      isUkl1 = 0;
      serialNumber = "NULL";
    }
  }//if(0==status1)
  else {
    //We failed to read from the I2C bus at the appropriate location.
    //This could be because some of the prototype boards had a problem that meant it was moved to bus 1.
    //Try to read again, but this time on bus 1.
    bus = 1;
    dynClear(ddcData);
    int status2 = fwCcpc_I2CReadSub(ccpcName, bus, address, subAddress, size, pageSize, FWCCPC_I2C_COMBINED, nack, ddcData);
    if ( 0 == status2 ) {
      //Successfully read the serial number from the CCPC. Is it an UKL1?
      const unsigned ccpcSerNum = fwCcpc_convertByteToDec(ddcData);
      if ( (0x60100000 == (ccpcSerNum>>20)) || (0x1060 == (ccpcSerNum&0xF0FF)) ) {
	//Check whether the serial number indicates that it is a production or prototype board.
	if ( (0x00200000 == (ccpcSerNum&0x00F00000)) || (0x200 == (ccpcSerNum&0xF00)) ) {
	  //Production.
	  isUkl1 = FWUKL1_PRODUCTION;
	} else {
	  //Prototype
	  isUkl1 = FWUKL1_PROTOTYPE;
	}
	//Convert the serial number to a string.
	serialNumber = fwCcpc_convertDecToHex(ccpcSerNum);
      } else {
	//The CCPC is not attached to a UKL1, so we must tell everyone.
	isUkl1 = 0;
	serialNumber = "Null";
      }
    }//if(0==status2)
    else {
      //This could be a broken prototype board, indicate that we think that and return the appropriate serial number,
      //but do flag it as an error.
      isUkl1 = 1;
      serialNumber = "Prototype";
      fwException_raise(exceptionInfo, "WARNING", "fwUkl1_isUkl1Board(): A board that did not have a PROM at I2C bus address 0x50 was identified as a prototype UKL1 board.","2");
    }//else(0==status2)
  }//else(0==status1)

  return;
}

/*!
 * Adds a list of UKL1 boards to the system. It creates the hardware datapoint, subscribing it to the DIM server;
 * and the finite statemachine datapoint, connecting the callback function for state changes. The names of those
 * added are put in the library list of DIM server names.
 *
 * \param  tobeAddedUkl1List A dyn_string of names of the UKL1 boards as they appear in the DIM server that need to
 *           be added to the system. If a name is already found in the system it will be removed and then readded.
 * \param  versionList FWUKL1_PRODUCTION if the UKL1 board to be added is a production version, FWUKL1_PROTOTYPE
 *           if it is a prototype version.
 * \param  deleteExisting If the datapoint is found to exist it will be deleted and then recreated with the relevant
 *           settings. Default: FALSE.
 * \param  forceReconfigure If TRUE it will force the settings to be manually set and not to be loaded from the defaults
 *           for the fwHw datapoint type. FALSE and the settings are just loaded from the defaults. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_addUkl1Boards(dyn_string toBeAddedUkl1List, dyn_uint versionList, bool deleteExisting, bool forceReconfigure, dyn_string& exceptionInfo) {
  //Current length of the exception info, used to determine if more errors have appeared.
  int exceptSize = dynlen(exceptionInfo);

  //Store the number of elements in the list.
  const unsigned numToBeAdded = dynlen(toBeAddedUkl1List);
  //If the list was empty return!
  if ( 0 == numToBeAdded )
    return;

  //Loop over each board adding them to the system.
  for (unsigned board = 1; board <= numToBeAdded; ++board) {
    const string ukl1Name = toBeAddedUkl1List[board];
    const unsigned ukl1Version = versionList[board];
    //Create the hardware type datapoint and subscribe it to the DIM server.
    //We want to just load the new datapoint from the HW type defaults and if it does exist then lets delete it.
    _fwUkl1_createHwTypeCCPCUKL1Datapoint(ukl1Name, ukl1Version, deleteExisting, forceReconfigure, exceptionInfo);
    if ( exceptSize < dynlen(exceptionInfo) ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_addUkl1Boards(): Failed to setup " + ukl1Name + "'s connection with the DIM server. Communication with this UKL1 is not possible.", "2");
      exceptSize = dynlen(exceptionInfo);
    }

  }//for each board.

  //Add the boards as device units in the FSM tree.
  _fwUkl1_createUkl1FsmTreeDU(toBeAddedUkl1List, deleteExisting, exceptionInfo);
  if ( exceptSize < dynlen(exceptionInfo) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_addUkl1Boards(): Failed to add the boards to the FSM tree list.", "2");
  }
  
  return;

}

/*!
 * Deletes a list of UKL1 boards from the system. It removes the hardware datapoint, unsubscribing it from the DIM server;
 * and the finite statemachine datapoint, disconnecting the callback function for state changes. The names of those
 * added are removed from the list of UKL1 names stored in the library.
 *
 * \param  tobeDeletedUkl1List A dyn_string of names of the UKL1 boards as they appear in the DIM server that need to
 *           be removed from the system.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_deleteUkl1Boards(dyn_string toBeDeletedUkl1List, dyn_string& exceptionInfo) {
  //Store the number of elements in the list.
  const unsigned numToBeDeleted = dynlen(toBeDeletedUkl1List);
  //If the list was empty return!
  if ( 0 == numToBeDeleted )
    return;

  //Loop over each board removing them from the system.
  for (unsigned board = 1; board <= numToBeDeleted; ++board) {
    const string ukl1Name = toBeDeletedUkl1List[board];
    //Some datapoint and datapoint type related constants...
    //Name of the hardware type, as considered by the fwHw library.
    const string fwHwType = "UKL1";

    //Remove the hardware type datapoint and unsubscribes it from the DIM server.
    bool bStatus = fwHw_remove(fwHwType, ukl1Name);
    if ( !bStatus )
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_deleteUkl1Board(): Failed to remove " + ukl1Name + "'s DIM server subscriptions.", "2");

  }//for each board.

  //Remove the list of boards to the FSM tree.
  const int exceptSize = dynlen(exceptionInfo);
  _fwUkl1_deleteUkl1FsmTreeDU(toBeDeletedUkl1List, exceptionInfo);
  if ( exceptSize < dynlen(exceptionInfo) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_deleteUkl1Boards(): Failed to delete the boards from the FSM tree list.", "2");
  }
	
  return;
}

/*!
 * Returns the full list of names of UKL1 boards that are present in both the FSM tree and as a HWTypeCCPCUKL1 data point
 * and hence are fully accessible in the system.
 *
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_string List of all the names of the UKL1 board that has been retrieved.
 */
dyn_string fwUkl1_getUkl1NamesList(dyn_string& exceptionInfo) {
  //This will hold the list to that is returned.
  dyn_string namesList;

  //Get the list of HW DPs in the system.
  dyn_string hwList;
  if ( fwHw_get(FWUKL1_HW_TYPE, hwList) ) {
    //Now we have the hardware list check to see if it is a node. We don't care if it is local or remote, both are >0 so just check that.
    const int numHw = dynlen(hwList);
    //This keeps track of the index in the names list that we wish to add elements for.
    int nameIndex = 1;
    for (int hw = 1; hw <= numHw; ++hw) {
      if ( 0 < fwFsmTree_isNode(hwList[hw]) ) {
	//Add it to the list.
	namesList[nameIndex] = hwList[hw];
	++nameIndex;
      }//if isNode.
      //No need for else we don't keep track here of incomplete names.
    }//for hw.
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_getUkl1NamesList(): Failed to retrieve hardware for hardware type " + FWUKL1_HW_TYPE + ".", "2");
  }

  //Done.
  return namesList;
}

/*!
 * Returns number of UKL1 board names that are currently in the system.
 *
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return unsigned Number of UKL1 boards that are stored in the list.
 */
int fwUkl1_getNamesListSize(dyn_string& exceptionInfo) {
  return dynlen(fwUkl1_getUkl1NamesList(exceptionInfo));
}

/*!
 * Returns the names of the UKL1 board names that are current stored as hardware data points but are not present in the FSM tree.
 *
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_string List of all the names of the UKL1 board that are present only as hardware data points. Empty if there are none.
 */
dyn_string fwUkl1_getUkl1HwDpNamesOnlyList(dyn_string& exceptionInfo) {
  //This holds the list that is to be returned.
  dyn_string namesList;

  //Get the list of HW DPs in the system.
  dyn_string hwList;
  if ( fwHw_get(FWUKL1_HW_TYPE, hwList) ) {
    //Now we have the hardware list check to see if it is a node. 0 indicates that it is not.
    const int numHw = dynlen(hwList);
    //This keeps track of the index in the names list that we wish to add elements for.
    int nameIndex = 1;
    for (int hw = 1; hw <= numHw; ++hw) {
      if ( 0 == fwFsmTree_isNode(hwList[hw]) ) {
	//Add it to the list.
	namesList[nameIndex] = hwList[hw];
	++nameIndex;
      }//if is not node.
      //No need for else we don't keep track here of complete names.
    }//for hw.
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_getUkl1HwDpNamesOnlyList(): Failed to retrieve hardware for hardware type " + FWUKL1_HW_TYPE + ".", "2");
  }

  //Hopefully this is empty.
  return namesList;
}

/*!
 * Returns the names of the UKL1 board names that are current present in the FSM tree, but are not stored as hardware data points.
 *
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_string List of all the names of the UKL1 board that are present only as hardware data points. Empty if there are none.
 */
dyn_string fwUkl1_getUkl1FsmTreeNamesOnlyList(dyn_string& exceptionInfo) {
  //This holds the list that is to be returned.
  dyn_string namesList;

  //Get the list of children in the FSM tree node.
  //The function also returns by reference the type of each child, CU, LU or DU. We don't care, should only ever be a DU.
  dyn_int junkChildrenTypes;
  dyn_string fsmList = fwCU_getChildren(junkChildrenTypes, FWUKL1_CU);
  //There is a special FSM child called `FWUKL1_CU'_confDB, which should not be returned.
  //It will never exist in the hardware, but should always exist in the FSM tree.
  int confDbIndex = dynContains(fsmList, (FWUKL1_CU + "_ConfDB"));
  dynRemove(fsmList, confDbIndex);
  //Now get the list of HW DPs for comparison with.
  dyn_string hwList;
  if ( fwHw_get(FWUKL1_HW_TYPE, hwList) ) {
    //How many FSM children are there? Required to know if we are to loop over them.
    const int numChildren = dynlen(fsmList);
    //This keeps track of the index in the names list that we wish to add elements for.
    int nameIndex = 1;
    for (int child = 1; child <= numChildren; ++child) {
      if ( 0 == dynContains(hwList, fsmList[child]) ) {
	//The FSM child is not present as a harware DP.
	namesList[nameIndex] = fsmList[child];
	++nameIndex;
      }//if is not in the hardware list.
      //No need for else we don't keep track here of complete names.
    }//for children.
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_getUkl1FsmTreeNamesOnlyList(): Failed to retrieve hardware for hardware type " + FWUKL1_HW_TYPE + ".", "2");
  }

  //Hopefully this is empty.
  return namesList;
}

// ============
//   INTERNAL
// ============

/*!
 * This is used to add list of UKL1 boards to the appropriate FSM tree. It will add the UKL1 boards as device units (DUs) under
 * the control unit (CU) defined by FWUKL1_CU.
 *
 * \param  ukl1List Names of UKL1 boards to be added to the FSM tree. Should be as they appear in the DIM server.
 * \param  deleteExisting If TRUE then if the DU is found in the tree it will be deleted before it is readded. If FALSE then the settings
 *           for the DU will just be reset. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * The DUs added to the tree will appear as their DIM server name, with the panel fwUkl1DUOverview.pnl added for
 * User interaction with the hardware. The label for the DUs will be the name of the UKL1 board as it appears in the DIM server.\n
 * The FSM CU node may have to be stopped for this function to complete and the FWUKL1_CU node will be restarted once the
 * operation is complete. No other nodes in the tree will be affected.\n
 * The behaviour of this function is dependent upon the value of deleteExisting, the behaviour is described below:
 *   \li Creation - If the DU does not exist it will be created with the settings as described above.
 *         In this case the FSM tree node is modified and requires it to be stopped and restarted.
 *   \li Hard recreation - DU exists, deleteExisting=TRUE. The DU is first deleted and then the function then proceeds as for creation.
 *         This requires the FSM tree node to be stopped and restarted.
 *   \li Soft recreation - DU exists, deleteExisting=FALSE. This does not require the FSM tree node to be stopped or restarted, it will
 *         simply ensure that the User panel and label are set correctly. Default: FALSE.
 *
 * Note that if the CU (as defined by FWUKL1_CU) does not exist then it will be created and the function will proceed as for creation.
 * The function will also refresh the FSM tree list in the device editor and navigator (DEN) and also will display the CU panel which
 * will refresh it if the list changed.
 */
void _fwUkl1_createUkl1FsmTreeDU(dyn_string ukl1List, bool deleteExisting, dyn_string& exceptionInfo) {
  //Don't do anything if the list is empty.
  if ( 0 == dynlen(ukl1List) ) {
    return;
  }

  //It is possible to call this function and for it to have no effect on the FSM tree node. In this case there is
  //no point in stopping, regenerating and restarting the FSM tree node. This variable is set to true if a restart
  //is required.
  bool restart = FALSE;
  //This flag is used to determine if we should create the DU or not.
  bool createDu = FALSE;
  //This flag is used to signal if there is no CU.
  bool noCU = FALSE;
  //First we must determine if the FWUKL1_CU node exists in the FSM tree.
  const int nodeExists = fwFsmTree_isNode(FWUKL1_CU);
  //If the node does not exist then we must create it.
  if ( 0 == nodeExists ) {
    //Stop the FSM tree to do this. Can't have been stopped by this stage.
    fwFsmTree_stopTree();
    //Tree was stopped will need to regenerate and restart.
    restart = TRUE;
    //Create the node as a control unit (last arg=1) and put it under the root node `FSM'.
    const string createdNode = fwFsmTree_addNode("FSM", FWUKL1_CU, "DAQ_Domain_v1", 1);
    if ( "" != createdNode ) {
      //It created the node and now we must set the label and UI panel.
      fwFsmTree_setNodeLabel(FWUKL1_CU, "RICH L1");
      fwFsmTree_setNodePanel(FWUKL1_CU, "fwUkl1CUOverview.pnl");
    } else {
      //Tell everyone we have no CU.
      noCU = TRUE;
      //Failed to create the node, update the status value and tell everyone.
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createUkl1FsmTreeDU(): Failed to create the control unit in the FSM state machine.", "1");
    }
  }//if(0==nodeExists)

  //The next section can be done regardless of whether the node exist section was called or not,
  //but if it failed then we don't have a CU FSM tree node to add the DUs to so check the status is still good.
  if ( !noCU ) {
    //Loop over all the names given in the ukl1List and determine what is the appropriate behaviour for each.
    const int numUkl1s = dynlen(ukl1List);
    for (int ukl1 = 1; ukl1 <= numUkl1s; ++ukl1) {
      //Get the name of the UKL1 board as it will appear in the FSM.
      const string ukl1FsmName = ukl1List[ukl1];
      //Check whether the DU exists.
      const int duExists = fwFsmTree_isNode(ukl1FsmName);
      if ( 0 == duExists ) {
	//If the node does not exist then we must create it.
	createDu = TRUE;
      }//if(0==duExists)
      else {
	//Determine whether or not to delete the DU before it is recreated.
	if ( deleteExisting ) {
	  //Remove the DU, which will require the node to be stopped first if it has not already been stopped.
	  if (!restart) {
	    //Must always stop everything.
	    fwFsmTree_stopTree();
	    //We are going to have to restart now.
	    restart = TRUE;
	  }
	  //Remove the DU from the tree, accept the default of removing a node recursively from the tree as the DU shouldn't have any children!
	  fwFsmTree_removeNode(FWUKL1_CU, ukl1FsmName);
	  //Check the DU was actually deleted.
	  const int duDeleted = fwFsmTree_isNode(ukl1FsmName);
	  if ( 0 == duDeleted ) {
	    //DU was deleted so create it.
	    createDu = TRUE;
	  } else {
	    //The DU still exists in the tree when it shouldn't. Warn everyone and update the status.
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_createUkl1FsmTreeDU(): Failed to delete " + ukl1Name + " from the FSM tree. UKL1 boards may exist in the list that no longer exist in the system.", "2");
	  }
	}//if(deleteExisting)
      }//else(0==duExists)
	
      //If we need to create the DU then we best had.
      if ( createDu ) {
	//We need to be stopped by here if we are not already.
	if (!restart) {
	  //Must always stop everything.
	  fwFsmTree_stopTree();
	  //We are going to have to restart now.
	  restart = TRUE;
	}
	//Add the board to the tree as a HwTypeCCPCUKL1 type and as a device unit (last arg=0).
	fwFsmTree_addNode(FWUKL1_CU, ukl1FsmName, "HwTypeCCPCUKL1", 0);
      }//if(createDu)
	
      //Check to see if the node was added. Done for all cases.
      const int duPresent = fwFsmTree_isNode(ukl1FsmName);
      if ( 0 != duPresent ) {
	//Added the board :) Now set the label and the User panel.
	fwFsmTree_setNodeLabel(ukl1FsmName, ukl1List[ukl1]);
	fwFsmTree_setNodePanel(ukl1FsmName, "fwUkl1/fwUkl1DUOverview.pnl");
      } else {
	//Failed to add the DU. Warn people and update the status appropriately.
	fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createUkl1FsmTreeDU(): Failed to add " + ukl1Name + " to the FSM tree. Communication with this board will not be possible via FSM.", "2");
      }
    }//for(numUkl1s)
  }//if(noCU) Don't bother with an else we already have the appropriate status and will have signaled the error.

  //We should only regenerate and restart the FSM tree node if it has actually been stopped and modified.
  if (restart) {
    //Commented functions should not need to be performed left here for reference...
    //fwFsmTree_generateAll();
    //Generate the FSM types.
    //fwFsmTree_generateFsmTypes();
    //Generate the FSM for CU
    fwFsmTree_generateTreeNode(FWUKL1_CU);
    // Start FSM processes for the CU sub-tree
    fwFsmTree_startTree();
  }

  //Refresh the Tree in case the DEN is running, this can be done regardless. It ensures that the tree display is correct.
  fwFsmTree_refreshTree();
  //Don't try and view the CU panel as if it is run in a script in the console it has no concept of a UI and ths can't start it.


  //Now we are done.
  return;

}

/*
 * This is used to delete a list UKL1 boards from the appropriate FSM tree. It will look in the tree FWUKL1_CU
 *
 * \param  ukl1List List of names of UKL1 boards to be deleted from the FSM tree. Names should be as they appear in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 */
void _fwUkl1_deleteUkl1FsmTreeDU(dyn_string ukl1List, dyn_string& exceptionInfo) {
  if ( 0 == dynlen(ukl1List) ) {
    //Nothing to do so just exit.
    return;
  }

  //First we must determine if the FWUKL1_CU node exists in the FSM tree.
  const int nodeExists = fwFsmTree_isNode(FWUKL1_CU);
	
  //Does the node we wish to delete from exist?
  if ( 0 != nodeExists ) {
    //The node exists, either locally or remotely.
    //Stop it so we can edit it, accept the default of stopping all children.
    fwFsmTree_stopTree();
    //Loop over all the names given in the ukl1List and determine what is the appropriate behaviour for each.
    const int numUkl1s = dynlen(ukl1List);
    for (int ukl1 = 1; ukl1 <= numUkl1s; ++ukl1) {
      //Convert the ukl1Name into the name as it will appear in the FSM.
      const string ukl1FsmName = ukl1List[ukl1];
      //Check whether the DU exists.
      const int duExists = fwFsmTree_isNode(ukl1FsmName);
      if ( 0 != duExists ) {
	//Remove the DU from the tree, accept the default of removing a node recursively from the tree as the DU shouldn't have any children!
	fwFsmTree_removeNode(FWUKL1_CU, ukl1FsmName);
	//Check the DU was actually deleted.
	const int duDeleted = fwFsmTree_isNode(ukl1FsmName);
	if ( 0 != duDeleted ) {
	  //The DU still exists in the tree when it shouldn't. Warn everyone and update the status.
	  fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_deleteUkl1FsmTreeDU(): Failed to delete " + ukl1Name + " from the FSM tree. There is no physical board in the crate, but it will still appear in the FSM tree.", "2");
	}
      }//if(0==duExists)
      //No else, if it doesn't exist then we don't care about it. We could just be asking for the removal of a hardware only board.
    }//for(numUkl1s)

    //Only any point of doing this if the node exists.
    //Generate the FSM for CU.
    fwFsmTree_generateTreeNode(FWUKL1_CU);
    //Start FSM processes for the CU sub-tree. We could have deleted only a couple of boards.
    fwFsmTree_startTreeNode(FWUKL1_CU);
    //Refresh the Tree in case the DEN is running, this can be done regardless.
    fwFsmTree_refreshTree();
    //Don't try and reopen the CU, we might not be able to achieve that if we are running in a console script.
  }//if(0!=nodeExists)
  else {
    //The node does not exist so we cannot delete any boards. Warn people, it will happen on a fresh install though.
    fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_deleteUkl1FsmTreeDU(): No UKL1 node found. Cannot delete device units. If this is a clean install this not a problem.", "1");
  }//else(0!=nodeExists)


  return;
}

/*!
 * This will create a datapoint from the HwTypeCCPCUKL1 datapoint type, which represents a UKL1 board, with all
 * the appropriate settings and subscribe these registers to the DIM server.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version If FWUKL1_PRODUCTION it is a production board, if it is FWUKL1_PROTOTYPE it is a prototype board,
 *           this affects the I2C bus that needs to be used.
 * \param  deleteExisting If the datapoint is found to exist it will be deleted and then recreated with the relevant
 *           settings.
 * \param  forceReconfigure If TRUE it will force the settings to be set from hardcoded settings and not to be loaded from
 *           the defaults for the fwHw datapoint type, it will also reset the default settings to those hardcoded in the library.
 *	     FALSE and the settings are just loaded from the defaults.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This function has the following possible ways of operating:
 *    \li Hard creation - Datapoint does not exist and the datapoint is created, the settings are loaded from the library
 *          and saved to defaults, all registers are subscribed to the DIM server. forceReconfigure=TRUE, deleteExisting has
 *          no relevance in this case.
 *    \li Soft creation - Datapoint does not exist and the datapoint is created, the settings are loaded from the
 *          defaults, all registers are subscribed to the DIM server. forceReconfigure=FALSE, deleteExisting has
 *          no relevance in this case.
 *    \li Full hard recreation - Datapoint exists, deleteExisting=TRUE, forceReconfigure=TRUE the datapoint registers
 *          are unsubscribed from the DIM server and then the datapoint is deleted. It then proceeds as for
 *          hard creation.
 *    \li Full soft recreation - Datapoint exists, deleteExisting=TRUE, forceReconfigure=FALSE the datapoint registers
 *          are unsubscribed from the DIM server and then the datapoint is deleted. It then proceeds as for
 *          soft creation.
 *    \li Hard recreation - Datapoint exists, deleteExisting=FALSE, forceReconfigure=TRUE, all the settings and their
 *          defaults are reset to the settings as for a full creation and DIM server subscriptions are recreated.
 *          The datapoint is never deleted.
 *    \li Soft recreation - Datapoint exists, deleteExisting=FALSE, forceReconfigure=FALSE, all the settings are reloaded
 *          from their default values. The datapoint is never deleted and the DIM server subscriptions are not recreated.
 *
 * The name of the created datapoint will be the concatantion of FWUKL1_DPNAME_PREFIX and ukl1Name. The setting placed in this
 * datapoint set the name of the UKL1 board as in the DIM server and the address of all the registers that
 * the datapoint can access. Only status registers are set up for monitoring and the datapoints do not contain
 * any values, hence none are written to the board. Note that no inverse is provided for this function
 * i.e._fwUkl1_deleteHwTypeCCPCUKL1Datapoint as this is done via fwHw_remove.
 */
void _fwUkl1_createHwTypeCCPCUKL1Datapoint(string ukl1Name, unsigned ukl1Version, bool deleteExisting, bool forceReconfigure, dyn_string& exceptionInfo) {
  if ( "" == ukl1Name ) {
    //Nothing to delete.
    return;
  }

  //Some datapoint and datapoint type related constants...
  //Name of the hardware type, as considered by the fwHw library.
  const string fwHwType = "UKL1";
  //Name of the datapoint type we are creating from.
  const string dptName = FWHW_TYPE_DPT + FWHW_CCPC_DESCRIPTION + fwHwType;
  //Name of the default settings data point.
  const string defaultSettingsDpName = FWHW_DEFAULT_SETTINGS_DP + FWHW_CCPC_DESCRIPTION + "UKL1";

  //Check to see if the default settings datapoint actually exists. If it doesn't then we should add it
  //to the system, otherwise it will cause much confusion later... especially as there is no error checking
  //associated with loading/saving to the default values!
  if ( !dpExists(defaultSettingsDpName) ) {
    //Potentially override the User choice, but then we must...
    forceReconfigure = TRUE;
    //Create the hardware datapoint, but make sure we don't try and load the default settings as we are creating them.
    if ( fwHw_create(fwHwType, defaultSettingsDpName, FALSE) ) {
      //Successfully recreated the defaults datapoint, better warning people it happened as it is a bit odd.
      fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Had to recreate the UKL1 hardware data point's default settings. This indicates that the install may not have proceeded correctly, this specific error should not cause any further problems it may indicate a more serious problem however.", "2");
    } else {
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to recreate the default settings for the UKL1 hardware data point. This should not have to have needed to be done in the first place (install of fwUkl1 may not have completed successfully) and failure to create it indicates a more serious problem.", "2");
    }
  }

  //Now check to see if the datapoint exists.
  if ( dpExists(ukl1Name) ) {
    //Datapoint exists. There are three possible cases that need to be handled here...

    //Soft recreation.
    if ( !deleteExisting && !forceReconfigure ) {
      //No real need to add anything to the exception here.
      _fwUkl1_configureHwTypeCCPCUKL1Settings(ukl1Name, ukl1Version, FALSE, exceptionInfo);

      //Ensure the DIM services are setup. If they are it will be ignored otherwise they will be set up.
      if ( !fwHw_subscribe(ukl1Name) )
	fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to subscribe " + ukl1Name + " to the DIM server, communication with this UKL1 board will not be possible.", "2");

    }//if ( !deleteExisting && !forceReconfigure )

    //Hard recreation.
    else if ( !deleteExisting && forceReconfigure ) {
      //Reconfigure all the settings from fwUkl1 lib values and save as defaults.
      _fwUkl1_configureHwTypeCCPCUKL1Settings(ukl1Name, ukl1Version, TRUE, exceptionInfo);

      //Now resubscribe to the server.
      if ( !fwHw_subscribe(ukl1Name) )
	fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to subscribe " + ukl1Name + " to the DIM server, communication with this UKL1 board will not be possible.", "2");

    }//else if ( !deleteExisting && forceReconfigure )

    //Full hard&soft recreation, hardness determined by forceReconfigure during recreation.
    else if ( deleteExisting ) {
      //Here we want to delete the existing datapoint and recreate it.
      //fwHw_remove deletes the hardware data point and completely unsubscribes it from the DIM server.
      bool status = fwHw_remove(fwHwType, ukl1Name);
      if ( !status )
	fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to delete " + ukl1Name + "'s hardware data point. Attempting to reconfiguring and use the existing data point.", "2");

      //Whether we deleted the data point or not we should try and recreate it by calling ourselves again.
      //Behaviour depends on forceReconfigure, we definately don't want to try and delete the data point.
      //If we do try and delete the data point we could potentially get into an infinite loop if some how it isn't removed.
      //By not asking to delete it, if it isn't removed then we can just reconfigure it and won't get to this stage again.
      //Used to check no exceptions occurred during this function call.
      const int exceptSize = dynlen(exceptionInfo);
      _fwUkl1_createHwTypeCCPCUKL1Datapoint(ukl1Name, ukl1Version, FALSE, forceReconfigure, exceptionInfo);
      if ( exceptSize < dynlen(exceptionInfo) )
	fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Problems were encountered while recreating " + ukl1Name + "'s hardware datapoint. Check previous errors for details.", "2");

      //Just return as we have done everything we want to.
      return;
    }// else if ( deleteExisting )

  }//if ( dpExists(ukl1Name) )

  else {
    //Hard creation.
    if ( forceReconfigure ) {
      //DP doesn't exist so create it, but don't reload from defaults we need to set those up first.
      if ( fwHw_create(fwHwType, ukl1Name, FALSE) ) {
	//Set all the values on the datapoint to something sensible, it will create a detailed enough exception tree.
	_fwUkl1_configureHwTypeCCPCUKL1Settings(ukl1Name, ukl1Version, TRUE, exceptionInfo);
      } else {
	fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to create the hardware data point for " + ukl1Name + ", communication with this board is not possible.", "1");
      }
    }// if ( forceReconfigure )

    //Soft creation
    else {
      //Use the hardware tool to create the datapoint and don't save to defaults.
      if ( fwHw_create(fwHwType, ukl1Name, FALSE) ) {
	//Set all the values on the datapoint to something sensible, it will create a detailed enough exception tree.
	_fwUkl1_configureHwTypeCCPCUKL1Settings(ukl1Name, ukl1Version, FALSE, exceptionInfo);
      } else {
	fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to create the hardware data point for " + ukl1Name + ", communication with this board is not possible.", "1");
      }
    }// else (forceReconfigure)

    //Now subscribe all the registers to the DIM server. DIM server will not do anything if the registers are already found to be subscribed.
    if ( !fwHw_subscribe(ukl1Name) ) {
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_createHwTypeCCPCUKL1Datapoint(): Failed to subscribe " + ukl1Name + " to the DIM server, communication with this board is not possible.", "2");
    }

  }// else ( dpExists(ukl1Name) )

  //If we got this far then everything must have been created and subscribed without issue.
  return;
}

/*!
 * This is used to configure the register settings of the HwTypeCCPCUKL1 datapoint type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server. The appropriate
 *           datapoint to be set will be found from this.
 * \param  ukl1Verision Verision number of the UKL1 board, either FWUKL1_PRODUCTION or FWUKL1_PROTOTYPE.
 * \param  saveAsDefaults If true then the settings loaded into the datapoint are also saved as
 *           its default values.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1Settings(string ukl1Name, unsigned ukl1Version, bool saveAsDefaults, dyn_string& exceptionInfo) {
  //Return status of the function.
  int status = 0;
  //Base name of the fwHwType that is being set.
  const string fwHwType = "UKL1";

  //Decide how best to load the settings.
  //This can be used to mark that we failed to load from the default settings and should try to load from hardware defaults.
  bool failedFromDefaults = FALSE;
  //This is the case if we wish to load from the default settings.
  //Can only be done for production boards, when we don't want to save the settings to the defaults.
  if ( !saveAsDefaults && (FWUKL1_PRODUCTION == ukl1Version) ) {
    //In this case we can just reload from the default settings.
    if ( !fwHw_applyDefaultSettings(fwHwType, ukl1Name) ) {
      failedFromDefaults = TRUE;
      fwException_raise(exceptionInfo, "WARNING", "Failed to apply default settings to " + ukl1Name + ", attempting to configure the board from hardcoded defaults.", "2");
    }
  }

  //This case is when we wish to reload the default settings or we wish to configure a prototype board.
  //The prototypes require a different I2C bus number to that stored in the default settings.
  //If we wish to reset the defaults to the hardcoded settings then we can't load from them!
  if ( saveAsDefaults || (FWUKL1_PROTOTYPE == ukl1Version) || failedFromDefaults ) {
    //Even if these functions fail we will try to configure the rest of the settings and can check the error
    //log higher up to see if enough failed to indicate a major problem.
    //This notes the size of the exception now so we can check if any more appeared in the log after all these function calls.
    const int exceptSize = dynlen(exceptionInfo);

    //Write the common settings to the all the datapoints on the board.
    _fwUkl1_configureHwTypeCCPCUKL1CommonSettings(ukl1Name, exceptionInfo);

    //Initialise the FSM state.
    _fwUkl1_configureHwTypeCCPCUKL1FsmState(ukl1Name, ukl1Version, exceptionInfo);

    //Now we must do the specific settings for all of the hardware type included in the UKL1 type.
    //GBE
    _fwUkl1_configureHwTypeCCPCUKL1GBESpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
    //TTCrx
    _fwUkl1_configureHwTypeCCPCUKL1TTCRXSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
    //FE FPGA
    _fwUkl1_configureHwTypeCCPCUKL1FeFpgaSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
    //BE FPGA
    _fwUkl1_configureHwTypeCCPCUKL1BeFpgaSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
    //UKL1 PROM
    _fwUkl1_configureHwTypeCCPCUKL1PROMSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
    //UKL1 temperature sensors.
    _fwUkl1_configureHwTypeCCPCUKL1TempSensorSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);

    //Check to see if any exceptions were risen.
    if ( exceptSize >= dynlen(exceptionInfo) ) {
      //No new exceptions were generated so set them as the defaults, if we are asked to.
      //Having done all that do we want to save them as the defaults?
      if (saveAsDefaults) {
	//Will get the settings that we have just set and save them to the default values.
	//At present it appears that this function just returns 1, update this to check for errors once it can actually show them!
	if ( 1 != fwHw_saveSettingsAsDefault(fwHwType, ukl1Name) )
	  fwException_raise(exceptionInfo, "WARNING", "fwUkl1_configureHwTypeCCPCUKL1Settings(): " + ukl1Name + "'s default settings were not saved correctly, this could mean they are corrupt and the data point should not be configured from the default settings until this problem is fixed.","2");
      }
    } else {
      fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_configureHwTypeCCPCUKL1Settings(): " + ukl1Name + " will operate with limited functionality as some (or all) of the registers were not configured correctly. DIM communication will not be possible for those that configuration failed.", "2");
      if (saveAsDefaults)
	fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_configureHwTypeCCPCUKL1Settings(): " + ukl1Name + "'s default settings were not saved as the data point from which the settings would be saved from was not configured correctly hence corrupt defaults could be created.", "2");
    }

  }

  //Configured the datapoint!
  return;

}

/*!
 * This will set the common settings for the fwHw type UKL1 and all included fwHw types.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1CommonSettings(string ukl1Name, dyn_string& exceptionInfo) {
  //The common setting to be written.
  dyn_string commonSettings;
  commonSettings[1] = ukl1Name;
  //Write the common setting, this is done recursively so all the subtypes have their common setting
  //updated to.
  if ( !fwHw_setCommonSettings(ukl1Name, commonSettings, TRUE) )
    fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1CommonSettings(): Failed to set the CCPC name for " + ukl1Name + ", this will prevent correct communication with this UKL1 board via the DIM server.", "1");
}

/*!
 * This will set the specific settings for the status data point element that the UKL1 HW type has.
 * It controls the FSM interactions. It should be ensured that the ukl1Name HW data point exists before calling this function.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1FsmState(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name and add the name of the DPE to the end.
  const string dpName = ukl1Name + ".status";
  //If reset that value back to 0, their return values match ours.
  if ( 0 != dpSet(dpName, FWUKL1_NOT_READY_STATE) )
    fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1FsmState(): Failed to initialise the UKL1 FSM state to NOT_READY for " + ukl1Name + ", problems accessing the DP were encountered.", "1");
}

/*!
 * This will set the specific settings for the fwHw type GBE that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1GBESpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".GBE";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1]  = "JTAGID";
  registerList[2]  = "MACSoftReset";
  registerList[3]  = "RxFifoPrtReset";
  registerList[4]  = "TxFifoPrtReset";
  registerList[5]  = "PortEnable";
  registerList[6]  = "RxFifoEnable";
  registerList[7]  = "Mode";
  registerList[8]  = "Clock";
  registerList[9]  = "SPI3ConfigTrnmtGlobal";
  registerList[10] = "SPI3ConfigRcv";
  registerList[11] = "RxStatus";
  registerList[12] = "PHYCtrl";
  registerList[13] = "PHYData";
  registerList[14] = "MDIOControl";
  //This contains the address of all the registers to be set.
  dyn_uint regAddrList;
  regAddrList[1]  = 0x50c;
  regAddrList[2]  = 0x505;
  regAddrList[3]  = 0x59e;
  regAddrList[4]  = 0x620;
  regAddrList[5]  = 0x500;
  regAddrList[6]  = 0x5b3;
  regAddrList[7]  = 0x501;
  regAddrList[8]  = 0x794;
  regAddrList[9]  = 0x700;
  regAddrList[10] = 0x701;
  regAddrList[11] = 0x793;
  regAddrList[12] = 0x680;
  regAddrList[13] = 0x681;
  regAddrList[14] = 0x683;
  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the address to be filled in later.
  //2nd element is the number of 32-bit words for command.
  //3rd element is the check for change rate, set to 2 to match Tell1 change rate and to be compatible with their panels.
  //4th element is send on change only setting.
  dyn_int specificSetting = makeDynAnytype(0x00, 1, 2, 1);
  dyn_dyn_int specificSettings;
  //Loop over all the registers
  for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
    //Put the appropriate register address in the settings array.
    specificSetting[1] = regAddrList[count];
    //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
    specificSettings[count] = specificSetting;
  }

  //Now write all these to the datapoint.
  //Set the general GBE port settings.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
    fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1GBESpecificSettings(): Failed to setup the specific settings for the GBE ethernet registers for " + ukl1Name + ", communication with these registers is not possible.", "1");

  //Now setup the PROM, this will indicate its own errors don't need to add anything here.
  _fwUkl1_configureHwTypeCCPCUKL1GBEPROMSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);
  //Now setup the four ports, this will indicate its own errors don't need to add anything here.
  _fwUkl1_configureHwTypeCCPCUKL1GBEPORTSpecificSettings(ukl1Name, exceptionInfo);

  //And we are done.
}

/*!
 * This will set the specific settings for the fwHw GBE PORM type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1GBEPROMSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".GBE.PROM";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1] = "Id";
  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the bus number that the PROM is on, always either 0 or 1 depending on the version.
  int bus = 0;
  if ( FWUKL1_PROTOTYPE == ukl1Version )
    bus = 1;
  //2nd element is the address of the PROM on the bus, always 0x57.
  //3rd element is the address of the PROM registers at register 0x57, always 0.
  //4th element is the number of bytes that should be written to that register, 46 for compatibility with Tell1.
  //5th element is the page size, again 46 for compatibility with Tell1.
  //6th element is the the type of I2C interface to use, always FWCCPC_I2C_COMBINED for the PROM.
  //7th element is the negative acknowledge, NACK, leave it enabled as I think it tells us if it failed. Always 1. 
  //8th element is the refresh rate, always 2.
  //9th element is the on data change settings always 1.
  dyn_int specificSetting = makeDynAnytype(bus, 0x57, 0, 46, 46, FWCCPC_I2C_COMBINED, 1, 2, 1);
  dyn_dyn_int specificSettings;
  specificSettings[1] = specificSetting;

  //Now write all these to the datapoint.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
     fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1GBEPROMSpecificSettings(): Failed to setup the specific settings for the GBE ethernet PROM for " + ukl1Name + ", communication with the PROM is not possible.", "1");

}

/*!
 * This will set the specific settings for the fwHw GBE PORT type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1GBEPORTSpecificSettings(string ukl1Name, dyn_string& exceptionInfo) {
  //Loop over the four ports.
  for (unsigned port = 0; port < FWUKL1_GBE_PORTS; ++port) {
    //Convert the UKL1 board name into the datapoint name.
    const string dpName = ukl1Name + ".GBE.PORT" + port;

    //Set the specific settings.
    //This contains a list of all the register names that are to be set.
    dyn_string registerList;
    registerList[1]  = "Duplex";
    registerList[2]  = "MAC";
    registerList[3]  = "Config";
    registerList[4]  = "TXFifoThreshold";
    registerList[5]  = "TXFifoLowWtrmrk";
    registerList[6]  = "TXFifoHighWtrmrk";
    registerList[7]  = "PHYControl";
    registerList[8]  = "PHYStatus";
    registerList[9]  = "TXStat";
    registerList[10] = "RXStat";

    //This contains the address of all the registers to be set.
    dyn_uint regAddrList;
    //Calculate the start address in the block that contains the per port settings.
    const unsigned offset = (0x80*(port));
    regAddrList[1]  = offset + 0x02;
    regAddrList[2]  = offset + 0x10;
    regAddrList[3]  = offset + 0x18;
    regAddrList[4]  = 0x614 + port;
    regAddrList[5]  = 0x60a + port;
    regAddrList[6]  = 0x600 + port;
    regAddrList[7]  = offset + 0x60;
    regAddrList[8]  = offset + 0x61;
    regAddrList[9]  = offset + 0x40;
    regAddrList[10] = offset + 0x20;

    //This contains the size in bytes of the register to be read.
    dyn_uint regSizeList;
    regSizeList[1]  = 1;
    regSizeList[2]  = 1;
    regSizeList[3]  = 1;
    regSizeList[4]  = 1;
    regSizeList[5]  = 1;
    regSizeList[6]  = 1;
    regSizeList[7]  = 1;
    regSizeList[8]  = 1;
    regSizeList[9]  = 25;
    regSizeList[10] = 26;

    //Now we create the settings, these are actually almost the same for each except for the address.
    //1st element is the address to be filled in later.
    //2nd element is the number of 32-bit words for command.
    //3rd element is the check for change rate, 4th is send on change only setting.
    dyn_int specificSetting = makeDynAnytype(0x00, 1, 2, 1);
    dyn_dyn_int specificSettings;
    //Loop over all the registers
    for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
      //Put the appropriate register address in the settings array.
      specificSetting[1] = regAddrList[count];
      //Put the appropriate size of the register in.
      specificSetting[2] = regSizeList[count];
      //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
      specificSettings[count] = specificSetting;
    }
    
    //Now write all these to the datapoint.
    if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1GBEPORTSpecificSettings(): Failed to setup the specific settings for the GBE ethernet port " + port + " for " + ukl1Name + ", communication with this port is not possible.", "1");

  }

}
  
/*!
 * This will set the specific settings for the fwHw IngressFPGA type that has been inserted into the
 * UKL1 fwHw type.
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1FeFpgaSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Loop over the ingress/FE FPGAs.
  for (unsigned feFpga = 0; feFpga < FWUKL1_FRONT_FPGAS; ++feFpga) {
    //Convert the UKL1 board name into the datapoint name.
    const string dpName = ukl1Name + ".FeFpga" + feFpga;

    //Set the specific settings.
    //This contains a list of all the register names that are to be set.
    dyn_string registerList;
    registerList[1] = "Disable";
    registerList[2] = "Emulator";
    registerList[3] = "EmulatorHpdPixelMode";
    registerList[4] = "EmulatorEvent";
    registerList[5] = "ResetStatusBufferPointer";
    //This contains the address of all the registers to be set.
    dyn_uint regAddrList;
    regAddrList[1] = fwUkl1_FpgaizeAddress(24);
    regAddrList[2] = fwUkl1_FpgaizeAddress(20);
    regAddrList[3] = fwUkl1_FpgaizeAddress(20);
    regAddrList[4] = fwUkl1_FpgaizeAddress(20);
    regAddrList[5] = fwUkl1_FpgaizeAddress(21);
    //Now we create the settings, these are actually almost the same for each except for the address.
    //1st element is the address to be filled in later.
    //2nd element is the number of 32-bit words for command.
    //3rd element is the check for change rate (2secs), 4th is send on change only setting.
    dyn_int specificSetting = makeDynAnytype(0x00, FWCCPC_LBUS_BITS_32, 1, 2, 1);
    dyn_dyn_int specificSettings;
    //Loop over all the registers
    for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
      //Put the appropriate register address in the settings array.
      specificSetting[1] = regAddrList[count];
      //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
      specificSettings[count] = specificSetting;
    }

    //Now write all these to the datapoint.
    if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1FeFpgaSpecificSettings(): Failed to setup the specific settings for FE FPGA number " + feFpga + " for " + ukl1Name + ", communication with this FPGA is not possible.", "1");
    
    //Now setup the nine channels, don't bother dealing with their error messages it is dealt with higher up.
    _fwUkl1_configureHwTypeCCPCUKL1ChannelSpecificSettings(ukl1Name, feFpga, exceptionInfo);

  }//for feFpga

  return;
}

/*!
 * This will set the specific settings for the fwHw InputChannel type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  feFpga Number of the FE/ingress FPGA that the channels are being configured for.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1ChannelSpecificSettings(string ukl1Name, unsigned feFpga, dyn_string& exceptionInfo) {
  //Loop over the input channels.
  for (unsigned channel = 0; channel < FWUKL1_FPGA_CHANNELS; ++channel) {
    //Convert the UKL1 board name into the datapoint name.
    const string dpName = ukl1Name + ".FeFpga" + feFpga + ".Channel"+ channel;

    //Set the specific settings.
    //This contains a list of all the register names that are to be set.
    dyn_string registerList;
    registerList[1] = "Disable";

    //This contains the address of all the registers to be set.
    dyn_uint regAddrList;
    regAddrList[1] = fwUkl1_FpgaizeAddress(25+feFpga);
    //Now we create the settings, these are actually almost the same for each except for the address.
    //1st element is the address to be filled in later.
    //2nd element is the number of 32-bit words for command.
    //3rd element is the check for change rate (2secs), 4th is send on change only setting.
    dyn_int specificSetting = makeDynAnytype(0x00, FWCCPC_LBUS_BITS_32, 1, 2, 1);
    dyn_dyn_int specificSettings;
    //Loop over all the registers
    for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
      //Put the appropriate register address in the settings array.
      specificSetting[1] = regAddrList[count];
      //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
      specificSettings[count] = specificSetting;
    }
    
    //Now write all these to the datapoint.
    if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1ChannelSpecificSettings(): Failed to setup the specific settings for FE FPGA " + feFpga + ", channel " + channel + " for " + ukl1Name + ", communication with this FPGA is not possible.", "1");
  }//for channel

}

/*!
 * This will set the specific settings for the fwHw EgressFPGA type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1BeFpgaSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".BeFpga";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1]  = "MacDestinationAddress";
  registerList[2]  = "MacSourceAddress";
  registerList[3]  = "IpDestinationAddress";
  registerList[4]  = "IpSourceAddress";
  registerList[5]  = "Protocol";
  registerList[6]  = "TypeOfService";
  registerList[7]  = "TimeToLive";
  registerList[8]  = "Type";
  registerList[9]  = "Version";
  registerList[10] = "FixedMepDestinationAddress";
  registerList[11] = "DontWaitForMepDestBrdcst";
  registerList[12] = "Throttle";
  registerList[13] = "ThrottlePolarity";
  registerList[14] = "TfcDecoder";
  registerList[15] = "PollGbePorts";
  registerList[16] = "PartitionId";
  registerList[17] = "MepPackingFactor";
  registerList[18] = "ZeroSuppression";
  registerList[19] = "HpdPixelMode";
  registerList[20] = "Latency";
  registerList[21] = "L1Id";
  registerList[22] = "DataFragmentSize";

  //This contains the address of all the registers to be set.
  dyn_int regAddrList;
  regAddrList[1]  = fwUkl1_FpgaizeAddress(0);
  regAddrList[2]  = fwUkl1_FpgaizeAddress(3);
  regAddrList[3]  = fwUkl1_FpgaizeAddress(6);
  regAddrList[4]  = fwUkl1_FpgaizeAddress(8);
  regAddrList[5]  = fwUkl1_FpgaizeAddress(10);
  regAddrList[6]  = fwUkl1_FpgaizeAddress(11);
  regAddrList[7]  = fwUkl1_FpgaizeAddress(12);
  regAddrList[8]  = fwUkl1_FpgaizeAddress(13);
  regAddrList[9]  = fwUkl1_FpgaizeAddress(14);
  regAddrList[10] = fwUkl1_FpgaizeAddress(15);
  regAddrList[11] = fwUkl1_FpgaizeAddress(15);
  regAddrList[12] = fwUkl1_FpgaizeAddress(15);
  regAddrList[13] = fwUkl1_FpgaizeAddress(15);
  regAddrList[14] = fwUkl1_FpgaizeAddress(15);
  regAddrList[15] = fwUkl1_FpgaizeAddress(15);
  regAddrList[16] = fwUkl1_FpgaizeAddress(16);
  regAddrList[17] = fwUkl1_FpgaizeAddress(18);
  regAddrList[18] = fwUkl1_FpgaizeAddress(19);
  regAddrList[19] = fwUkl1_FpgaizeAddress(19);
  regAddrList[20] = fwUkl1_FpgaizeAddress(22);
  regAddrList[21] = fwUkl1_FpgaizeAddress(23);
  regAddrList[22] = fwUkl1_FpgaizeAddress(29);

  //This contains a list of the number of 16-bit words per register.
  dyn_uint byteNumList;
  byteNumList[1]  = 3;
  byteNumList[2]  = 3;
  byteNumList[3]  = 2;
  byteNumList[4]  = 2;
  byteNumList[5]  = 1;
  byteNumList[6]  = 1;
  byteNumList[7]  = 1;
  byteNumList[8]  = 1;
  byteNumList[9]  = 1;
  byteNumList[10] = 1;
  byteNumList[11] = 1;
  byteNumList[12] = 1;
  byteNumList[13] = 1;
  byteNumList[14] = 1;
  byteNumList[15] = 1;
  byteNumList[16] = 2;
  byteNumList[17] = 1;
  byteNumList[18] = 1;
  byteNumList[19] = 1;
  byteNumList[20] = 1;
  byteNumList[21] = 1;
  byteNumList[22] = 1;

  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the address to be filled in later.
  //2nd element is the number of 32-bit words for command.
  //3rd element is the check for change rate (2secs), 4th is send on change only setting.
  dyn_uint specificSetting = makeDynAnytype(0x00, FWCCPC_LBUS_BITS_32, 1, 2, 1);
  dyn_dyn_uint specificSettings;
  //Loop over all the registers
  for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
    //Put the appropriate register address in the settings array.
    specificSetting[1] = regAddrList[count];
    //Put the appropriate size of register into the array.
    specificSetting[3] = byteNumList[count];
    //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
    specificSettings[count] = specificSetting;
  }

  //Setup the BE FPGA status registers, don't worry about the errors from this they will be logged appropriately.
  _fwUkl1_configureHwTypeCCPCUKL1BeFpgaStatusSpecificSettings(ukl1Name, ukl1Version, exceptionInfo);

  //Now write all the read write register settings to the datapoint.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1BeFpgaSpecificSettings(): Failed to setup the specific settings for the BE FPGA for " + ukl1Name + ", communication with this FPGA is not possible.", "1");

}

/*!
 * This will set the specific settings for the fwHw EgressFPGAStatus type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1BeFpgaStatusSpecificSettings(string ukl1Name, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".BeFpga.BeFpgaStatus";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1]  = "TTCId";
  registerList[2]  = "TfcFifoRequest";
  registerList[3]  = "EventMergerReadState";

  registerList[4]  = "SPI3TxBusy";
  registerList[5]  = "MepFifoFull";
  registerList[6]  = "Throttled";

  registerList[7]  = "TfcL0Reset";
  registerList[8]  = "TfcL1Reset";
  registerList[9]  = "TfcBShort";
  registerList[11] = "Tpa";
  registerList[12] = "FragmentReady";
  registerList[13] = "MepReady";
  registerList[14] = "FragmentRequest";
  registerList[15] = "EventQValid";

  registerList[16] = "FragmenterState";

  registerList[17] = "EthernetQValid";
  registerList[18] = "SequenceError";
  registerList[19] = "MepOccupancy";
  registerList[20] = "TfcFifoOccupany";


  //This contains the address of all the registers to be set.
  dyn_int regAddrList;
  regAddrList[1]  = fwUkl1_FpgaizeAddress(32);
  regAddrList[2]  = fwUkl1_FpgaizeAddress(32);
  regAddrList[3]  = fwUkl1_FpgaizeAddress(32);

  regAddrList[4]  = fwUkl1_FpgaizeAddress(34);
  regAddrList[5]  = fwUkl1_FpgaizeAddress(34);
  regAddrList[6]  = fwUkl1_FpgaizeAddress(34);

  regAddrList[7]  = fwUkl1_FpgaizeAddress(35);
  regAddrList[8]  = fwUkl1_FpgaizeAddress(35);
  regAddrList[9]  = fwUkl1_FpgaizeAddress(35);
  regAddrList[10] = fwUkl1_FpgaizeAddress(35);
  regAddrList[11] = fwUkl1_FpgaizeAddress(35);
  regAddrList[12] = fwUkl1_FpgaizeAddress(35);
  regAddrList[13] = fwUkl1_FpgaizeAddress(35);
  regAddrList[14] = fwUkl1_FpgaizeAddress(35);
  regAddrList[15] = fwUkl1_FpgaizeAddress(35);

  regAddrList[16] = fwUkl1_FpgaizeAddress(36);

  regAddrList[17] = fwUkl1_FpgaizeAddress(37);
  regAddrList[18] = fwUkl1_FpgaizeAddress(37);
  regAddrList[19] = fwUkl1_FpgaizeAddress(37);
  regAddrList[20] = fwUkl1_FpgaizeAddress(37);
  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the address to be filled in later.
  //2nd element is the number of 32-bit words for command.
  //3rd element is the rate at which it checks for change, 2secs always
  //4th is send on change only setting, again set.
  dyn_uint specificSetting = makeDynAnytype(0x00, FWCCPC_LBUS_BITS_32, 1, 2, 1);
  dyn_dyn_uint specificSettings;
  //Loop over all the registers
  for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
    //Put the appropriate register address in the settings array.
    specificSetting[1] = regAddrList[count];
    //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
    specificSettings[count] = specificSetting;
  }

  //Now write all these to the datapoint.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1BeFpgaStatusSpecificSettings(): Failed to setup the specific settings for the BE FPGA for " + ukl1Name + ", communication with this FPGA is not possible.", "1");

}

/*!
 * This will set the specific settings for the fwHw EEPORM type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 *           documentation) are used.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1PROMSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".PROM";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1] = "Id";
  //The bus number that the PROM is on is determined by whether it is production or prototype.
  //Zero for production
  int bus = 0;
  //One for prototype.
  if ( FWUKL1_PROTOTYPE == ukl1Version )
    bus = 1;
  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the bus number that the PROM is on, dependent on whether it is a production or prototype board.
  //2nd element is the address of the PROM on the bus, always 0x50.
  //3rd element is the address of the PROM registers at register 0x50, always 0.
  //4th element is the number of bytes that should be written to that register, 4 which is the number of bytes stored in the PROM.
  //5th element is the page size, 4 the size of the number of bytes. Probably.
  //6th element is the the type of I2C interface to use, always FWCCPC_I2C_COMBINED for the PROM.
  //7th element is the negative acknowledge, NACK, leave it enabled as I think it tells us if it failed. Always 1. 
  //8th element is the refresh rate, always 2. Monitoring must be enabled to use this setting.
  //9th element is the on data change settings always 1. Monitoring must be enabled to use this setting.
  dyn_int specificSetting = makeDynAnytype(bus, 0x50, 0, 4, 4, FWCCPC_I2C_COMBINED, 1, 2, 1);
  dyn_dyn_int specificSettings;
  specificSettings[1] = specificSetting;
  
  //Now write all these to the datapoint.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1PROMSpecificSettings(): Failed to setup the specific settings for the UKL1 PROM for " + ukl1Name + ", communication with the PROM is not possible.", "1");

  return;
}

/*!
 * This will set the specific settings for the fwHw TEMPSENS types that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 *           documentation) are used.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1TempSensorSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Loop over all the available temperature sensors and configure them.
  for (int tempSens = 1; tempSens <= FWUKL1_TEMPERATURE_SENSORS; ++tempSens) {
    //Convert the UKL1 board name into the datapoint name.
    const string dpName = ukl1Name + ".Temp" + tempSens;

    //Set the specific settings.
    //This contains a list of all the register names that are to be set.
    dyn_string registerList;
    registerList[1] = "TempRead";
    //1st element is the bus number that the PROM is on, dependent on whether it is a production or prototype board.
    //Zero for production
    int bus = 0;
    //One for prototype.
    if ( FWUKL1_PROTOTYPE == ukl1Version )
      bus = 1;
    //2nd element is the address of the temperature sensor on the bus, varies depending on which sensor it is:
    unsigned address;
    if ( 1 == tempSens )
      address = 0x48;
    else if ( 2 == tempSens )
      address = 0x4a;
    else if ( 3 == tempSens )
      address = 0x4c;
    //3rd element is the address of the temperature data on the register, always 0.
    //4th element is the number of bytes that should be read from the register, 1 as the temperature is stored as an 8-bit number.
    //5th element is the page size, -1 which is infinite and we can ignore it.
    //6th element is the the type of I2C interface to use, always FWCCPC_I2C_COMBINED for the temperature sensor.
    //7th element is the negative acknowledge, NACK, leave it enabled as I think it tells us if it failed. Always 1. 
    //8th element is the refresh rate, always 2. Monitoring must be enabled to use this setting.
    //9th element is the on data change settings always 1. Monitoring must be enabled to use this setting.
    dyn_int specificSetting = makeDynAnytype(bus, address, 0, 1, -1, FWCCPC_I2C_COMBINED, 1, 2, 1);
    dyn_dyn_int specificSettings;
    specificSettings[1] = specificSetting;
  
    //Now write all these to the datapoint.
    if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
      fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1TempSensorSpecificSettings(): Failed to setup the specific settings for the temperature sensor number " + tempSens + " for " + ukl1Name + ", communication with the sensor is not possible.", "1");

  }
}

/*!
 * This will set the specific settings for the fwHw TTCRX type that has been inserted into the
 * UKL1 fwHw type.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  ukl1Version Indicates whether we have a production or prototype board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_configureHwTypeCCPCUKL1TTCRXSpecificSettings(string ukl1Name, unsigned ukl1Version, dyn_string& exceptionInfo) {
  //Convert the UKL1 board name into the datapoint name.
  const string dpName = ukl1Name + ".TTCRX";

  //Set the specific settings.
  //This contains a list of all the register names that are to be set.
  dyn_string registerList;
  registerList[1]  = "IACId";
  registerList[2]  = "I2CId";
  registerList[3]  = "FineDelay1";
  registerList[4]  = "FineDelay2";
  registerList[5]  = "CoarseDelay";
  registerList[6]  = "Control";
  registerList[7]  = "Status";
  registerList[8]  = "Config1";
  registerList[9]  = "Config2";
  registerList[10] = "Config3";
  registerList[11] = "SingleErrCnt";
  registerList[12] = "DoubleErrCnt";
  registerList[13] = "SEUErrCnt";
  registerList[14] = "BunchCnt";
  registerList[15] = "EventCnt";
  //This contains the address of all the registers to be set.
  dyn_uint regAddrList;
  regAddrList[1]  = 16;
  regAddrList[2]  = 18;
  regAddrList[3]  = 0;
  regAddrList[4]  = 1;
  regAddrList[5]  = 2;
  regAddrList[6]  = 3;
  regAddrList[7]  = 22;
  regAddrList[8]  = 19;
  regAddrList[9]  = 20;
  regAddrList[10] = 21;
  regAddrList[11] = 8;
  regAddrList[12] = 10;
  regAddrList[13] = 11;
  regAddrList[14] = 24;
  regAddrList[15] = 26;
  //This contains the number of bytes that need to be set for the each of the registers.
  dyn_uint numBytes;
  numBytes[1]  = 2;
  numBytes[2]  = 1;
  numBytes[3]  = 1;
  numBytes[4]  = 1;
  numBytes[5]  = 1;
  numBytes[6]  = 1;
  numBytes[7]  = 1;
  numBytes[8]  = 1;
  numBytes[9]  = 1;
  numBytes[10] = 1;
  numBytes[11] = 2;
  numBytes[12] = 1;
  numBytes[13] = 1;
  numBytes[14] = 2;
  numBytes[15] = 3;

  //Now we create the settings, these are actually almost the same for each except for the address.
  //1st element is the bus number that the TTCrx is on,determined by whether it is a prototype or not.
  //Zero for production
  int bus = 0;
  //One for prototype.
  if ( FWUKL1_PROTOTYPE == ukl1Version )
    bus = 1;
  //2nd element is the address of the TTCrx on the bus, again always 0x42 the address the register addresses should be written to.
  //3rd element is the address of the TTCrx registers on the TTCrx chip, variable.
  //4th element is the number of bytes that should be written to that register, again variable.
  //5th element is the page size, don't really know but if -1 it is infinite and we can ignore it.
  //6th element is the the type of I2C interface to use, always FWCCPC_I2C_SEPARATED for the TTCrx chip.
  //7th element is the negative acknowledge, NACK, leave it enabled as I think it tells us if it failed. Always 1. 
  //8th element is the refresh rate, always 2. Have to enable monitoring to eactually use this setting.
  //9th element is the on data change settings always 1. Have to enable monitoring to eactually use this setting.
  dyn_int specificSetting = makeDynAnytype(bus, 0x42, 0, 0, -1, FWCCPC_I2C_SEPARATED, 1, 2, 1);
  dyn_dyn_int specificSettings;
  //Loop over all the registers
  for (unsigned count = 1; count <= dynlen(regAddrList); ++count) {
    //Put the appropriate register address in the settings array.
    specificSetting[3] = regAddrList[count];
    specificSetting[4] = numBytes[count];
    //Now put the specific settings in the appropriate element of the array containing all the settings for each register.
    specificSettings[count] = specificSetting;
  }

  //Now write all these to the datapoint.
  if ( !fwHw_setSpecificSettings(dpName, registerList, specificSettings) )
    fwException_raise(exceptionInfo, "ERROR", "_fwUkl1_configureHwTypeCCPCUKL1TtcrxSpecificSettings(): Failed to setup the specific settings for the UKL1 TTCrx chip for " + ukl1Name + ", communication with the TTCrx is not possible.", "1");

}

//@}


// ================
//  UKL1 INTERFACE
// ================

/** @defgroup SectionInterface
 *    This section provides functions for communicating with the hardware on a specified UKL1
 *    board, with the board always identified by its name as found in the DIM server.
 *    It provides the ability to access all registers on the hardware and to perform other
 *    hardware related tasks such as resetting or reloading the UKL1s. It also provides
 *    access to the recipes saved for the UKL1 hardware type. This is done via the UKL1
 *    board name as found in the DIM server and the name of the specific recipe that is being
 *    address. Recipe and hardware access are setup such that they are almost identical function
 *    prototypes and it is only the addition of the recipe name and choice of verifying the hardware
 *    writes that causes them to be different. This hides the subtlies of writing to the hardware
 *    of database from the User.\n
 *
 *    There are two methods for reading the values a register. It can either be done directly
 *    or via monitoring.
 *    In the case of direct reading the function fwCcpc_read() or fwHw_getRecipe() is used and the
 *    values returned by this function are parsed by fwUkl1_parseReadRegister()to remove the relevant
 *    bits from the register. It removes a continuous sections of bits and returns it as a hex string,
 *    this can then be converted into an appropriate display value. Further multiple sections of the
 *    defined size can be removed, via fwUkl1_parseReadRegisterSections().
 *    In the case of monitoring callback functions can be setup in order to parse the data of a
 *    given data point only when it changes and again fwUkl1_parseReadRegister can be used to
 *    manipulate the register data to get at the desired information or fwUkl1_parseReadRegisterSections()
 *    if data is to be removed for example in the case of FE FPGA broadcast data. A read, using fwCcpc_read,
 *    can be used to force a reset of all monitored value, in this case the data returned by
 *    fwCcpc_read can just be ignored.
 *    This method is recommended for dealing with status registers.
 *    The library provides a function to read all the registers for a given area of the UKL1
 *    e.g. all the BE FPGA register, and then returns them in a predefined formatted state.
 *    This is to be used in the case that monitoring is not required or the User does not wish
 *    to know or does not know the data is organised across the registers. The values returned
 *    will be in strings will often contain a description of the state rather than just a number.
 *    Again values for similar element e.g. the input channel disable states can be returned
 *    as comma separated values.
 *    This method is recommended for reading from configuration registers.\n
 *
 *    The library provides a mechanism for writing to the registers on the UKL1. It provides a
 *    wrapper for the fwCcpc_write mechanism however it provides a layer of abstraction that does
 *    not require the User to understand the layout of the registers and which registers contain
 *    multiple settings. It takes a string that typically describes the registers state and is
 *    identical to the value read in the case of the read/write configuration registers. This
 *    function is advised to be used when writing to the configuration registers.
 *
 *    All errors are returned by the framework defined exception handling mechanism taking an
 *    argument exceptionInfo by reference, which will contain details of an errors that occurred.
 *  @{
 */

// ==========================
//   DATA PARSING FUNCTIONS
// ==========================

// =============
//    GENERAL
// =============

/*!
 * The purpose of this function is to take the data read from the UKL1 registers and to allow multiple groups of continuous
 * bits to be return in the form of a integer array. The function poses the ability to deal with registers of any size e.g. the
 * 32-bit registers from the GBE or FPGAs, or the 8-bit registers from the I2C. It is also capable of dealing with the format
 * of the FPGA registers as read from the UKL1s and a boolean switch it used to indicate that these must be handled. Typically
 * this will be used with monitored registers.
 *
 * \param  dcData Data read from the register as a byte array.
 * \param  numSections This is the number of times the defined section of data is to be read. It should be be defined by the
 *           the library constants that define the number of components on the board. For example FWUKL1_FRONT_FPGAS to retrieve
 *           the same setting for all the FE FPGAs when stored in the same register. This is set to 1 when only a single section
 *           need be removed. Each section is returned as an element in returned array. The first element contains the bits
 *           at startBit then subsequent elements contain the sections starting at the next section of bits.
 * \param  sectionGap The number of bits between the last bit in a section and the first bit in the next section.
 * \param  startBit Number of the first bit to be read from the register, where bit 0 is the least significant bit.
 * \param  length Number of bits to be read, set to 1 if only startBit is to be read. Maximum 32.
 * \param  fpga Boolean value to indicate whether we are parsing FPGA registers or not. This must be set correctly otherwise
 *           the data will not be properly interpreted.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return int Unsigned integer representation of the bits if the function is successful or -1 if it fails.
 *
 * This function primarily designed to be use with registers that contain multiple settings in a single register that are a constant
 * number of bits apart and are the same number of bits in size. This is most effective when dealing with the FE FPGA or channel
 * registers to implement a broadcast mode.\n
 * The reason for the FPGA flag is that the UKL1 FPGA registers are read as if they were 32-bit registers and not the 16-bit registers
 * that they are physically (this is because sequential addresses are 32-bits apart and not 16!). This function will remove the extra two
 * junk words that are read and parse only the remaining data in this case. Do not remove these extra words manually before hand.
 * This function can return sections that are a maximum of 32-bits in length. The only registers that are bigger than this are the
 * MAC addresses and these have their own dedicated convertion functions. The input size of the register is unlimited.
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  //Fictional UKL1 board name, as found in the DIM server.
 *  const string ukl1Name = "pcukl101";
 *  //Register name needs to be the full DP name, hence requires the HW DP name prefix defined by this library.
 *  dyn_string registerNames;
 *  registerNames[1] = FWUKL1_HW_DP_NAME_PREFIX + ukl1Name + ".FeFpga0.disable"
 *  //Stores the read data.
 *  dyn_char readData;
 *  //Holds errors.
 *  dyn_int callStatusList;
 *  //Read and save the status.
 *  const bool readStatus = fwCcpc_read(registerNames, ddcData, callStatusList);
 *  //Check for errors...
 *  if (!readStatus)
 *    (...)
 *
 *  //Data read from the disable register on the FE FPGA contains the disable settings for all the FE FPGA disable setting.
 *  //we split this data up into 4 settings, thus requiring only one read to get all the data.
 *  //Number of sections wanted is the number of FE FPGAs.
 *  const unsigned numSections = FWUKL1_FRONT_FPGAS;
 *  //The settings are in adjacent bits.
 *  const unsigned sectionGap = 0;
 *  //The data is a single bit, the first one starting at bit zero.
 *  const unsigned startBit = 0;
 *  const unsigned length   = 1;
 *  //The it is a setting on an FPGA register.
 *  const bool fpga = TRUE;
 *  //Check for errors.
 *  dyn_string exceptionInfo;
 *  //Convert the data.
 *  const dyn_int feFpgaDisableData = fwUkl1_parseReadRegisterSections(ddcData[1], numSections, sectionGap, startBit, length, fpga, exceptionInfo);
 *  //Check for errors.
 *  if ( 0 < dynlen(exceptionInfo) )
 *    (...)
 *
 *  //Let us deal with all four settings.
 *  for (int feFgpa = 1; feFpga <= FWUKL1_FRONT_FPGAS; ++feFpga) {
 *    if ( 1 == feFpgaDisableData[feFpga] )
 *      (...disabled code...)
 *    else if ( 0 == feFpgaDisableData[feFpga] )
 *      (...enabled code...)
 *    else
 *      (...-1 would be seen if an error occur,
 *          which should have been handled after function call,
 *          need to check anyway otherwise it might be dealt with
 *          by an above case...)
 *
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
 */
dyn_int fwUkl1_parseReadRegisterSections(const dyn_char& dcData, unsigned numSections, unsigned sectionGap, unsigned startBit, unsigned length, bool fpga, dyn_string& exceptionInfo) {
  //This will hold the output data.
  dyn_int output;
  //Loop over the number of sections that we wish to get.
  for (int sec = 1; sec <= numSections; ++sec) {
    //First get the requested bits.
    output[sec] = fwUkl1_parseReadRegister(dcData, startBit, length, fpga, exceptionInfo);

    //Move the start bit to the start of the next section.
    //Add each time the total number of bits that need to be read and skipped.
    //Again minus 1 because of different bases, 0 and 1, for bit position and number of bits.
    startBit += (length-1 + sectionGap);
  }

  //Not going to concern ourselves with errors as the parseReadRegister() will tag them appropriately.

  //Return the data.
  return output;
}

/*!
 * The purpose of this function is to take the data read from the UKL1 registers and to allow a specific group of continuous
 * bits to be returned as a integer. The function poses the ability to deal with registers of any size e.g. the
 * 32-bit registers from the GBE or FPGAs, or the 8-bit registers from the I2C. It is also capable of dealing with the format
 * of the FPGA registers as read from the UKL1s and a boolean switch it used to indicate that these must be handled. Typically
 * this will be used with monitored registers.
 *
 * \param  dcData Data read from the register as a byte array.
 * \param  startBit Number of the first bit to be read from the register, where bit 0 is the least significant bit.
 * \param  length Number of bits to be read, set to 1 if only startBit is to be read. Maximum 32.
 * \param  numSections This is the number of times the defined section of data is to be read. It should be be defined by the
 *           the library constants that define the number of components on the board. For example FWUKL1_FRONT_FPGAS to retrieve
 *           the same setting for all the FE FPGAs when stored in the same register. This is set to 1 when only a single section
 *           need be removed. Multiple sections are separated by commas.
 * \param  fpga Boolean value to indicate whether we are parsing FPGA registers or not. This must be set correctly otherwise
 *           the data will not be properly interpreted.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return int Unsigned integer representation of the bits if the function is successful or -1 if it fails.
 *
 * The reason for the FPGA flag is that the UKL1 FPGA registers are read as if they were 32-bit registers and not the 16-bit registers
 * that they are physically (this is because sequential addresses are 32-bits apart and not 16!). This function will remove the extra two
 * junk words that are read and parse only the remaining data in this case. Do not remove these extra words manually before hand.
 * This function can return sections that are a maximum of 32-bits in length. The only registers that are bigger than this are the
 * MAC addresses and these have their own dedicated convertion functions. The input size of the register is unlimited.
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  //Fictional UKL1 board name, as found in the DIM server.
 *  const string ukl1Name = "pcukl101";
 *  //Register name needs to be the full DP name, hence requires the HW DP name prefix defined by this library.
 *  dyn_string registerNames;
 *  registerNames[1] = FWUKL1_HW_DP_NAME_PREFIX + ukl1Name + ".BeFpga.ZeroSuppression"
 *  //Stores the read data.
 *  dyn_char readData;
 *  //Holds errors.
 *  dyn_int callStatusList;
 *  //Read and save the status.
 *  const bool readStatus = fwCcpc_read(registerNames, ddcData, callStatusList);
 *  //Check for errors...
 *  if (!readStatus)
 *    (...)
 *
 *  //Data read from the zero suppression register contains also the HPD pixel mode data which we are not interested in,
 *  //hence this function will return only the zero suppression data.
 *  //Zero suppression data spans 4 bits, starting with the least significant bit.
 *  const unsigned startBit = 0;
 *  const unsigned length   = 4;
 *  //The zero suppression setting is on a FPGA register.
 *  const bool fpga = TRUE;
 *  //Check for errors.
 *  dyn_string exceptionInfo;
 *  //Convert the data.
 *  const int zeroSuppData = fwUkl1_parseReadRegister(ddcData[1], startBit, length, fpga, exceptionInfo);
 *  //Check for errors.
 *  if ( 0 < dynlen(exceptionInfo) )
 *    (...)
 *
 *  //If zero suppression is enabled then 0xf would be returned.
 *  if ( 0xf == zeroSuppData )
 *    (...zero suppressed code...)
 *  else if ( 0 == zeroSuppData )
 *    (...none zero suppressed code...)
 *  else
 *    (...-1 would be seen if an error occur,
 *        which should have been handled after function call,
 *        need to check anyway otherwise it might be dealt with
 *        by an above case...)
 *
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
 */
int fwUkl1_parseReadRegister(dyn_char dcData, unsigned startBit, unsigned length, bool fpga, dyn_string& exceptionInfo) {
  if ( 32 < length ) {
    fwException_raise(exceptionInfo, "WARNING", "fwUkl1_parseReadRegister(): Maximum length of returned bits is 32-bits, requested length was " + length + ", which is too large. Reducing length to 32-bits, returned value will be clipped at 32-bits.", "2");
    length = 32;
  }

  //First we must remove the extra bits.
  //Check that we don't gain any more exception data...
  const int exceptSize = dynlen(exceptionInfo);
  //If we are dealing with an FPGA the we need to convert from 16- to 32-bit words.
  if (fpga) {
    fwUkl1_convert32To16BitWords(dcData, exceptionInfo);
    if ( exceptSize < dynlen(exceptionInfo) ) {
      //Return that we failed, indicate that it was this function that called the byte removal function before it failed.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseReadRegiser(): Aborted function, failed to properly parse the FPGA register data.", "2");
      return -1;
    }
  }

  //In order to be able to span byte boundaries we need to know where we are going to start and finish in terms
  //of the appropriate byte. These constants will ease the navigation of the bit-field.

  //Determine how many bits we have, there are eight bits in each element as each element contains a char.
  const int numBytes = dynlen(dcData);
  if ( -1 == numBytes ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseReadRegister(): Aborted function, failed to determine number of bytes in the input array.", "2");
    return -1;
  }
  //How many bits in total were we given.
  const unsigned numBits  = 8 * numBytes;
  //The last bit is the number of bits we wish to remove offset by the starting bit number.
  //We subtract 1 as if we just remove 1 bit then the start and end bit are the same.
  unsigned endBit = startBit + length - 1;
  if ( numBits <= endBit ) {
    //Can't end at a point passed the number of bits that we have.
    //The last bit is 1 less than the number of bits in the register as we start counting at 0.
    fwException_raise(exceptionInfo, "WARNING", "fwUkl1_parseReadRegiser(): Last bit to be read (" + endBit + ") is beyond the last bit in the register (" + (numBits-1) + "). Reducing last bit to be read to " + (numBits-1) + ", returned value will be clipped.", "2");
    //numBits is starts at 1, whereas endBit starts at 0.
    endBit = numBits -1;
  }
  //The first byte that we need to read from. Remembering that we are using a big endian numbering scheme.
  const unsigned startByte = numBytes - (unsigned)startBit/8;
  //Which is the first bit we need to read from in the starting byte.
  const unsigned startBitInByte = (unsigned)startBit%8;
  //This is the last byte that we need to read from.
  const unsigned endByte   = numBytes - (unsigned)endBit/8;
  //This is the last bit in the last byte that we need to read from.
  const unsigned endBitInByte = (unsigned)endBit%8;

  //This are used to enable us to write to the output data word.
  //The word that will be returned.
  unsigned uData = 0;
  //Marks where the next region of bits should start from in the word.
  unsigned nextBitInWord = 0;

  //Now we need to loop from the startByte to the endByte, decrement as we are big endian.
  for (unsigned byte = startByte; byte >= endByte; --byte) {
    //There are three cases that we need to consider.
    if ( startByte == endByte ) {
      //In this case we know are working within the same byte so we just align the desired start bit to the start of the byte.
      //Then create a mask of 1's of the numbers of bits we wish to extract and the operate on the bit aligned byte.
      uData |= (dcData[byte]>>startBitInByte) & ~(~0 << length);
    }//if(startByte==endByte)
    else if ( byte == startByte ) {
      //In this case we need to add from the startBit to the end of the byte to the output word.
      //Align the startBit with the beginning of the byte, don't need to mask as we want the rest of the byte to be output.
      uData |= (dcData[byte]>>startBitInByte) << nextBitInWord;
      //Advance the nextBit so where we know where to OR the next words into.
      nextBitInWord += (8-startBit);
    }//else if(byte==startByte)
    else if ( byte == endByte ) {
      //In this case we need to add from the first bit in the byte to the endBit.
      //Create a mask that has 1's covering the bits from the start bit in the byte to the endBitInByte and then operate on the bit.
      //Plus 1 because the shift is 1 based and the bit counting 0.
      uData |= (dcData[byte] & ~(~0 << (endBitInByte+1))) << nextBitInWord;
    }//else if(byte==endByte)
    else {
      //This case we can just take the whole byte and write it into the appropriate place in our word.
      //First convert is to a 32-bit number so that is can be bit shifted appropriately i.e. to bits higher than 8.
      uData |= (unsigned)dcData[byte] << nextBitInWord;
      //Have advanced 8 bits.
      nextBitInWord += 8;
    }//else
  }

  return uData;
}

/*!
 * Takes a colon delimited hex MAC address, as a string, and removes the colons
 * to return a continuous hex string. It also performs checks on the data format.
 *
 * \param  macAddress The string that contains the MAC address to be parsed.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string MAC address without the colons. If the MAC address contains no colons
 *           it will be returned as it was. If the MAC address does not look like a MAC
 *           address `Fail' will be returned.
 *
 * Expects the MAC address to be of the form 00:0e:0c:b0:2d:19 or 000e0cb02d19
 * i.e. 12 characters long.
 * \b Example
 * \code
 *   (...)
 *   //A colon delimited MAC address that we wish to convert.
 *   const string macAddress1 = 00:0e:0c:b0:2d:19;
 *   //A colonless MAC address.
 *   const string hexAddress1 = fwUkl1_convertMacToHex(macAddress1, exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Will not have any error...
 *   DebugTN(hexAddress1);  //000e0cb02d19
 *
 *   //Invalid colon separated MAC address.
 *   const string macAddress2 = 00:0e:0c:b0;
 *   const string hexAddress2 = fwUkl1_convertMacToHex(macAddress2, exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle error...
 *   DebugTN(hexAddress2);  //Fail
 *   (...)
 *  \endcode
 */
string fwUkl1_convertMacToHex(const string& macAddress, dyn_string& exceptionInfo) {
  //Check that it contains the appropriate number of characters, either with or without colons.
  if ( (12 != strlen(macAddress)) && (17 != strlen(macAddress)) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): MAC address is " + strlen(macAddress) + " characters long. Should be 12 or 17 characters.", "1");
    return "Fail";
  }

  //Will hold the MAC address once it has been parsed.
  string parsedMacAddress;
  //Try to split the string using a colon as a delimiter.
  dyn_string tmpMac = strsplit(macAddress, ":");
  //Check it has found the appropriate number of parts.
  const int lenMac = dynlen(tmpMac);
  if ( 1 == lenMac ) {
    //String contained no colons, it is the right lenght so it can be returned as it was given.
    parsedMacAddress = macAddress;
  }
  else if ( 6 == lenMac ) {
    //Contains 6 parts, put them all into one string.
    int status = sprintf(parsedMacAddress, "%02s%02s%02s%02s%02s%02s", tmpMac[1], tmpMac[2], tmpMac[3], tmpMac[4], tmpMac[5], tmpMac[6]);
    //Check for errors.
    if ( -1 == status ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): Failed to reconstruct MAC address as hex string.", "1");
      parsedMacAddress = "Fail";
    }
  }
  else {
    //Not properly formed.
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertMacToHex(): Found only " + lenMac + " colon delimited sections, should contain 1 or 6 for a valid MAC address.", "1");
    parsedMacAddress = "Fail";
  }

  return parsedMacAddress;

}

/*!
 * Takes an array of bytes and returns them as a colon delimited MAC address.
 *
 * \param  macAddress The array containing the bytes to be parsed.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string Expects a 6 element array containing 3 elements with 16-bit words of data in each.
 *           Returns `Fail' if the input data is misformed.
 *
 * This code requires that the extra bytes generated from a fwCcpc_read() are first removed, via fwUkl1_convert32To16BitWords().
 * Generally it is expected that this function will only be used internally and fwUkl1_parseBeFpgaReadRegisters() will be used if dealing
 * with the BE FPGA registers. It is not defined an internal function as it mirrors the convert MAC to byte function, which is not internal.
 *
 * \b Example
 * \code
 *   (...)
 *   //Fictional UKL1 board name, as found in the DIM server.
 *   const string ukl1Name = "pcukl101";
 *   //Register name needs to be the full DP name, hence requires the HW DP name prefix defined by this library.
 *   dyn_string registerNames;
 *   registerNames[1] = FWUKL1_HW_DP_NAME_PREFIX + ukl1Name + ".BeFpga.MacDestinationAddress"
 *   //Stores the read data.
 *   dyn_char readData;
 *   //Holds errors.
 *   dyn_int callStatusList;
 *   //Read and save the status.
 *   const bool readStatus = fwCcpc_read(registerNames, ddcData, callStatusList);
 *   //Check for errors...
 *   if (!readStatus)
 *     (...)
 *
 *   fwUkl1_convert32To16BitWord(ddcData[1], exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle errors, should be none if read successful...
 *
 *   const string macAddress = fwUkl1_convertByteToMac(ddcData[1], exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle errors, should be none if read successful...
 *   DebugTN(macAddress); //Colon delimited MAC address e.g. 00:0E:0C:B0:2D:19
 *
 *   (...)
 *  \endcode
 */
string fwUkl1_convertByteToMac(const dyn_char& macAddress, dyn_string& exceptionInfo) {
  //If the masks are in the data then they will be at the front i.e. lowest element numbers.
  //Determine the number of elements in the array and count backwards from there.
  //This will allow absolute indexing whether the masks are present or not.
  const int size = dynlen(macAddress);
  //Check that it contains the appropriate number of elements.
  //Technically if it came from the configuration database it could contain the write mask tagged on the end.
  //This doesn't cause a problem as we use absolute indexes later, just require there to be 6 elements minimum (and the MAC to be in those).
  if ( 6 > size ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToMac(): Input array contains " + size + " elements. Should be 6 or 12 elements big.", "1");
    return "Fail";
  }

  //Now we need to put the data into the appropriate format.
  string output;
  //This should be in a big endian order once it reaches here.
  output =  fwCcpc_convertByteToHex(macAddress[size - 5]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 4]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 3]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 2]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size - 1]);
  output += ":";
  output += fwCcpc_convertByteToHex(macAddress[size]);

  return output;

}

/*!
 * Takes a standard formed IP(4) address, as a string, and convert it to a continuous hex
 * string. Also performs syntacs checking on the given string.
 *
 * \param  ipAddress The string that contains the IP address to be parsed. Should be of the form:
 *           XXX.XXX.XXX.XXX, where X represents a decimal number, e.g. 192.168.192.192.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string If the IP address does not look like a IP address `Fail' will be returned, otherwise the hex string.
 *
 * Expects the MAC address to be of the form 192.168.255.255
 * \b Example
 * \code
 *   (...)
 *   //An IP address that we wish to convert.
 *   const string ipAddress1 = 192.168.255.0;
 *   //The equivalent hex string.
 *   const string hexAddress1 = fwUkl1_convertIpToHex(ipAddress1, exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Will not have any error...
 *   DebugTN(hexAddress1);  //c0a8ff00
 *
 *   //Invalid IP address.
 *   const string ipAddress2 = 192.168.;
 *   const string hexAddress2 = fwUkl1_convertIpToHex(ipAddress2, exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle error...
 *   DebugTN(hexAddress2);  //Fail
 *   (...)
 *  \endcode
 */
string fwUkl1_convertIpToHex(const string& ipAddress, dyn_string& exceptionInfo) {
  //Check that it is less than the maximum length of an IP(4) address.
  if ( 16 < strlen(ipAddress) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertIpToHex(): IP address length, " + strlen(ipAddress) + " is incorrect. Should be no more than 16 characters long.", "1");
    return "Fail";
  }
  //Try to split the string using a . as a delimiter.
  dyn_string tmpIp = strsplit(ipAddress, ".");
  //Check it has found the appropriate number of parts.
  const int lenIp = dynlen(tmpIp);
  if ( 4 == lenIp ) {
    //Contains 4 parts, put them all into one string.
    string output;
    for (unsigned element = 1; element <= dynlen(tmpIp); ++element) {
      //Take a section of the IP and convert it to a hex string.
      string tmpSection = fwCcpc_convertDecToHex(tmpIp[element]);
      //If it does not contain 2 characters pad the string.
      if ( 2 != strlen(tmpSection) )
	tmpSection = "0" + tmpSection;
      output += tmpSection;
    }
    return output;
  }
  else {
    //Not properly formed.
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertIpToHex(): Found " + lenIp + " full stop delimited sections, require 4.", "1");
    return "Fail";
  }

}

/*!
 * Takes an array of bytes, and returns them as a full stop delimited IP address.
 *
 * \param  ipAddress The array containing the bytes to be parsed.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string Expects a 4 element array. Returns `Fail' if the input data is missformed.
 *
 * This code requires that the extra bytes generated from a fwCcpc_read() are first removed, via fwUkl1_convert32To16BitWords().
 * Generally it is expected that this function will only be used internally and fwUkl1_parseBeFpgaReadRegister() will be used if dealing
 * with the BE FPGA registers. It is not defined an internal function as it mirrors the convert IP to byte function, which is not internal.
 *
 * \b Example
 * \code
 *   (...)
 *   //Fictional UKL1 board name, as found in the DIM server.
 *   const string ukl1Name = "pcukl101";
 *   //Register name needs to be the full DP name, hence requires the HW DP name prefix defined by this library.
 *   dyn_string registerNames;
 *   registerNames[1] = FWUKL1_HW_DP_NAME_PREFIX + ukl1Name + ".BeFpga.IpDestinationAddress"
 *   //Stores the read data.
 *   dyn_char readData;
 *   //Holds errors.
 *   dyn_int callStatusList;
 *   //Read and save the status.
 *   const bool readStatus = fwCcpc_read(registerNames, ddcData, callStatusList);
 *   //Check for errors...
 *   if (!readStatus)
 *     (...)
 *
 *   fwUkl1_convert32To16BitWord(ddcData[1], exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle errors, should be none if read successful...
 *
 *   const string ipAddress = fwUkl1_convertByteToIp(ddcData[1], exceptionInfo);
 *   if ( 0 != dynlen(exceptionInfo) )
 *     //...Handle errors, should be none if read successful...
 *   DebugTN(ipAddress); //Colon delimited MAC address e.g. 192.168.255.0
 *
 *   (...)
 *  \endcode
 */
string fwUkl1_convertByteToIp(const dyn_char& ipAddress, dyn_string& exceptionInfo) {
  //If the masks are in the data then they will be at the front i.e. lowest element numbers.
  //Determine the number of elements in the array and count backwards from there.
  //This will allow absolute indexing whether the masks are present or not.
  const int size = dynlen(ipAddress);
  //Check that it contains the appropriate number of elements.
  //Technically if it came from the configuration database it could contain the write mask tagged on the end.
  //This doesn't cause a problem as we use absolute indexes later, just require there to be 4 elements minimum (and the IP to be in those).
  if ( 4 > size ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToIp(): Input array contains " + size + " elements. Should be 4 elements big.", "1");
    return "Fail";
  }

  //Now we need to put the data into the appropriate format.
  string output;
  output += fwCcpc_convertByteToDec(ipAddress[size - 3]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size - 2]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size - 1]);
  output += ".";
  output += fwCcpc_convertByteToDec(ipAddress[size]);

  return output;
}

/*!
 * This takes the data that is read by from the L1 board FPGAs (LBUS), which is read as a series of 32-bit words,
 * and converts it to a series of 16-bit words can then be decoded to return the register value.
 *
 * \param  dcData The dyn_char array that contains the bytes read back from the hardware, will have the upper 2-bytes
 *           removed from the array. Assume byte ordering that matches that done read by the FW CCPC library. Return
 *           by reference.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_char The dyn_char that now has its bytes in an order that can be decoded by the fwCcpc library
 *           to return a string representing the hex number.
 *
 * The function performs two tasks. First it considers the byte array as a group of 32-bit numbers and removes
 * the upper two bytes to reduce them to a 16-bit number. The resulting array must then be word swapped, however
 * the words must maintain their internal byte ordering.
 */
void fwUkl1_convert32To16BitWords(dyn_char& dcData, dyn_string& exceptionInfo) {
  //Convert the read data from a series of 32-bit words to a series of 16-bit words.

  //Check that we have a multiple of 32-bits.
  if ( 0 != (dynlen(dcData)%4) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convert32To16BitWords(): Provided array does not contain enough data to make a 32-bit word, cannot correctly convert to 16-bit word. " + dynlen(dcData) + " elements present in given array.", "1");
    return;
  }

  //The loop must run over the number of 32-bit words in the read data.
  const unsigned total32Words = dynlen(dcData)/4;
  for (unsigned word = 0; word < total32Words; ++word) {
    //We need to remove the first two elements in that word.
    //This will follow the sequence of odd numbers.
    //After we delete the first element the second element becomes the first, so we always delete the same one.
    dynRemove(dcData, 2*word+1);
    dynRemove(dcData, 2*word+1);
  }
  //In the event that we have read back more than one word then we will need to swap the ordering of the 16-bit
  //words contained within the array.
  const unsigned total16Words = dynlen(dcData)/2;
  if (1 < total16Words) {
    //Copy the data into a temporary array.
    const dyn_char tmpData = dcData;
    //Loop over all the 16-bit words of the array and put them back into the original in the reverse order.
    const unsigned totalElements = dynlen(dcData);
    for (unsigned word = 0; word < totalElements; word+=2) {
      //Need to keep the byte ordering within the word, but word swap the array. 16-bit words.
      dcData[word+1] = tmpData[totalElements-word-1];
      dcData[word+2] = tmpData[totalElements-word];
    }
  }
  //Done.
  return;
}

/*!
 * Converts an array of bytes into a hex string and treats the dyn_char array as if it were little endian.
 *
 * \param  dcData The array of bytes to be converted.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return string Hex string of the dyn_char. Will return `Fail' if the data could not be converted.
 */
string fwUkl1_convertByteToHexLittleEndian(const dyn_char& dcData, dyn_string& exceptionInfo) {
  //We must reverse the byte ordering of the array.
  //For this we will need to know its size.
  const int numBytes = dynlen(dcData);
  if ( 0 < numBytes ) {
    //Have a none zero number of bytes.
    //This stores the reversed bytes.
    dyn_char swappedDcData;
    //Loop through them swapping the first element for last, 2nd for 2nd to last etc..
    for (int byte = 1; byte <= numBytes; ++byte) {
      swappedDcData[numBytes-(byte-1)] = dcData[byte];
    }
  }
  else if ( 0 == numBytes )
    return "";
  else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_convertByteToHexLittleEndian(): Failed to determine size of byte array, could not convert to hex string.", "1");
    return "Fail";
  }

  return fwCcpc_convertByteToHex(swappedDcData);
}


/*!
 * Takes a string and ensures that it contains enough characters for the number of bytes specified
 * by numBytes. Adds `0' to the beginnig if it is too short and removes characters from the beginning
 * if it is too long. Must be in hex.
 *
 * \param  sData The string that contains the data to be checked. Returned by reference.
 * \param  numBytes The number of bytes that the returned character string should return.
 * \return void
 *
 * This is for use with strings that represent a hex number that should be a set number of bits long.
 * It assumes that the lowest order bit is the last (reading left to right) character and hence its behaviour is
 * to move the last character always towards the final character in the padded string (the 2nd, 4th, 6th etc character).
 * This is achieved by padding or removing at the beginning of the string.
 *
 * \b Example:
 * \code
 *   (...)
 *   //Declare some data e.g. User input from a panel.
 *   const string hexData1 = "1234";
 *   //Wish to use the fwCcpc library to write this to a 32-bit registers hence is needs to be zero padded
 *   //to be 4-bytes big (8-characters).
 *   const string paddedData1 = fwUkl1_padString(hexData1, 4);
 *   DebugTN(paddedData1); // "00001234"
 *
 *   //Declare some more data e.g. User input from a panel.
 *   const string hexData2 = "1234";
 *   //Wish to use the fwCcpc library to write this to an 8-bit registers hence it need to be 1-byte (2-characters) big.
 *   const string paddedData2 = fwUkl1_padString(hexData2, 1);
 *   //Here the data is misformed and it is assumed that the least significant bit (last character) in the given number
 *   //should remain the least significant bit.
 *   DebugTN(paddedData2); // "34"
 */
void fwUkl1_padString(string& sData, unsigned numBytes) {
  //First determine the length.
  const int datalen = strlen(sData);
  //Create a temporary string to store the original string.
  const string tmpOrig = sData;
  //Determine the number of characters that must be in the final string,
  //each character represents half a byte or nibble.
  const int numChar = numBytes*2;
  //Perform the checking and padding/trimming.
  if ( numChar > datalen ) {
    //The data string is too short so increase its length by padding at the beginning, highest order bytes.
    //A temporary string to hold the amount of padding.
    string tmpPad;
    for (int character = 1; character <= (numChar-datalen); ++character)
      tmpPad += "0";
    //Now insert the padding into the data string.
    sData = tmpPad + tmpOrig;
  }
  else if ( numChar < datalen ) {
    //Remove the initial characters of the string until it reaches the appropriate length, highest order bytes.
    sData = substr(sData, (datalen-numChar), datalen);
  } else {
    //Do nothing, it is the right length.
  }

  //Return the resulting string.
  return;
}

/*!
 * This takes a array of string where each element of the array contains the relevant settings for either
 * the FE FPGAs, the channels on a FE FPGA or the GBE ports and converts these to an array of strings.
 * The returned result is an array of array of strings that contain all the relevant settings.
 *
 * \param  origData The dyn_string that contains the data to be parsed. Each element should contain a
 *           comma separated list of settings.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_dyn_string Each element contains a dyn_string, where each element contains the setting for
 *           an individual unit.
 *
 * The input strings must contain a comma separated list of the unit settings. The function checks that
 * there are either four or nine settings found, if there are none then the data format is considered to be
 * invalid and for that specific element an empty string is return. It will continue to parse the rest of
 * the data if an error is found. The ordering of the output data is the first element in the array corresponds
 * to the first entry in the string reading left to right.
 */
dyn_dyn_string fwUkl1_parseBroadcastData(const dyn_string& origData, dyn_string& exceptionInfo) {
  //Holds the result that we will be returning.
  dyn_dyn_string parsedData;

  //First determine the number strings we will need to parse.
  const unsigned totalLength = dynlen(origData);
  if ( -1 == totalLength ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBroadcastData(): Failed to determine number of elements in input array, unknown number of etries to parse, aborting.", "1");
    //Parsed data should be empty in this case.
    return parsedData;
  }

  //Loop over the array and deal with each element.
  for (unsigned element = 1; element <= totalLength; ++element) {
    //Holds the result of the string parsing.
    //Split the string up using comma's as a deliminator.
    parsedData[element] = strsplit(origData[element], ',');
    //Check we have enough elements for either all the FE FPGAs, GBE ports (both 4) or FE FPGA
    //channels (9).
    const unsigned foundParts = dynlen(parsedData[element]);
    if ( (1 != foundParts) && (4 != foundParts) && (9 != foundParts) ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBroadcastData(): Failed to parse the broadcast string `" + origData[element] + "'. Found " + foundParts + " sections in the string. Must have 1, 4 or 9. This element will contain the original string.", "2");
      //Parsed data wasn't good so return an empty array instead.
      parsedData[element] = makeDynChar(origData[element]);
    }
  }
  //Finally return the parsed strings.
  return parsedData;
}


// ===========
//    FPGAS
// ===========

/*!
 * This function takes the data for the various BE FPGA registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register. The appropriate name prefix is added and
 *           returned by reference. The following values are permitted:
 *   \li MacDestinationAddress A 12 character string that gives the hex number for the MAC address, can either be a continuous
 *         string of the address or may posses a colon delimiter every two characters,
 *         e.g. 000e0cb02d19 or 00:0e:0c:b0:2d:19.
 *   \li MacSourceAddress A 12 character string that gives the hex number for the MAC address, can either be a continuous
 *         string of the address or may posses a colon delimiter every two characters,
 *         e.g. 000e0cb02d19 or 00:0e:0c:b0:2d:19.
 *   \li IpDestinationAddress A string that contains either the address in hex format e.g. c0a0c0c0 or in a typical IP4 format
 *         e.g. 192.168.192.192.
 *   \li IpSourceAddress A string that contains either the address in hex format e.g. c0a0c0c0 or in a typical IP4 format
 *         e.g. 192.168.192.192.
 *   \li Protocol Default format described below.
 *   \li TypeOfService Default format described below.
 *   \li TimeToLive Default format described below.
 *   \li Type Default format described below.
 *   \li Version Default format described below.
 *   \li PartitionId Must contain an 8 character string representing the two 16-bit words of data for the partition ID.
 *   \li L1Id An 8-bit value, 2-character string. This will be present in the output data packet and MUST be unique for each L1.
 *         This is set correctly by the FSM and generally should not be User defined, except maybe when the User is an expert.
 *   \li ZeroSuppression String should be either `Enable' or `Disable' to enable or disable zero suppression across
 *         all the Ingress/FE FPGAs. Case sensitive.
 *   \li HpdPixelMode String should be either `ALICE' or `LHCb', which should correspond to the mode of operation
 *         required from the pixel chips. Case sensitive.
 *   \li Latency Default format described below.
 *   \li MepPackingFactor Default format described below.
 *   \li FixedMepDestinationAddress `Enable' or `Disable' to enable or disable a fixed multi-event packet destination address.
 *   \li DontWaitForMepDestBrdcst `Don't wait' or `Wait' to enable or disable the waiting for a multi-event packet destination
 *         address broadcast, respectively.
 *   \li Throttle `Enable' or `Disable' to enable or disable the throttle.
 *   \li ThrottlePolarity `Invert' or `Normal' to select the polarity of the throttle output.
 *   \li TfcDecoder `Enable' or `Disable' to enable or disable the decoding of TFC commands. The board will not operate without
 *         this setting enabled.
 *   \li PollGbePorts `Enable' or `Disable' to enable the Egress FPGA to send data to the GBE card for transmission.
 *   \li DataFragmentSize A decimal number specifying the size of the data packet that is transmitted via the GBE card.
 *         It is the size of the fragment of the multi-event packet that is transmitted. Max 9000.
 * \param sData Data to be written to the registers in the form of a string. The default format to be used is a hex string,
 *          lower or upper case that need not be zero padded to the appropriate register size.
 * \param ddcData Data converted into a dyn_dyn_char for use with the fwCcpc_write function. Returned by reference.
 * \param masks Masks that are associated with the write data also to be used with the fwCcpc_write function.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array. The returned ddcData is not quite suitable for writting to the database and the masks must first
 * be appeneded to the end of element in the ddcData, this must be done by the User.
 */
void fwUkl1_parseBeFpgaWriteRegisters(string ukl1Name, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".BeFpga.";

  //Add prefix to each name and convert the data into a byte array.
  //This indicates if there was a formatting error and we should remove those elements from the list or if they can be parsed.
  bool formatError;
  //Evaluate the size each time incase there is a problem with any of the data formating and an element is removed.
  //PVSS will update the size accordingly and this way we won't go out of bounds.
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //Generally we will succeed so just mark it good here and it can be marked back if necessary.
    //Masks should only be created if no error is found.
    formatError = FALSE;

    //Check for the MAC addresses being written.
    if ( ("MacDestinationAddress" == registerNames[element]) || ("MacSourceAddress" == registerNames[element]) ) {
      //This is potentially has colons in it that may need to be removed.
      sData[element] = fwUkl1_convertMacToHex(sData[element], exceptionInfo);
      if ( "Fail" != sData[element] ) {
	//Need a mask for a 96-bit word write, 3 32-bit words.
	masks[element] = makeDynChar(0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff);
	//Need to make sure the three 16-bit words that make up the data are padded to 32-bit words.
	//This relies on a deep knowledge of the way the fwCcpc library and L1 hardware works.
	//It is explained in the main library documentation.
	sData[element] = "0000" + substr(sData[element], 8, 4) + "0000" + substr(sData[element], 4, 4) +
	  "0000" + substr(sData[element], 0, 4);
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format, removed from list of writes.", "2");
      }
    }
    //Check for the IP addresses.
    else if ( ("IpDestinationAddress" == registerNames[element]) || ("IpSourceAddress" == registerNames[element]) ) {
      //This is potentially in decimal with full stop delimiters.
      sData[element] = fwUkl1_convertIpToHex(sData[element], exceptionInfo);
      if ( "Fail" != sData[element] ) {
	//Need a 64-bit mask for this.
	masks[element] = makeDynChar(0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff);
	//Need to make sure the two 16-bit words that make up the data are padded to 32-bit words.
	//This relies on a deep knowledge of the way the fwCcpc library and L1 hardware works.
	//It is explained in the main library documentation.
	sData[element] = "0000" + substr(sData[element], 4, 4) + "0000" + substr(sData[element], 0, 4);
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format, removed from list of writes.", "2");
      }
    }
    //Check for the partition ID.
    else if ( "PartitionId" == registerNames[element] ) {
      //Must be 8 characters big to be a valid ID, which is 4 bytes. This function will just make it fit if it is too big.
      fwUkl1_padString(sData[element], 4);
      //This needs to be written to two 32-bit registers with the two 16-bits words of the given partition ID
      //in each. Need to pad the beginning with four 0 characters and between the 1st and 2nd words.
      sData[element] = "0000" + substr(sData[element], 4, 4) + "0000" + substr(sData[element], 0, 4);
      //Here we need a 64-bit mask.
      masks[element] = makeDynChar(0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0xff, 0xff);
    }
    //Check for the zero suppression mode.
    else if ( "ZeroSuppression" == registerNames[element] ) {
      //Convert the requested setting into a padded hex string.
      if ( "Disable" == sData[element] ) {
	//For this we need to write to only the first four bits.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x0f);
	sData[element] = "00000000";
      } else if ( "Enable" == sData[element] ) {
	//For this we need to write to only the first four bits.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x0f);
	sData[element] = "0000000f";
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Disable' or `Enable', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check for a change of pixel mode.
    else if ( "HpdPixelMode" == registerNames[element] ) {
      //Convert the requested setting into a padded hex string.
      if ( "ALICE" == sData[element] ) {
	//For this we need to write to only bits 4 to 7.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0xf0);
	sData[element] = "000000f0";
      } else if ( "LHCb" == sData[element] ) {
	//For this we need to write to only bits 4 to 7.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0xf0);
	sData[element] = "00000000";
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `ALICE' or `LHCb', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check for fixing MEP destination address.
    else if ( "FixedMepDestinationAddress" == registerNames[element] ) {
      if ( "Disable" == sData[element] ) {
	//Write to bit 1.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x01);
	sData[element] = "00000000";
      } else if ( "Enable" == sData[element] ) {
	//Write to bit 1.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x01);
	sData[element] = "00000001";
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Disable' or `Enable', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check for MEP destination broadcast wait setting.
    else if ( "DontWaitForMepDestBrdcst" == registerNames[element] ) {
      if ( "Don't wait" == sData[element] ) {
	//Write to bit 2.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x02);
	sData[element] = "00000002";
      } else if ( "Wait" == sData[element] ) {
	//Write to bit 2.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x02);
	sData[element] = "00000000";
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Don't wait' or `Wait', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check for throttle enabling.
    else if ( "Throttle" == registerNames[element] ) {
      if ( "Disable" == sData[element] ) {
	//Write to bits 3 and 4.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x0c);
	sData[element] = "00000000";
      } else if ( "Enable" == sData[element] ) {
	//Write to bits 3 and 4.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x0c);
	sData[element] = "0000000c";
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Disable' or `Enable', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check for throttle invertion.
    else if ( "ThrottlePolarity" == registerNames[element] ) {
      if ( "Invert" == sData[element] ) {
	//Write to bits 5 and 6.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x30);
	sData[element] = "00000030";
      } else if ( "Normal" == sData[element] ) {
	//Write to bits 5 and 6.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x30);
	sData[element] = "00000000";
      } else {
 	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Invert' or `Normal', got " + sData[element] + ". Removed from list of writes.", "2");
     }
    }
    //Check if TFC decoder is enabled.
    else if ( "TfcDecoder" == registerNames[element] ) {
      if ( "Disable" == sData[element] ) {
	//Write to bit 6.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x40);
	sData[element] = "00000000";
      } else if ( "Enable" == sData[element] ) {
	//Write to bit 6.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x40);
	sData[element] = "00000040";
      } else {
 	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Invert' or `Normal', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Check if transmission to the GBE output ports is enabled.
    else if ( "PollGbePorts" == registerNames[element] ) {
      if ( "Disable" == sData[element] ) {
	//Write to bit 7.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x80);
	sData[element] = "00000000";
      } else if ( "Enable" == sData[element] ) {
	//Write to bit 7.
	masks[element] = makeDynChar(0x00, 0x00, 0x00, 0x80);
	sData[element] = "00000080";
      } else {
 	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Disable' or `Enable', got " + sData[element] + ". Removed from list of writes.", "2");
      }
    }
    //Data fragment size is in decimal.
    else if ( "DataFragmentSize" == registerNames[element] ) {
      //Convert from decimal to hex.
      sData[element] = fwCcpc_convertDecToHex(sData[element]);
      //Pad the string.
      fwUkl1_padString(sData[element], 4);
      //Create a 16-bit mask.
      masks[element]   = makeDynChar(0x00, 0x00, 0xff, 0xff);
    }
    //L1 ID is only 2-characters big.
    else if ( "L1Id" == registerNames[element] ) {
      //Create an 8-bit mask.
      masks[element]   = makeDynChar(0x00, 0x00, 0x00, 0xff);
      //Pad the string.
      fwUkl1_padString(sData[element], 4);
    }
    //All the others just require a 16-bit mask and padding to 32-bit word.
    else {
      //Need a 16-bit mask.
      masks[element]   = makeDynChar(0x00, 0x00, 0xff, 0xff);
      //Pad the string, it still must be 32-bits long, the mask ensures we ignore the upper 16-bits.
      fwUkl1_padString(sData[element], 4);
    }

    if ( !formatError ) {
      //Put the data and register names in a format that the write can use.
      //Add the prefix to convert it to a datapoint name.
      registerNames[element] = prefix + registerNames[element];
      //Convert from a hex string to a byte array.
      ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
    } else {
      //Failed to parse the register, exception has already been raised...
      //Remove from register list.
      dynRemove(registerNames, element);
      //Remove the original data so it everything is kept in the same order as the registerNames
      dynRemove(sData, element);
      //Masks and ddcData aren't written so they are not affected.
      //Decrement the element number as we need to start this loop at the same number as before.
      --element;
    }

  }//for element

  //and we are done.
  return;

}

/*!
 * Takes the data read from the BE FPGA configuration registers using fwCcpc_read or from a recipe and converts them to
 * strings to provide a description of the register setting. It also splits settings that occupy the same register address
 * into their own description.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the names and the BE FPGA registers to be read from as named in the UKL1 hardware
 *           datapoint. They can either be the full name or just the register name. The following registers can be parsed
 *           (their names are given in the short form, just the DP element) and also a description of the return value given by sData:
 *   \li MacDestinationAddress A 12 character string that gives the hex number for the MAC address, can either be a continuous
 *         string of the address or may posses a colon delimiter every two characters,
 *         e.g. 000e0cb02d19 or 00:0e:0c:b0:2d:19.
 *   \li MacSourceAddress A 12 character string that gives the hex number for the MAC address, can either be a continuous
 *         string of the address or may posses a colon delimiter every two characters,
 *         e.g. 000e0cb02d19 or 00:0e:0c:b0:2d:19.
 *   \li IpDestinationAddress A string that contains either the address in hex format e.g. c0a0c0c0 or in a typical IP4 format
 *         e.g. 192.168.192.192.
 *   \li IpSourceAddress A string that contains either the address in hex format e.g. c0a0c0c0 or in a typical IP4 format
 *         e.g. 192.168.192.192.
 *   \li Protocol Default format described below.
 *   \li TypeOfService Default format described below.
 *   \li TimeToLive Default format described below.
 *   \li Type Default format described below.
 *   \li Version Default format described below.
 *   \li PartitionId Must contain an 8 character string representing the two 16-bit words of data for the partition ID.
 *   \li L1Id An 8-bit value, 2-character string. This will be present in the output data packet and MUST be unique for each L1.
 *         This is set correctly by the FSM and generally should not be User defined, except maybe when the User is an expert.
 *   \li ZeroSuppression String should be either `Enable' or `Disable' to enable or disable zero suppression across
 *         all the Ingress/FE FPGAs. Case sensitive.
 *   \li HpdPixelMode String should be either `ALICE' or `LHCb', which should correspond to the mode of operation
 *         required from the pixel chips. Case sensitive.
 *   \li Latency Default format described below.
 *   \li MepPackingFactor Default format described below.
 *   \li FixedMepDestinationAddress `Enable' or `Disable' to enable or disable a fixed multi-event packet destination address.
 *   \li DontWaitForMepDestBrdcst `Don't wait' or `Wait' to enable or disable the waiting for a multi-event packet destination
 *         address broadcast, respectively.
 *   \li Throttle `Enable' or `Disable' to enable or disable the throttle.
 *   \li ThrottlePolarity `Invert' or `Normal' to select the polarity of the throttle output.
 *   \li TfcDecoder `Enable' or `Disable' to enable or disable the decoding of TFC commands. The board will not operate without
 *         this setting enabled.
 *   \li PollGbePorts `Enable' or `Disable' to enable the Egress FPGA to send data to the GBE card for transmission.
 *   \li DataFragmentSize A decimal number specifying the size of the data packet that is transmitted via the GBE card.
 *         It is the size of the fragment of the multi-event packet that is transmitted. Max 9000.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring, that the fwCcpc_read function is not used and in this case an empty
 *           array should be passed and it will be ignored. Any errors found in this array will be reported via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * It does not matter if data is being parsed from a recipe or from the hardware the results are the same. The recipe data must first
 * be converted from a string to a dyn_char (can be achieved through fwCcpc_convertHexToByte()).
 */
void fwUkl1_parseBeFpgaReadRegisters(string ukl1Name, const dyn_string& registerNames, dyn_dyn_char ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if the full DPE name is given we match to it.
  const string prefix = ukl1Name + ".BeFpga.";
  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Convert the data from a series of 32-bit words to a series of 16-bit words.
    const int exceptSize = dynlen(exceptionInfo);
    fwUkl1_convert32To16BitWords(ddcData[element], exceptionInfo);

    //Need to know the size of the read data to allow absolute indexing of the given data.
    //It can be used to account for whether the mask data is in the lower elements if it comes from a recipe.
    const int size = dynlen(ddcData[element]);

    //Check that the data went through the first stage of parsing correctly.
    if ( (exceptSize != dynlen(exceptionInfo)) || (exceptSize == -1) ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): Failed to convert the read 32-bit data to the 16-bit data stored in the FE FPGA registers.", "2");
      data[element] = "Fail";
    }
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    else if ( checkStatus && (0 != callStatusList[element]) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_ParseBeFpgaReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
      } else {
	errMsg += "an unknown error.";
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error.
      data[element] = "Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    //Parse the MAC addresses.
    else if ( ("MacDestinationAddress" == registerNames[element]) || ("MacSourceAddress" == registerNames[element]) ||
	      (prefix + "MacDestinationAddress" == registerNames[element]) || (prefix + "MacSourceAddress" == registerNames[element]) ) {
      //Now convert the remaining 48-bits into a MAC address.
      //Don't worry about errors the exceptionInfo should be appropriately filled if there was an error plus the string reads `Fail'.
      data[element] = fwUkl1_convertByteToMac(ddcData[element], exceptionInfo);
    }
    //Parse the IP addresses.
    else if ( ("IpDestinationAddress" == registerNames[element]) || ("IpSourceAddress" == registerNames[element]) ||
	      (prefix + "IpDestinationAddress" == registerNames[element]) || (prefix + "IpSourceAddress" == registerNames[element]) ) {
      //Convert the given data into a full stop delimited IP.
      //Don't worry about exceptions the convertion function will ensure `Fail' is returned in an error and describe what went wrong.
      data[element] = fwUkl1_convertByteToIp(ddcData[element], exceptionInfo);
    }
    //Parse the zero suppression data.
    else if ( ("ZeroSuppression" == registerNames[element]) || (prefix + "ZeroSuppression" == registerNames[element]) ) {
      //Convert the requested setting into a padded hex string.
      if ( 0x00 == (ddcData[element][size] & 0x0f) ) {
	data[element] = "Disable";
      } else if ( 0x0f == (ddcData[element][size] & 0x0f) ) {
	data[element] = "Enable";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Parse LHCb mode settings.
    else if ( ("HpdPixelMode" == registerNames[element]) || (prefix + "HpdPixelMode" == registerNames[element]) ) {
      //Convert the requested setting into a padded hex string.
      if ( 0xf0 == (ddcData[element][size] & 0xf0) ) {
	data[element] = "ALICE";
      } else if ( 0x00 == (ddcData[element][size] & 0xf0) ) {
	data[element] = "LHCb";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Parse enable forced fixed MEP destination address settings.
    else if ( ("FixedMepDestinationAddress" == registerNames[element]) || (prefix + "FixedMepDestinationAddress" == registerNames[element]) ) {
      //Is it enabled or disabled.
      if ( 0x01 == (ddcData[element][size] & 0x01) ) {
	data[element] = "Enable";
      } else if ( 0x00 == (ddcData[element][size] & 0x01) ) {
	data[element] = "Disable";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Parse don't wait for MEP destination broadcast settings.
    else if ( ("DontWaitForMepDestBrdcst" == registerNames[element]) || (prefix + "DontWaitForMepDestBrdcst" == registerNames[element]) ) {
      //Is it waiting for not.
      if ( 0x02 == (ddcData[element][size] & 0x02) ) {
	data[element] = "Don't wait";
      } else if ( 0x00 == (ddcData[element][size] & 0x02) ) {
	data[element] = "Wait";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Check to see if the throttle is enabled.
    else if ( ("Throttle" == registerNames[element]) || (prefix + "Throttle" == registerNames[element]) ) {
      //Is it enabled or disabled.
      if ( 0x0c == (ddcData[element][size] & 0x0c) ) {
	data[element] = "Enable";
      } else if ( 0x00 == (ddcData[element][size] & 0x0c) ) {
	data[element] = "Disable";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Check polarity of throttle.
    else if ( ("ThrottlePolarity" == registerNames[element]) || (prefix + "ThrottlePolarity" == registerNames[element]) ) {
      //Is it enabled or disabled.
      if ( 0x30 == (ddcData[element][size] & 0x30) ) {
	data[element] = "Invert";
      } else if ( 0x00 == (ddcData[element][size] & 0x30) ) {
	data[element] = "Normal";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Check to see if the TFC is enabled.
    else if ( (prefix + "TfcDecoder" == registerNames[element]) ) {
      //Is it enabled or disabled.
      if ( 0x40 == (ddcData[element][size] & 0x40) ) {
	data[element] = "Enable";
      } else if ( 0x00 == (ddcData[element][size] & 0x40) ) {
	data[element] = "Disable";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Check to see if the transmission of data to the GBE is enabled.
    else if ( ("PollGbePorts" == registerNames[element]) || (prefix + "PollGbePorts" == registerNames[element]) ) {
      //Is it enabled or disabled.
      if ( 0x80 == (ddcData[element][size] & 0x80) ) {
	data[element] = "Enable";
      } else if ( 0x00 == (ddcData[element][size] & 0x80) ) {
	data[element] = "Disable";
      } else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	data[element] = "Fail";
      }
    }
    //Check the size of the data fragment size, it needs to be in decimal.
    else if ( ("DataFragmentSize" == registerNames[element]) || (prefix + "DataFragmentSize" == registerNames[element]) ) {
      if ( size == 4 ) {
	//Value is twice as big as we expect, hence the mask must be present. Zero the first two elements to cure this problem.
	ddcData[element][1] = 0;
	ddcData[element][2] = 0;
      }//No else as it is just right in any other case.
      data[element] = fwCcpc_convertByteToDec(ddcData[element]);
    }
    //This is only 8-bits big.
    else if ( ("L1Id" == registerNames[element]) || (prefix + "L1Id" == registerNames[element]) ) {
      //Zero the elements until only the last on is left as this is the actual L1 ID.
      const unsigned maxCount = dynlen(ddcData[element]);
      //Notice that the loop deliberately goes to less than element count. This is to ensures that the whole array is not zerod.
      for (unsigned count = 1; count < maxCount; ++count) {
	ddcData[element][count] = 0;
      }
      //Also remove the leading zeros.
      data[element] = strltrim(fwCcpc_convertByteToHex(ddcData[element]), "0");
      //Check we shouldn't be returning zero!
      if ("" == data[element])
	data[element] = "0";
    }
    //Can just return the read data to a hex string as the extra words have already been removed.
    else {
      //Need to deal with the mask being present here......
      if ( (size == 4) && ("PartitionId" != registerNames[element]) && (prefix + "PartitionId" != registerNames[element]) ) {
	//In this case we must have a 16-bit register and the mask from the recipe.
	//Only the partition ID is 32-bits big and the register name is not that.
	//Zero the first two elements to get ride of the mask.
	ddcData[element][1] = 0;
	ddcData[element][2] = 0;
      }
      if ( size == 8 ) {
	//In this case it is the partition ID and we can delete the first 4 elements.
	ddcData[element][1] = 0;
	ddcData[element][2] = 0;
	ddcData[element][3] = 0;
	ddcData[element][4] = 0;
      }
      //The other case is 2 elements, which is just a 16-bit register or 4 (32-bits) and it is the partition ID.
      //In both of these cases the data can be left alone.
      data[element] = strltrim(fwCcpc_convertByteToHex(ddcData[element]), "0");
      //Check we shouldn't be returning zero!
      if ("" == data[element])
	data[element] = "0";
    }
  }//for size of arrays.

  //And we are done.
  return;
}


/*!
 * This function takes the data for the various FE FPGA registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3, FWUKL1_BROADCAST_FPGA.
 * \param  registerName List of FE FPGA register names that are to be written to. Returned with the appropriate name
 *           prefix by reference. The allowed values are described below.
 *   \li Disable Accepts the text `Disable' or `Enable' to disable or enable a FE FPGA.
 *   \li Emulator Accepts the text `Disable' or `Enable' to disable or enable input data emulation on that FE FPGA, respectively.
 *         This setting disables the input data from the optical inputs on the 9 associated inputs channels and then sends emulated
 *         data through the L1 data processing chain.
 *   \li EmulatorHpdPixelMode Accepts `LHCb' or `ALICE'. The type of emulated data that will be sent. This will not automatically
 *         enable the event emulation.
 *   \li EmulatorEvent Accepts `Send' or `Don't send' to enable the sending of emulated events. This can only occur if the emulator is
 *         enabled via the `Emulator' setting. `Don't send' will prevent events being sent through the chain even if the emulator is
 *         enabled.
 *   \li ResetStatusBufferPointer Accepts `Assert' or `Deassert', which will either assert or deassert the reset for the buffer pointer
 *         for that FE FPGA. This must be done before reading the status information for a FE FPGA or it will produce undefined data.
 * \param  sData Data to be written to the registers in the form of a string. The default case is a 4-character hex string,
 *           comma separated in the case of broadcast data.
 * \param  ddcData Data converted into a dyn_dyn_char for use with the fwCcpc_write function. Returned by reference.
 * \param  masks Masks that are associated with the write data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array.
 */
void fwUkl1_parseFeFpgaWriteRegisters(string ukl1Name, unsigned feFpga, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  //Either we need to write to a specific FE FPGA or in the case of broadcasting we can  write to FE FPGA 0 as all FE FPGA settings
  //exist in the same BE FPGA register and just set different bits. We can always write to the FE FPGA 0 DP and with the
  //appropriate mask and data structure it can write them all at the same time.
  const unsigned selectedFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
  //Determines which FPGA we should loop to when parsing the data.
  const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
  const string prefix = ukl1Name + ".FeFpga" + selectedFpga + ".";

  //First we need to parse the broadcast data to break it up into its component FPGA settings.
  dyn_dyn_string sParsedData = fwUkl1_parseBroadcastData(sData, exceptionInfo);
  //Don't bother dealing with the errors here it will be picked up later when looking at the
  //data and the User will see the extra exception details.

  //Add prefix to each name and convert the data into a byte array.
  //This indicates if there was a formatting error and we should remove those elements from the list or if they can be parsed.
  bool formatError;
  //Evaluate the size each time incase there is a problem with any of the data formating and an element is removed.
  //PVSS will update the size accordingly and this way we won't go out of bounds.
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //This way we only need to indicate if something went wrong.
    formatError = FALSE;
    //Disabling the FPGA.
    if ( "Disable" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned disable = 0;
      //Now loop over the appropriate range of FPGAs.
      //Loop should start at the FE FPGA we selected in name prefix.
      for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 1<<loopFpga;
	//Create the data to write. In the event that we are only writing to 1 FPGA then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "Disable" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Set the relevant bit high.
	  disable |= 1<<loopFpga;
	} else if ( "Enable" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected Disable or Enable, got " + sParsedData[element][loopFpga+1-selectedFpga] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(disable);
	fwUkl1_padString(sData[element], 4);
      }
    }
    //Enable/disable the L0 emulator setting.
    else if ( "Emulator" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned disable = 0;
      //Now loop over the appropriate range of FPGAs.
      for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 1<<(4*loopFpga);
	//Create the data to write. In the event that we are only writing to 1 FPGA then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "Disable" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Set the relevant bit high.
	  disable |= 1<<(4*loopFpga);
	} else if ( "Enable" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Disable' or `Enable', got " + sParsedData[element][loopFpga+1-selectedFpga] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(disable);
	fwUkl1_padString(sData[element], 4);
      }
    }
    //Set the pixel mode for the emulated event.
    else if ( "EmulatorHpdPixelMode" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned mode = 0;
      //Now loop over the appropriate range of FPGAs.
      //Loop should start at the FE FPGA we selected in name prefix.
      for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 2<<(4*loopFpga);
	//Create the data to write. In the event that we are only writing to 1 FPGA then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "ALICE" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Set the relevant bit high.
	  mode |= 2<<(4*loopFpga);
	} else if ( "LHCb" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `ALICE' or `LHCb', got " + sParsedData[element][loopFpga+1-selectedFpga] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(mode);
	fwUkl1_padString(sData[element], 4);
      }
    }
    //Enable/disable the L0 emulator setting.
    else if ( "EmulatorEvent" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned send = 0;
      //Now loop over the appropriate range of FPGAs.
      //Loop should start at the FE FPGA we selected in name prefix.
      for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 4<<(4*loopFpga);
	//Create the data to write. In the event that we are only writing to 1 FPGA then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "Send" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Set the relevant bit high.
	  send |= 4<<(4*loopFpga);
	} else if ( "Don't send" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Don't send' or `Send', got " + sParsedData[element][loopFpga+1-selectedFpga] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(send);
	fwUkl1_padString(sData[element], 4);
      }
    }
    //Asserts or deasserts the reset for the FE status buffer pointer.
    else if ( "ResetStatusBufferPointer" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned assert = 0;
      //Now loop over the appropriate range of FPGAs.
      //Loop should start at the FE FPGA we selected in name prefix.
      for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 1<<loopFpga;
	//Create the data to write. In the event that we are only writing to 1 FPGA then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "Assert" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Set the relevant bit high.
	  assert |= 1<<loopFpga;
	} else if ( "Deassert" == sParsedData[element][loopFpga+1-selectedFpga] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. Expected `Deassert' or `Assert', got " + sParsedData[element][loopFpga+1-selectedFpga] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(assert);
	fwUkl1_padString(sData[element], 4);
      }
    }
    else {
      //Check that the data was parsed actually contains something.
      if ( 0 != dynlen(sParsedData[element]) ) {
	//These are useful for holding the mask and data to be written as it is determined.
	unsigned mask = 0;
	unsigned number = 0;
	//Now loop over the appropriate range of FPGAs.
	//Loop should start at the FE FPGA we selected in name prefix.
	for (unsigned loopFpga = selectedFpga; loopFpga <= endFpga; ++loopFpga) {
	  //Determine the number of characters in the string we wish to put in the register.
	  const unsigned numChar = strlen(sParsedData[element][loopFpga+1-selectedFpga]);
	  //Create the mask to cover the found number of characters.
	  //Store it as a single number and convert it to a hex string at the end.
	  //Create word of 1's then move enough 0's in to cover the bits we wish to write to.
	  //Invert that to get the 1's at the beginning and then shift the those 1's by an amount determined by the
	  //FPGA number. It assumes the data to be written is the same size for each FPGA and is padded to the appropriate length.
	  mask |= ~(~0<<numChar)<<(loopFpga*numChar);
	  //This is a bit of a funny way to do things, but there are no generic registers for FE specific commands,
	  //so this might make sense one day. At least it will do something sensible one day...
	  //Convert the four numbers got from the string to decimal and perform a bitwise OR. once they have been aligned
	  //to the appropriate byte boundary.
	  number |= fwCcpc_convertHexToDec(sParsedData[element][loopFpga+1-selectedFpga])<<(loopFpga*numChar);
	}
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(number);
	fwUkl1_padString(sData[element], 4);
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaWriteRegisters(): Failed to parse " + registerNames[element] + " due to an incorrect data format. No data present in array. Removed from list of writes.", "2");
      }
    }

    if ( !formatError ) {
      //Put the data and register names in a format that the write can use.
      //Add the prefix to convert it to a datapoint name.
      registerNames[element] = prefix + registerNames[element];
      //Convert from a hex string to a byte array.
      ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
    } else {
      //Failed to parse the register, exception has already been raised...
      //Remove from register list.
      dynRemove(registerNames, element);
      //Remove the original data so it everything is kept in the same order as the registerNames
      dynRemove(sData, element);
      //Masks and ddcData aren't written so they are not affected.
      //Decrement the element number as we need to start this loop at the same number as before.
      --element;
    }

  }//for element

  return;
}

/*!
 * Takes the data read from the FE FPGA configuration registers using fwCcpc_read or a recipe and converts them to string to
 * provide a description of the register setting. It also splits settings that occupy the same register address into their
 * own description.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3, FWUKL1_BROADCAST_FPGA.
 * \param  registerNames dyn_string that contains the names and the FE FPGA registers to be read from as named in the UKL1 hardware
 *           datapoint. They can either be the full name or just the register name. The following registers can be parsed
 *           (their names are given in the short form, just the DP element) and also a description of the return value given by sData:
 *           It will be zero padded to the nearest 16-bit word, except in the following cases:
 *   \li Disable `Enable' or `Disable' depending on the state of the FPGA.
 *   \li Emulator `Disable' or `Enable' depending on whether the L0 emulation is active or not.
 *   \li EmulatorHpdPixelMode `LHCb' or `ALICE' depending on which HPD pixel mode is being simulated in the events.
 *   \li EmulatorEvent `Send' or `Don't send' depending on whether or not events are being sent when in L0 emulation mode.
 *   \li ResetStatusBufferPointer `Assert' or `Deassert' depending on whether the reset is asserted or not.
 *           In broadcast mode the returned data elements will contain a comma delimited list of the settings for each
 *           individual FE FPGA.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring, that the fwCcpc_read function is not used and in this case an empty
 *           array should be passed and it will be ignored. Any errors found in this array will be reported via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseFeFpgaReadRegisters(string ukl1Name, unsigned feFpga, const dyn_string& registerNames, dyn_dyn_char ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  //Either we need to write to a specific FE FPGA or in the case of broadcasting we can  write to FE FPGA 0 as all FE FPGA settings
  //exist in the same BE FPGA register and just set different bits. We can always write to the FE FPGA 0 DP and with the
  //appropriate mask and data structure it can write them all at the same time.
  const unsigned selectedFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
  //Determines which FPGA we should loop to when parsing the data.
  const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
  const string prefix = ukl1Name + ".FeFpga" + selectedFpga + ".";

  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Convert the data from a series of 32-bit words to a series of 16-bit words.
    const int exceptSize = dynlen(exceptionInfo);
    fwUkl1_convert32To16BitWords(ddcData[element], exceptionInfo);

    //Check that the data went through the first stage of parsing correctly.
    if ( (exceptSize != dynlen(exceptionInfo)) || (exceptSize == -1) ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): Failed to convert the read 32-bit data to the 16-bit data stored in the FE FPGA registers.", "2");
      //Mark the error, either one or all FE FPGAs.
      if ( FWUKL1_BROADCAST_FPGA != feFpga )
	data[element] = "Fail";
      else
	data[element] = "Fail,Fail,Fail,Fail";
    }
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    else if ( checkStatus && (0 != callStatusList[element]) && (0 != dynlen(ddcData[element])) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
      } else {
	errMsg += "an unknown error.";
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error, either one or all FE FPGAs.
      if ( FWUKL1_BROADCAST_FPGA != feFpga )
	data[element] = "Fail";
      else
	data[element] = "Fail,Fail,Fail,Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    else if ( ("Disable" == registerNames[element]) || ((prefix + "Disable") == registerNames[element]) ) {
      //Is it enabled or disabled?
      //First convert it to a single number.
      unsigned fpgaStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x01 == ((fpgaStates>>loopFpga) & 0x01) ) {
	  data[element] += "Disable";
	} else if ( 0x00 == ((fpgaStates>>loopFpga) & 0x01) ) {
	  data[element] += "Enable";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    else if ( ("Emulator" == registerNames[element]) || ((prefix + "Emulator") == registerNames[element]) ) {
      //Is it enabled or disabled?
      //First convert it to a single number.
      unsigned fpgaStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x01 == (fpgaStates>>(4*loopFpga) & 0x01) ) {
	  data[element] += "Disable";
	} else if ( 0x00 == (fpgaStates>>(4*loopFpga) & 0x01) ) {
	  data[element] += "Enable";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    else if ( ("EmulatorHpdPixelMode" == registerNames[element]) || (prefix + "EmulatorHpdPixelMode" == registerNames[element]) ) {
      //Is it in ALICE or LHCb mode?
      //First convert it to a single number.
      unsigned fpgaStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x02 == (fpgaStates>>(4*loopFpga) & 0x02) ) {
	  data[element] += "ALICE";
	} else if ( 0x00 == (fpgaStates>>(4*loopFpga) & 0x02) ) {
	  data[element] += "LHCb";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    else if ( ("EmulatorEvent" == registerNames[element]) || (prefix + "EmulatorEvent" == registerNames[element]) ) {
      //Is it send events or not?
      //First convert it to a single number.
      unsigned fpgaStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x04 == (fpgaStates>>(4*loopFpga) & 0x04) ) {
	  data[element] += "Send";
	} else if ( 0x00 == (fpgaStates>>(4*loopFpga) & 0x04) ) {
	  data[element] += "Don't send";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    //Is the FE status buffer pointer reset asserted or deasserted.
    else if ( ("ResetStatusBufferPointer" == registerNames[element]) || (prefix + "ResetStatusBufferPointer" == registerNames[element]) ) {
      //First convert it to a single number.
      unsigned fpgaStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x01 == ((fpgaStates>>loopFpga) & 0x01) ) {
	  data[element] += "Assert";
	} else if ( 0x00 == ((fpgaStates>>loopFpga) & 0x01) ) {
	  sData[element] += "Deassert";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    //Can just return the read data to a hex string as the extra words have already been removed.
    else {
      //Assume the same data should be repeated four times if broadcasting.
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of FPGAs.
      //First determine the starting number.
      const unsigned startFpga = FWUKL1_BROADCAST_FPGA==feFpga ? 0:feFpga;
      //Determine the end of the loop.
      const unsigned endFpga = FWUKL1_BROADCAST_FPGA==feFpga ? (FWUKL1_FRONT_FPGAS-1):feFpga;
      for (unsigned loopFpga = startFpga; loopFpga <= endFpga; ++loopFpga) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	data[element] = fwCcpc_convertByteToHex(ddcData[element]);
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
  }//for element

  //and we are done.
  return;
}

/*!
 * This function takes the data for the various FE FPGA channel registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  channel Input channel number on a FE FPGA that is to be written to. Range: 0 to 8, FWUKL1_BROADCAST_CHANNEL.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3.
 * \param  registerName List of FE FPGA channel register names that are to be written to. Returned with the appropriate name
 *           prefix by reference. The allowed values are:
 *   \li Disable Accepts the text `Disable' or `Enable' to disable or enable a FE FPGA.
 * \param  sData Data to be written to the registers in the form of a hex string. The default value is a 4-character
 *           hex string, other values are accepted for specific registers as described above.
 * \param  ddcData Data converted into a dyn_dyn_char for use with the fwCcpc_write function. Returned by reference.
 * \param  masks Masks that are associated with the write data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array.
 */
void fwUkl1_parseFeFpgaChannelWriteRegisters(string ukl1Name, unsigned channel, unsigned feFpga, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  //Either we need to write to a specific FE FPGA channel or in the case of broadcasting we can write to FE FPGA channel 0 as all
  //FE FPGA channel settings are at the same register address, we merely need to write to the first channel with the appropriate data.
  //This is why broadcasting over the UKL1 channels is not any better, as you still have to write to the four registers anyway.
  const unsigned selectedChannel = FWUKL1_BROADCAST_CHANNEL==channel ? 0:channel;
  //Determine the end of the loop, which parsing the data.
  const unsigned endChannel = FWUKL1_BROADCAST_CHANNEL==channel ? (FWUKL1_FPGA_CHANNELS-1):channel;
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".FeFpga" + feFpga + ".Channel" + selectedChannel + ".";
  //First we need to parse the broadcast data to break it up into its component channel settings.
  dyn_dyn_string sParsedData = fwUkl1_parseBroadcastData(sData, exceptionInfo);

  //Don't bother dealing with the errors here it will be picked up later when looking at the
  //data and the User will see the extra exception details.

  //Add prefix to each name and convert the data into a byte array.
  //This indicates if there was a formatting error and we should remove those elements from the list or if they can be parsed.
  bool formatError;
  //Evaluate the size each time incase there is a problem with any of the data formating and an element is removed.
  //PVSS will update the size accordingly and this way we won't go out of bounds.
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //This way we only need to indicate if something went wrong.
    formatError = FALSE;
    if ( "Disable" == registerNames[element] ) {
      //These are useful for holding the mask and data to be written as it is determined.
      unsigned mask = 0;
      unsigned disable = 0;
      //Now loop over the appropriate range of FPGAs.
      //Loop should start at the FE FPGA we selected in name prefix.
      for (unsigned loopChannel = selectedChannel; loopChannel <= endChannel; ++loopChannel) {
	//Create the mask. Store it as a single number and convert it to a hex string at the end.
	mask |= 1<<loopChannel;
	//Create the data to write. In the event that we are only writing to 1 channel then we must always take
	//the data from the 1st element, as this is the only element that will contain any data.
	if ( "Disable" == sParsedData[element][loopChannel+1-selectedChannel] ) {
	  //Set the relevant bit high.
	  disable |= 1<<loopChannel;
	} else if ( "Enable" == sParsedData[element][loopChannel+1-selectedChannel] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaChannelWriteRegister(): Failed to parse " + registerNames[element] + ". Expected `Disable' or `Enable', got " + sParsedData[element][loopChannel+1-selectedChannel] + ". Removed from list of writes.", "2");
	}
      }
      if ( !formatError ) {
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(disable);
	fwUkl1_padString(sData[element], 4);
      }
    }
    else {
      //This doesn't really have any meaning as there are no channel settings that take numbers
      //if there do exist such registers one day then who ever is reading this just might as well
      //rewrite this section as it is unlikely to make sense for what is required it is just a guess...
      //Check that the data was parsed actually contains something.
      if ( 0 != dynlen(sParsedData[element]) ) {
	//These are useful for holding the mask and data to be written as it is determined.
	unsigned mask = 0;
	unsigned number = 0;
	//Now loop over the appropriate range of FPGAs.
	//Loop should start at the FE FPGA we selected in name prefix.
	for (unsigned loopChannel = selectedChannel; loopChannel <= endChannel; ++loopChannel) {
	  //Determine the number of characters in the string we wish to put in the register.
	  const unsigned numChar = strlen(sParsedData[element][loopChannel+1-selectedChannel]);
	  //Create the mask to cover the found number of characters.
	  //Store it as a single number and convert it to a hex string at the end.
	  //Create word of 1's then move enough 0's in to cover the bits we wish to write to.
	  //Invert that to get the 1's at the beginning and then shift the those 1's by an amount determined by the
	  //FPGA number. It assumes the data to be written is the same size for each FPGA and is padded to the appropriate length.
	  mask |= ~(~0<<numChar)<<(loopChannel*numChar);
	  //This is a bit of a funny way to do things, but there are no generic registers for FE channel specific commands,
	  //so this might make sense one day. At least it will do something sensible...
	  //Convert the numbers got from the string to decimal and perform a bitwise OR. once they have been aligned
	  //to the appropriate byte boundary.
	  number |= fwCcpc_convertHexToDec(sParsedData[element][loopFpga+1-selectedFpga])<<(loopChannel*numChar);
	}
	//Convert the mask data to a string.
	//Convert the mask data to a string.
	string sTmpMask = fwCcpc_convertDecToHex(mask);
	fwUkl1_padString(sTmpMask,4);
	masks[element] = fwCcpc_convertHexToByte(sTmpMask);
	//Convert the register data to a string.
	sData[element] = fwCcpc_convertDecToHex(number);
	fwUkl1_padString(sData[element], 4);
      } else {
	formatError = TRUE;
	//Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaChannelWriteRegister(): Failed to parse " + registerNames[element] + " due to an incorrect data format. No data present in array. Removed from list of writes.", "2");
      }
    }

    if ( !formatError ) {
      //Put the data and register names in a format that the write can use.
      //Add the prefix to convert it to a datapoint name.
      registerNames[element] = prefix + registerNames[element];
      //Convert from a hex string to a byte array.
      ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
    } else {
      //Failed to parse the register, exception has already been raised...
      //Remove from register list.
      dynRemove(registerNames, element);
      //Remove the original data so it everything is kept in the same order as the registerNames
      dynRemove(sData, element);
      //Masks and ddcData aren't written so they are not affected.
      //Decrement the element number as we need to start this loop at the same number as before.
      --element;
    }

  }//for element

  return;
}

/*!
 * Takes the data read from the FE FPGA channel configuration registers using fwCcpc_read or from a recipe and converts
 * them to string to provide a description of the register setting. It also splits settings that occupy the same register
 * address into their own description.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  channel Input channel number on a FE FPGA that is to be written to. Range: 0 to 8, FWUKL1_BROADCAST_CHANNEL.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3.
 * \param  registerNames dyn_string that contains the names and the FE FPGA registers to be read from as named in the UKL1 hardware
 *           datapoint. They can either be the full name or just the register name. The following registers can be parsed
 *           (their names are given in the short form, just the DP element) and also a description of the return value given by sData:
 *           It will be zero padded to the nearest 16-bit word, except in the following cases:
 *   \li Disable `Enable' or `Disable' depending on the state of the channel.
 *           In the event that a broadcast is issued the return data will be comma separated.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring, that the fwCcpc_read function is not used and in this case an empty
 *           array should be passed and it will be ignored. Any errors found in this array will be reported via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseFeFpgaChannelReadRegisters(string ukl1Name, unsigned channel, unsigned feFpga, const dyn_string& registerNames, dyn_dyn_char ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if the full DPE is given we can match the register name.
  //Either we need to write to a specific FE FPGA channel or in the case of broadcasting we can write to FE FPGA channel 0 as all
  //FE FPGA channel settings are at the same register address, we merely need to write to the first channel with the appropriate data.
  //This is why broadcasting over the UKL1 channels is not any better, as you still have to write to the four registers anyway.
  const unsigned selectedChannel = FWUKL1_BROADCAST_CHANNEL==channel ? 0:channel;
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".FeFpga" + feFpga + ".Channel" + selectedChannel + ".";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);

  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Convert the data from a series of 32-bit words to a series of 16-bit words.
    const int exceptSize = dynlen(exceptionInfo);
    fwUkl1_convert32To16BitWords(ddcData[element], exceptionInfo);

    //Check that the data went through the first stage of parsing correctly.
    if ( (exceptSize != dynlen(exceptionInfo)) || (exceptSize == -1) ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaChannelReadRegisters(): Failed to convert the read 32-bit data to the 16-bit data stored in the FE FPGA registers.", "2");
      //Mark the error, either one or all FE FPGAs.
      if ( FWUKL1_BROADCAST_CHANNEL != channel )
	data[element] = "Fail";
      else
	data[element] = "Fail,Fail,Fail,Fail,Fail,Fail,Fail,Fail,Fail";
    }
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    else if ( (checkStatus && (0 != callStatusList[element])) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_parseFeFpgaChannelReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
      } else {
	errMsg += "an unknown error.";
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error, either one or all FE FPGAs.
      if ( FWUKL1_BROADCAST_CHANNEL != channel )
	data[element] = "Fail";
      else
	data[element] = "Fail,Fail,Fail,Fail,Fail,Fail,Fail,Fail,Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaChannelReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    else if ( ("Disable" == registerNames[element]) || ((prefix + "Disable") == registerNames[element]) ) {
      //Is it enabled or disabled?
      //First convert it to a single number.
      unsigned channelStates = fwCcpc_convertByteToDec(ddcData[element]);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of channels.
      //Start at the selected channel.
      //Determine the end of the loop.
      const unsigned endChannel = FWUKL1_BROADCAST_CHANNEL==channel ? (FWUKL1_FPGA_CHANNELS-1):channel;
      for (unsigned loopChannel = selectedChannel; loopChannel <= endChannel; ++loopChannel) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x01 == ((channelStates>>loopChannel) & 0x01) ) {
	  data[element] += "Disable";
	} else if ( 0x00 == ((channelStates>>loopChannel) & 0x01) ) {
	  data[element] += "Enable";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseFeFpgaChannelReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    //Can just return the read data to a hex string as the extra words have already been removed.
    //Should really do something for broadcasting, but there isn't actually anything to do.
    else {
      data[element] = fwCcpc_convertByteToHex(ddcData[element]);
    }

  }//for element

  //and we are done.
  return;
}

/*!
 * This takes the data read from the BE FPGA status register in the format as read from a data point and
 * converts it to a string value for display.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint elements to be
 *           read from, which corresponds to the appropriate hardware register. The accepted names are:
 *    \li TTCId Returns the first byte of the TTCrx ID as a hex string.
 *    \li TfcFifoRequest Returns `Ready' or `Not Ready'.
 *    \li SPI3TxBusy Returns `Busy' if the SPI3 transmit is busy or `Not busy' if it is not. Busy indicates
 *          data is being transmited from BE FPGA to the GBE.
 *    \li MepFifoFull Returns `Full' if the MEP FIFO is full or `Not full' if it is not. Indicates the amount of data
 *          present in the MEP FIFO.
 *    \li Throttled Returns `Throttled' or `Not throttled' depending on the state of the throttle.
 *    \li TfcL0Reset `Asserted' or `Not Asserted'.
 *    \li TfcL1Reset `Asserted' or `Not Asserted'.
 *    \li TfcBShort Decimal number.
 *    \li Tpa Decimal number.
 *    \li FragmentReady `Ready' or `Not Ready'.
 *    \li MepReady `Ready' or `Not Ready'.
 *    \li FragmentRequest `Requested' or `Not requested'.
 *    \li EventQValid `Valid' or `Not Valid'
 *    \li FragmenterState Decimal number.
 *    \li EthernetQValid `Valid' or `Not Valid'.
 *    \li SequenceError `Error' or `No Error'.
 *    \li MepCount Decimal number.
 *    \li TfcFifoOccupancy Decimal number.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring, that the fwCcpc_read function is not used and in this case an empty
 *           array should be passed and it will be ignored. Any errors found in this array will be reported via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseBeFpgaStatusRegisters(string ukl1Name, const dyn_string& registerNames, const dyn_dyn_char& ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if we are given the full DPE name we can match against it.
  const string prefix = ukl1Name + ".BeFpga.BeFpgaStatus.";
  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Used to check whether the function executed without a problem.
  int status = 0;
  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    if ( checkStatus && (0 != callStatusList[element]) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
	status = 1;
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
	status = -1;
      } else {
	errMsg += "an unknown error.";
	status = -1;
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error.
      data[element] = "Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseBeFpgaStatusRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    else if ( ((prefix + "TTCId") == registerNames[element]) || ("TTCId" == registerNames[element]) ) {
      //Return the TTCId as an 8 bit number by removing the extra bytes.
      //Note the deliberate use of `<', this is so we don't delete all the elements!
      //Calculate the maximum number of iterations for the loop now, as otherwise the length calculation will change
      //with each loop.
      const unsigned totalBytes = dynlen(ddcData[element]);
      for (unsigned byte = 1; byte < totalBytes; ++byte) {
	//Always remove the first element, until we only have one left.
	dynRemove(ddcData[element], 1);
      }
      //Now convert it to hex.
      data[element] = fwCcpc_convertByteToHex(ddcData[element]);
    }
    else if ( ((prefix + "TfcFifoRequest") == registerNames[element]) || ("TfcFifoRequest" == registerNames[element]) ) {
      //This is the index of the element that should contain the appropriate status bit.
      const int index = 3;
      if ( 0x100 == (ddcData[element][index]&0x100) ) {
	data[element] = "Requested";
      } else if ( 0x00 == (ddcData[element][index]&0x100) ){
	data[element] = "Not Requested";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "SPI3TxBusy") == registerNames[element]) || ("SPI3TxBusy" == registerNames[element]) ) {
      //Wish to look at the LSB.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x01 == (ddcData[element][index]&0x01) ) {
	data[element] = "Busy";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "Not busy";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "MepFifoFull") == registerNames[element]) || ("MepFifoFull" == registerNames[element]) ) {
      //Wish to look at the LSB.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x02 == (ddcData[element][index]&0x02) ) {
	data[element] = "Full";
      } else if ( 0x00 == (ddcData[element][index]&0x02) ){
	data[element] = "Not full";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "Throttled") == registerNames[element]) || ("Throttled" == registerNames[element]) ) {
      //Are the throttle signals asserted.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x0c == (ddcData[element][index]&0x0c) ) {
	data[element] = "Throttled";
      } else if ( 0x00 == (ddcData[element][index]&0x0c) ){
	data[element] = "Not throttled";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "TfcL0Reset") == registerNames[element]) || ("TfcL0Reset" == registerNames[element]) ) {
      //Is the L0 reset asserted?
      //Need to look at the LSB. High is asserted, low is deasserted.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x01 == (ddcData[element][index]&0x01) ) {
	data[element] = "Asserted";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "Deasserted";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "TfcL1Reset") == registerNames[element]) || ("TfcL1Reset" == registerNames[element]) ) {
      //Is the L1 reset asserted?
      //Need to look at the 2nd bit. High is asserted, low is deasserted.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x02 == (ddcData[element][index]&0x02) ) {
	data[element] = "Asserted";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "Deasserted";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "TfcBShort") == registerNames[element]) || ("TfcBShort" == registerNames[element]) ) {
      //Last 6-bits of the last broadcast command to be recieved.
      //Convert it to an int.
      const unsigned bShort = fwCcpc_convertByteToDec(ddcData[element]);
      //Allow PVSS to do the conversion to a string for us, but need to isolate the relevant bits first.
      sData[element] = (bShort>>2)&0x3f;
    }
    else if ( ((prefix + "Tpa") == registerNames[element]) || ("Tpa" == registerNames[element]) ) {
      //don't know what TPA does. Will figure it out eventually, just return the 4 bits for now.
      //It is bits 16 to 20. Convert it to an int.
      const unsigned tpa = fwCcpc_convertByteToDec(ddcData[element]);
      //Allow PVSS to do the conversion to a string for us, but need to isolate the relevant bits first.
      sData[element] = (tpa>>16)&0x0f;
    }
    else if ( ((prefix + "FragmentReady") == registerNames[element]) || ("FragmentReady" == registerNames[element]) ) {
      //MEP fragment ready for transmission.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = 1;
      if ( 0x01 == (ddcData[element][index]&0x01) ) {
	data[element] = "Ready";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "Not Ready";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "MepReady") == registerNames[element]) || ("MepReady" == registerNames[element]) ) {
      //MEP ready for transmission.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = 1;
      if ( 0x02 == (ddcData[element][index]&0x02) ) {
	data[element] = "Ready";
      } else if ( 0x00 == (ddcData[element][index]&0x02) ){
	data[element] = "Not Ready";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "FragmentRequest") == registerNames[element]) || ("FragmentRequest" == registerNames[element]) ) {
      //Think this is when the TFC requests a fragment.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = 1;
      if ( 0x04 == (ddcData[element][index]&0x04) ) {
	data[element] = "Requested";
      } else if ( 0x00 == (ddcData[element][index]&0x04) ){
	data[element] = "Not Requested";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "EventQValid") == registerNames[element]) || ("EventQValid" == registerNames[element]) ) {
      //No idea.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = 1;
      if ( 0x08 == (ddcData[element][index]&0x08) ) {
	data[element] = "Valid";
      } else if ( 0x00 == (ddcData[element][index]&0x08) ){
	data[element] = "Not Vaild";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "FragmenterState") == registerNames[element]) || ("FragmenterState" == registerNames[element]) ) {
      //Provides details of the fragmenter state in the UKL1 firmware. Only a firmware engineer will understand this...
      const unsigned state = fwCcpc_convertByteToDec(ddcData[element]);
      //Allow PVSS to do the conversion to a string for us, but need to isolate the relevant bits first.
      sData[element] = state&0xff;
    }
    else if ( ((prefix + "EthernetQValid") == registerNames[element]) || ("EthernetQValid" == registerNames[element]) ) {
      //No idea.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x01 == (ddcData[element][index]&0x01) ) {
	data[element] = "Valid";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "Not Vaild";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "SequenceError") == registerNames[element]) || ("SequenceError" == registerNames[element]) ) {
      //No idea.
      //This is the index of the element that should contain the appropriate status bit.
      const int index = dynlen(ddcData[element]);
      if ( 0x01 == (ddcData[element][index]&0x01) ) {
	data[element] = "Error";
      } else if ( 0x00 == (ddcData[element][index]&0x01) ){
	data[element] = "No Error";
      } else {
	data[element] = "Fail";
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ParseBeFpgaStatusData(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element][1]) + ".", "2");
      }
    }
    else if ( ((prefix + "MepCount") == registerNames[element]) || ("MepCount" == registerNames[element]) ) {
      //Number of MEPs waiting to be transmitted.
      const unsigned count = fwCcpc_convertByteToDec(ddcData[element]);
      //Allow PVSS to do the conversion to a string for us, but need to isolate the relevant bits first.
      sData[element] = (count>>2)&0x03;
    }
    else if ( ((prefix + "TfcFifoOccupancy") == registerNames[element]) || ("TfcFifoOccupancy" == registerNames[element]) ) {
      //Not completely sure.
      const unsigned occ = fwCcpc_convertByteToDec(ddcData[element]);
      //Allow PVSS to do the conversion to a string for us, but need to isolate the relevant bits first.
      sData[element] = (occ>>8)&0xff;
    }
    else {
      //For those values that don't need to be treated.
      data[element] = fwCcpc_convertByteToHex(ddcData[element]);
    }
  }

  //Done.
  return;
}

// =========
//    GBE
// =========

/*!
 * This function takes the data for the various GBE registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerName List of GBE register names that are to be written to. Returned with the appropriate name
 *           prefix by reference. The allowed values are:
 *   \li JTAGID Default data format, see below.
 *   \li MACSoftReset Default data format, see below.
 *   \li RxFifoPrtReset Default data format, see below.
 *   \li TxFifoPrtReset Default data format, see below.
 *   \li PortEnable `Enable' or `Disable' in a comma delimited format. Values for all four ports must be present.
 *   \li RxFifoEnable Default data format, see below.
 *   \li Mode Default data format, see below.
 *   \li Clock Default data format, see below.
 *   \li SPI3ConfigTrnmtGlobal Default data format, see below.
 *   \li SPI3ConfigRcv Default data format, see below.
 *   \li RxStatus Default data format, see below.
 *   \li PHYCtrl Default data format, see below.
 *   \li PHYData Default data format, see below.
 *   \li MDIOControl Default data format, see below.
 * \param  sData Data to be written to the registers in the form of a hex string. The default value is an 8-character
 *           hex string, other values are accepted for specific registers as described above.
 * \param  ddcData sData converted into a dyn_dyn_char for use with the fwCcpc_write or fwHw_saveRecipe function. Returned by reference.
 * \param  masks Masks that are associated with the write data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array.
 */
void fwUkl1_parseGbeWriteRegisters(string ukl1Name, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".GBE.";

  //Add prefix to each name and convert the data into a byte array.
  //This indicates if there was a formatting error and we should remove those elements from the list or if they can be parsed.
  bool formatError;
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //This way we only need to indicate if something went wrong.
    formatError = FALSE;
    //Deal with the different formats.
    if ( "PortEnable" == registerNames[element] ) {
      //Just parse this data as broadcast data, all the other data can be left alone.
      //Technically this function takes a dyn_string, however attempting to makeDynString causes the function
      //to fail horribly. It is probably something to do with temporaries and const &.
      dyn_dyn_string sParsedData = fwUkl1_parseBroadcastData(sData[element], exceptionInfo);
      //Holds the data to be written as it is determined.
      unsigned disable = 0;
      //Now loop over the appropriate range of GBE ports.
      for (unsigned port = 0; port < FWUKL1_GBE_PORTS; ++port) {
	//Create the data to write.
	if ( "Enable" == sParsedData[1][port+1] ) {
	  //Set the relevant bit high.
	  disable |= 1<<port;
	} else if ( "Disable" == sParsedData[1][port+1] ) {
	  //Do nothing the bit is already zero.
	} else {
	  formatError = TRUE;
	  //Indicate who was lost before we remove the register name from the list, details of failure indicated by the parse IP address function.
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseGbeWriteRegister(): Failed to parse " + registerNames[element] + ". Expected `Disable' or `Enable', got " + sParsedData[1][port+1] + ". Removed from list of writes.", "2");
	}
      }
      //Now convert the int data back to a string.
      sData[element] = fwCcpc_convertDecToHex(disable);
    }

    if ( !formatError ) {
      //Put the data and register names in a format that the write can use.
      //Add the prefix to convert it to a datapoint name.
      registerNames[element] = prefix + registerNames[element];
      //Pad the string.
      fwUkl1_padString(sData[element], 4);
     //Convert from a hex string to a byte array.
      ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
      //Everything needs a 32-bit mask.
      masks[element]   = makeDynChar(0xff, 0xff, 0xff, 0xff);
    } else {
      //Failed to parse the register, exception has already been raised...
      //Remove from register list.
      dynRemove(registerNames, element);
      //Remove the original data so it everything is kept in the same order as the registerNames
      dynRemove(sData, element);
      //Masks and ddcData aren't written so they are not affected.
      //Decrement the element number as we need to start this loop at the same number as before.
      --element;
    }
  }

 return;
}

/*!
 * This takes the data read from the GBE registers in the format as read from fwCcpc_read() or fwHw_getRecipe() and
 * returns the values as a decriptive string.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint elements to be
 *           read from, which corresponds to the appropriate hardware register. The accepted names are:
 *   \li JTAGID Default data format, see below.
 *   \li MACSoftReset Default data format, see below.
 *   \li RxFifoPrtReset Default data format, see below.
 *   \li TxFifoPrtReset Default data format, see below.
 *   \li PortEnable `Enable' or `Disable' in a comma delimited format. Values for all four ports must be present.
 *   \li RxFifoEnable Default data format, see below.
 *   \li Mode Default data format, see below.
 *   \li Clock Default data format, see below.
 *   \li SPI3ConfigTrnmtGlobal Default data format, see below.
 *   \li SPI3ConfigRcv Default data format, see below.
 *   \li RxStatus Default data format, see below.
 *   \li PHYCtrl Default data format, see below.
 *   \li PHYData Default data format, see below.
 *   \li MDIOControl Default data format, see below.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring or reading from a recipe, that the fwCcpc_read function is not used and in
 *           this case an empty array should be passed and it will be ignored. Any errors found in this array will be reported
 *           via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseGbeReadRegisters(string ukl1Name, const dyn_string& registerNames, const dyn_dyn_char& ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if we are given the full DPE name we can match against it.
  const string prefix = ukl1Name + ".GBE.";
  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Used to check whether the function executed without a problem.
  int status = 0;

  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    if ( (checkStatus && (0 != callStatusList[element])) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_parseGbeReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
	status = 1;
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
	status = -1;
      } else {
	errMsg += "an unknown error.";
	status = -1;
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error.
      data[element] = "Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseGbeReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    //Deal with the port enable settings.
    else if (("PortEnable" == registerNames[element]) || ((prefix + "PortEnable") == registerNames[element]) ) {
      //Is it enabled or disabled?
      //First convert it to a string and trim off the last 8 characters if it is from the recipe, otherwise there will be no change.
      //First convert it to a single number.
      const string sTmp = substr(fwCcpc_convertByteToHex(ddcData[element]), 0, 8);
      unsigned portStates = fwCcpc_convertHexToDec(sTmp);
      //This is to mark whether it is the first time round the loop.
      bool firstTime = TRUE;
      //Now loop over the appropriate range of channels.
      for (unsigned port = 0; port < FWUKL1_GBE_PORTS; ++port) {
	//If it is not the first time round the loop we should add a comma before we write the next setting.
	if ( !firstTime )
	  data[element] += ",";
	if ( 0x01 == ((portStates>>port) & 0x01) ) {
	  data[element] += "Enable";
	} else if ( 0x00 == ((portStates>>port) & 0x01) ) {
	  data[element] += "Disable";
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseGbeReadRegisters(): " + registerNames[element] + " has unrecognised value: " + fwCcpc_convertByteToHex(ddcData[element]) + ".", "2");
	  data[element] += "Fail";
	}
	//If it is the first time, it isn't any more.
	if ( firstTime )
	  firstTime = FALSE;
      }
    }
    else {
      //Can just return the read data to a hex string as the extra words have already been removed.
      //Trim off the mask in the case of the recipe data, otherwise the string will only be that big.
      data[element] = substr(fwCcpc_convertByteToHex(ddcData[element]), 0, 8);
    }
  }//for element

  //and we are done.
  return;
}

/*!
 * This function takes the data for the various GBE registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  port Output port on the GBE card that is to be set.
 * \param  registerName List of GBE register names that are to be written to. Returned with the appropriate name
 *           prefix by reference. The allowed values are:
 *   \li Duplex Default data format, see below.
 *   \li MAC Default data format, see below.
 *   \li Config Default data format, see below.
 *   \li TXFifoThreshold Decimal number.
 *   \li TXFifoLowWtrmrk Decimal number.
 *   \li TXFifoHighWtrmrk Decimal number.
 *   \li PHYControl Default data format, see below.
 *   \li PHYStatus Default data format, see below.
 *   \li TXStat Default data format, see below.
 *   \li RXStat Default data format, see below.
 * \param  sData Data to be written to the registers in the form of a hex string. The default value is an 8-character
 *           hex string, other values are accepted for specific registers as described above.
 * \param  ddcData sData converted into a dyn_dyn_char for use with the fwCcpc_write function. Returned by reference.
 * \param  masks Masks that are associated with the write data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array.
 */
void fwUkl1_parseGbePortWriteRegisters(string ukl1Name, unsigned port, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".GBE.PORT" + port + ".";
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //Need a 32-bit mask, the same for all.
    masks[element]   = makeDynChar(0xff, 0xff, 0xff, 0xff);
    //Deal with the different formats.
    if ( ("TXFifoThreshold" == registerNames[element]) ||
	 ("TXFifoLowWtrmrk" == registerNames[element]) ||
	 ("TXFifoHighWtrmrk" == registerNames[element]) ) {
      //These will be decimal numbers.
      sData[element] = fwCcpc_convertDecToHex(sData[element]);
    }
    //Everything needs to be padded the string.
    fwUkl1_padString(sData[element], 4);
    //Put the data and register names in a format that the write can use.
    //Add the prefix to convert it to a datapoint name.
    registerNames[element] = prefix + registerNames[element];

    //Convert from a hex string to a byte array.
    ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
    
  }

  //done.
  return;
}

/*!
 * This takes the data read from the GBE port registers in the format as read from fwCcpc_read() or fwHw_getRecipe() and
 * returns the values as a decriptive string.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  port Output port on the GBE card that is to be set.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint elements to be
 *           read from, which corresponds to the appropriate hardware register. The accepted names are:
 *   \li Duplex Default data format, see below.
 *   \li MAC Default data format, see below.
 *   \li Config Default data format, see below.
 *   \li TXFifoThreshold Decimal number.
 *   \li TXFifoLowWtrmrk Decimal number.
 *   \li TXFifoHighWtrmrk Decimal number.
 *   \li PHYControl Default data format, see below.
 *   \li PHYStatus Default data format, see below.
 *   \li TXStat Default data format, see below.
 *   \li RXStat Default data format, see below.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring or reading from a recipe, that the fwCcpc_read function is not used and in
 *           this case an empty array should be passed and it will be ignored. Any errors found in this array will be reported
 *           via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseGbePortReadRegisters(string ukl1Name, unsigned port, const dyn_string& registerNames, const dyn_dyn_char& ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if we are given the full DPE name we can match against it.
  const string prefix = ukl1Name + ".GBE.PORT" + port + ".";
  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Used to check whether the function executed without a problem.
  int status = 0;

  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    if ( (checkStatus && (0 != callStatusList[element])) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_parseGbeReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
	status = 1;
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
	status = -1;
      } else {
	errMsg += "an unknown error.";
	status = -1;
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error.
      data[element] = "Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseGbePortReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    //Deal with the different formats.
    else if ( ( (prefix + "TXFifoThreshold" == registerNames[element]) ) || ("TXFifoThreshold" == registerNames[element]) ||
	      ( (prefix + "TXFifoLowWtrmrk") == registerNames[element] ) || ("TXFifoLowWtrmrk" == registerNames[element]) ||
	      ( (prefix + "TXFifoHighWtrmrk") == registerNames[element] ) || ("TXFifoHighWtrmrk" == registerNames[element]) ) {
      //These want to be decimal numbers. Only ever interested in the last 8 characters, from a recipe the mask would be there.
      const string sTmp = substr(fwCcpc_convertByteToHex(ddcData[element]), 0, 8);
      data[element] = fwCcpc_convertHexToDec(sTmp);
    }
    else {
      //Can just return the read data to a hex string.
      data[element] = substr(fwCcpc_convertByteToHex(ddcData[element]), 0, 8);
    }

  }//for element
  //done.
  return;
}

// =========
//    I2C
// =========

/*!
 * This function takes the data for the various TTCrx registers described by a string and converts it to data that can
 * used with fwCcpc_write() or fwHw_saveRecipe(). It allows data to be given in a more human readable format and converts
 * it to the byte array required for writting to the CCPC. It also returns the appropriate mask that must be used when
 * writing the data. If being used with fwHw_saveRecipe() then the mask will have to be appended to the data array.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerName List of GBE register names that are to be written to. Returned with the appropriate name
 *           prefix by reference. The allowed values are:
 *   \li IACId Requires a two byte word.
 *   \li I2CId Default data format, see below.
 *   \li FineDelay1 Default data format, see below.
 *   \li FineDelay2 Default data format, see below.
 *   \li CoarseDelay Default data format, see below.
 *   \li Control Default data format, see below.
 *   \li Status Default data format, see below.
 *   \li Config1 Default data format, see below.
 *   \li Config2 Default data format, see below.
 *   \li Config3 Default data format, see below.
 *   \li SingleErrCnt Default data format, see below.
 *   \li DoubleErrCnt Default data format, see below.
 *   \li SEUErrCnt Requires a two byte word, read only.
 *   \li BunchCnt Requires a two byte word, read only.
 *   \li EventCnt Requires a three byte word, read only.
 * \param  sData Data to be written to the registers in the form of a hex string. The default value is an 2-character
 *           hex string, other values are accepted for specific registers as described above.
 * \param  ddcData sData converted into a dyn_dyn_char for use with the fwCcpc_write function. Returned by reference.
 * \param  masks Masks that are associated with the write data.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * Improperly formatted data will be removed from the list of write registers and will be indicated by an increased size of the
 * exceptionInfo array.
 */
void fwUkl1_parseTtcrxWriteRegisters(string ukl1Name, dyn_string& registerNames, dyn_string sData, dyn_dyn_char& ddcData, dyn_dyn_char& masks, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".TTCRX.";

  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= dynlen(registerNames); ++element) {
    //The execptions are those that require more than one byte and hence a larger mask.
    if ( "IACId" == registerNames[element] ) {
      //This requires a two byte mask and padding.
      masks[element]   = makeDynChar(0xff, 0xff);
      fwUkl1_padString(sData[element], 2);
    }
    else if ( "SingleErrCnt" == registerNames[element] ) {
      //This requires a two byte mask and padding.
      masks[element]   = makeDynChar(0xff, 0xff);
      fwUkl1_padString(sData[element], 2);
    }
    else if ( "BunchCnt" == registerNames[element] ) {
      //This requires a two byte mask and padding.
      masks[element]   = makeDynChar(0xff, 0xff);
      fwUkl1_padString(sData[element], 2);
    }
    else if ( "EventCnt" == registerNames[element] ) {
      //This requires a three byte mask and padding.
      masks[element]   = makeDynChar(0xff, 0xff, 0xff);
      fwUkl1_padString(sData[element], 3);
    }
    else {
      //The standard only require an 8-bit mask.
      masks[element]   = makeDynChar(0xff);
      //Pad the string, such that it is only 1-byte big.
      fwUkl1_padString(sData[element], 1);
    }
    
    //Put the data and register names in a format that the write can use.
    //Add the prefix to convert it to a datapoint name.
    registerNames[element] = prefix + registerNames[element];
    //Convert from a hex string to a byte array.
    ddcData[element] = fwCcpc_convertHexToByte( sData[element] );
  }

  return;
}

/*!
 * This takes the data read from the TTCrx registers in the format as read from fwCcpc_read() or fwHw_getRecipe() and
 * returns the values as a decriptive string.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint elements to be
 *           read from, which corresponds to the appropriate hardware register. The accepted names are:
 *   \li IACId Requires a two byte word.
 *   \li I2CId Default data format, see below.
 *   \li FineDelay1 Default data format, see below.
 *   \li FineDelay2 Default data format, see below.
 *   \li CoarseDelay Default data format, see below.
 *   \li Control Default data format, see below.
 *   \li Status Default data format, see below.
 *   \li Config1 Default data format, see below.
 *   \li Config2 Default data format, see below.
 *   \li Config3 Default data format, see below.
 *   \li SingleErrCnt Default data format, see below.
 *   \li DoubleErrCnt Default data format, see below.
 *   \li SEUErrCnt Requires a two byte word, read only.
 *   \li BunchCnt Requires a two byte word, read only.
 *   \li EventCnt Requires a three byte word, read only.
 * \param  ddcData This contains the raw data that has been read back from the PVSS datapoint.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings. In the event that the read failed it will contain the text `Fail'.
 * \param  callStatusList This is the status list returned by the fwCcpc_read function and if the data to be parsed was generated
 *           by the use of this function then a callStatusList should be passed in order that errors can be checked for.
 *           It is possible, for example via monitoring or reading from a recipe, that the fwCcpc_read function is not used and in
 *           this case an empty array should be passed and it will be ignored. Any errors found in this array will be reported
 *           via the exceptionInfo.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_parseTtcrxReadRegisters(string ukl1Name, const dyn_string& registerNames, const dyn_dyn_char& ddcData, dyn_string& data, const dyn_int& callStatusList, dyn_string& exceptionInfo) {
  //Define the string to be added to each of the register names such that if we are given the full DPE name we can match against it.
  const string prefix = ukl1Name + ".TTCRX.";
  //Determine the number of registers for which we are reading the data.
  const unsigned numRegs = dynlen(registerNames);
  //Determine whether we should be checking the status list. If it is empty don't look in it.
  const bool checkStatus = (0==dynlen(callStatusList) ? FALSE:TRUE);
  //Used to check whether the function executed without a problem.
  int status = 0;

  //Loop over all the given registers and put the data into a single string for return.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //This series of `if' statements determines how to handle the possible raw status register data we may receive.
    //Check for errors. As we are using AND the second case will only be evaluated provided checkStatus is not empty
    //and hence we should avoid out of bounds errors.
    //Check for both the case when the register names has the appropriate prefix and for the case where it doesn't.
    if ( (checkStatus && (0 != callStatusList[element])) ) {
      //Construct the error message.
      string errMsg = "fwUkl1_parseTtcrxReadRegisters(): " + registerNames[element] + " encountered ";
      if (callStatusList[element] > 0) {
	errMsg += "a problem writing to the CCPC server.";
	status = 1;
      } else if (callStatusList[element] < 0) {
	errMsg += "a problem with PVSS.";
	status = -1;
      } else {
	errMsg += "an unknown error.";
	status = -1;
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
      //Mark the error.
      data[element] = "Fail";
    }
    else if ( 0 == dynlen(ddcData[element]) ) {
      //No data, raise an exception.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_parseTtcrxReadRegisters(): No data read to parse for " + registerNames[element] + ".", "2");
    }
    else {
      //Currently no exceptions, all either 1 or 2 bytes big. Just return their hex values.
      //These want to be decimal numbers. Only ever interested in the last 8 characters, from a recipe the mask would be there.
      data[element] = substr(fwCcpc_convertByteToHex(ddcData[element]), 0, 8);
    }

  }//for element.

  //Done.
  return;
}

// =============================
//   HARDWARE ACCESS FUNCTIONS
// =============================

// ===========
//    FPGAS
// ===========

/*!
 * Writes to a given set of UKL1 registers. Data should be in a form acceptable to the fwUkl1_parseBeFpgaWriteRegisters() as it will
 * first be passed through this function before being written to the hardware.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
* \param  registerNames dyn_string containing the names of the registers that need to be read. It need not contain the full
 *           path of the register i.e. the datapoint name and hardware chip (BeFpga, GBE) must be neglected. Care must be
 *           taken that the registers requested exist.
 * \param  data By default a 16-bit value that is to be written to the register represented as a 4-character hex string,
 *           it need not be zero padded. Certain settings accept or require an alternative format as described in the
 *           fwUkl1_parseBeFpgaRegisters function.
 * \param  verifyWrite If this is set to TRUE write will be verified to be successful. Not used if writeDb=TRUE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * If the recipe name does not exist it will be created. If the recipe type does not exist then the recipe may lose its type and
 * all data associated with it. No checks are performed on the recipe type at present.
 */
void fwUkl1_beFpgaWrite(const string& ukl1Name, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, const dyn_string& data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Check that the register names and data arrays are the same size.
  //Save only the length of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): Fail to write to registers as the register list to write to does not match the size of the data list to be written to them.", "1");
    return;
  }

  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseBeFpgaWriteRegisters(ukl1Name, registerNames, data, ddcData, masks, exceptionInfo);
  //Check that we were actually able to parse the data.
  if ( (exceptSize == dynlen(exceptionInfo)) && (exceptSize != -1) ) {
    //The data and registers have been parsed
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen)
	      fullDdcData[index] = ddcData[element];
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe
    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//See if we want to verify that the write was successfull.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_beFpgaWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): An unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_beFpgaWrite(): Failed to parse the data given to write, did not access hardware.", "1");
  }

  //Done.
  return;
}

/*!
 * This reads from a given set of registers on the BE FPGAs.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string containing the names of the registers that need to be read. It need not contain the full
 *           path of the register i.e. the datapoint name and hardware chip (BeFpga, GBE) must be neglected. Care must be
 *           taken that the registers requested exist.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register. The details of this string
 *           can be found in fwUkl1_parseBeFpgaReadRegister() documentation. Will return `Fail' in the event the read is not successful.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_beFpgaRead(const string& ukl1Name, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& excepitonInfo) {
  //Make sure we have an empty data array.
  dynClear(data);
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".BeFpga.";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);
  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwhw_getRecipe())
    else
      fwException_raise(exceptionInfo, "fwUkl1_beFpgaRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");
    
  }//if(readRecipe)
  else {
    //The read status will need to be checked to ensure that the function succeeded and there is a valid call list.
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }//else(readRecipe)

  //Now parse the read data to see if it is any good. Let the parsing function sort out any errors.
  fwUkl1_parseBeFpgaReadRegisters(ukl1Name, registerNames, ddcData, data, callStatusList, exceptionInfo);
	
  //And we are done, as I said the parsing function should have sorted out any errors.
  return;
}

/*!
 * This reads the BE FPGA status registers.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint elements to be
 *           read from, which corresponds to the appropriate hardware register. The accepted names can be found in the
 *           documentation for fwUkl1_parseBeFpgaStatusData.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will either be a hex representation, zero padded to the nearest byte boundary or a text description
 *           of the settings.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_beFpgaStatusRead(string ukl1Name, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Make sure we have an empty data array and status list.
  dynClear(data);
  dynClear(callStatusList);
  //Define the string to be added to each of the register names such that they are read from the appropriate datapoint element.
  const string prefix = ukl1Name + ".BeFpgaStatus.";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data in the form that the fwCcpc_read function returns.
  dyn_dyn_char ddcData;
  //Holds the status of each read, dealt with by the parsing function.
  dyn_int callStatusList;
  fwCcpc_read(registerNames, ddcData, callStatusList);

  fwUkl1_parseBeFpgaStatusData(ukl1Name, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //Return. Errors will have been dealt with by the parsing function.
  return;
}

/*!
 * Writes to a given set of UKL1 registers. Data should be in a form acceptable to the fwUkl1_parseFeFpgaWriteRegisters() as it will
 * first be passed through this function before being written to the hardware.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3, FWUKL1_BROADCAST_FPGA.  FWUKL1_BROADCAST_FPGA is a 
 *           software creatation that allows all the FE FPGAs to be written to in a single command. It saves on the number
 *           of hardware accesses by a factor of four.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
 * \param  registerNames dyn_string containing the names of the registers that need to be read. It need not contain the full
 *           path of the register i.e. the datapoint name and hardware chip (BeFpga, GBE) must be neglected. Care must be
 *           taken that the registers requested exist.
 * \param  data By default a 16-bit value that is to be written to the register represented as a hex string, it need not be zero padded.
 *           Certain settings accept or require an alternative format as described in fwUkl1_parseFeFpgaRegisters. For a single FPGA write
 *           the data should be in the format above with one element for each register. If broadcasting then the data must be comma
 *           delimited and four values MUST be present e.g. To disable a single front end `Enable,Enable,Disable,Enable'.
 * \param  verifyWrite If this is set to TRUE then the register will be read back from to ensure that
 *           the data written corresponds to that read. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * If the recipe name does not exist it will be created. If the recipe type does not exist then the recipe may lose its type and
 * all data associated with it. No checks are performed on the recipe type at present.
 */
void fwUkl1_feFpgaWrite(string ukl1Name, unsigned feFpga, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, const dyn_string& data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Ensure that we are in broadcast mode if writing to a recipe.
  if ( writeRecipe && (FWUKL1_BROADCAST_FPGA != feFpga) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): When writing to a configurationDB recipe broadcast mode must be used for the recipe.", "1");
    return;
  }
  //Check that the register names and data arrays are the same size.
  //Save only the lenght of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): Number of registers to write to, " + numRegs + ", is not equal to the amount of data given to be written, " + dynlen(data) + ".", "1");
    return;
  }


  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseFeFpgaWriteRegisters(ukl1Name, feFpga, registerNames, data, ddcData, masks, exceptionInfo);

  //Check that we were able to validate and parse the data before attempting to write.
  if ( exceptSize == dynlen(exceptionInfo) ) {
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen)
	      fullDdcData[index] = ddcData[element];
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe
    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//See if we want to verify that the write was successfull.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_feFpgaWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): A fatal error and an unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaWrite(): Failed to parse input data, not data was written.", "1");
  }

  //and we are done.
  return;
}

/*!
 * This reads from a given set of registers on the FE FPGAs. It is capable of reading from either an individual FE FPGA
 * or all FE FPGAs in broadcast mode. The source of the data can be either the UKL1 board or a recipe. If reading from
 * a recipe the broadcast FPGA must be used.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  feFpga FE FPGA number that is to be written to. Range: 0 to 3, FWUKL1_BROADCAST_FPGA.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_feFpgaRead(string ukl1Name, unsigned feFpga, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Ensure that we are in broadcast mode if writing to a recipe.
  if ( readRecipe && (FWUKL1_BROADCAST_FPGA != feFpga) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaRead(): When writing to a configurationDB recipe broadcast mode must be used for the recipe.", "1");
    return;
  }
  //Make sure we have an empty data array.
  dynClear(data);
  //Decide which FE FPGA we are going to read from. In the event of a broadcast because of the way the registers are set up in hardware
  //and implemented as datapoints we can just read from FE FPGA 0 and get all the FE FPGA settings provided we decode it properly.
  unsigned fpga = 0;
  //Only need to changed from initial value if not broadcasting.
  if ( FWUKL1_BROADCAST_FPGA != feFpga )
    fpga = feFpga;
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".FeFpga" + fpga + ".";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);
  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwhw_getRecipe())
    else
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");
    
  }//if(readRecipe)
  else {
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }

  //Parse the read data and let it deal with any errors that occur.
  fwUkl1_parseFeFpgaReadRegisters(ukl1Name, feFpga, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //and we are done.
  return;
}

/*!
 * This writes to a given set of registers on a FE FPGA channel. It can be used to write to either the registers on the UKL1
 * board or to a configurationDB recipe. If writing to a recipe broadcast mode must be used.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  channel Input channel number on a FE FPGA that is to be written to. Range: 0 to 8, FWUKL1_BROADCAST_CHANNEL.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register. The accepted values are described in
 *           fwUkl1_parseFeFpgaChannelRegisters();
 * \param  data By default a 16-bit value that is to be written to the register represented as a hex string, it need not be zero padded.
 *           Certain settings accept or require an alternative format as described in fwUkl1_parseFeFpgaChannelWriteRegisters() documentation.
 * \param  verifyWrite If this is set to TRUE then the register will be read back from to ensure that
 *           the data written corresponds to that read. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * If the recipe name does not exist it will be created. If the recipe type does not exist then the recipe may lose its type and
 * all data associated with it. No checks are performed on the recipe type at present.
 */
void fwUkl1_feFpgaChannelWrite(string ukl1Name, unsigned channel, unsigned feFpga, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, const dyn_string& data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Ensure that we are in broadcast mode if writing to a recipe.
  if ( writeRecipe && (FWUKL1_BROADCAST_CHANNEL != channel) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): When writing to a configurationDB recipe broadcast mode must be used for the recipe.", "1");
    return;
  }
  //Check that the register names and data arrays are the same size.
  //Save only the lenght of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): Number of registers to write to, " + numRegs + ", is not equal to the amount of data given to be written, " + dynlen(data) + ".", "1");
    return;
  }

  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseFeFpgaChannelWriteRegisters(ukl1Name, channel, feFpga, registerNames, data, ddcData, masks, exceptionInfo);
  //Check that we were able to parse the data.
  if ( exceptSize == dynlen(exceptionInfo) ) {
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen)
	      fullDdcData[index] = ddcData[element];
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe

    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//Check to see if we want to verify the write.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_feFpgaChannelWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): A fatal error and an unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelWrite(): Failed to parse input data, not data was written.", "1");
  }

  //and we are done.
  return;
}

/*!
 * This reads from a given set of registers on a FE FPGA channel. It can read from a single FE FPGA channel or all the channels on a FE FPGA
 * in broadcast mode. The reads can either be from the hardware or they can be done from a configurationDB recipe. If reading from a recipe
 * broadcast mode must be used.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  channel Input channel number on a FE FPGA that is to be written to. Range: 0 to 8, FWUKL1_BROADCAST_CHANNEL.
 * \param  feFpga Number of the FE FPGA that is to be written. Range:0 to 3.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_feFpgaChannelRead(string ukl1Name, unsigned channel, unsigned feFpga, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Ensure that we are in broadcast mode if writing to a recipe.
  if ( readRecipe && (FWUKL1_BROADCAST_CHANNEL != channel) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelRead(): When writing to a configurationDB recipe broadcast mode must be used for the recipe.", "1");
    return;
  }
  //Make sure we have an empty data array.
  dynClear(data);
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  //Either we need to write to a specific FE FPGA channel or in the case of broadcasting we can write to FE FPGA channel 0 as all
  //FE FPGA channel settings are at the same register address, we merely need to write to the first channel with the appropriate data.
  //This is why broadcasting over the UKL1 channels is not any better, as you still have to write to the four registers anyway.
  const unsigned selectedChannel = FWUKL1_BROADCAST_CHANNEL==channel ? 0:channel;
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".FeFpga" + feFpga + ".Channel" + selectedChannel + ".";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);
  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwHw_getRecipe)
    else
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_feFpgaChannelRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");
    
  }//if(readRecipe)
  else {
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }

  //Parse the data we just read and let it deal with any errors that occurred.
  fwUkl1_parseFeFpgaChannelReadRegisters(ukl1Name, channel, feFpga, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //and we are done
  return;
}


// =========
//    GBE
// =========

/*!
 * This writes to a given set of registers on the giga-bit ethernet (GBE) card. It can be used to write to either the
 * registers on the UKL1 board or to a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register. The accepted values are described in
 *           fwUkl1_parseGbeWriteRegisters();
 * \param  data By default a 32-bit value that is to be written to the register represented as a hex string, it need not be zero padded.
 *           Certain settings accept or require an alternative format as described in fwUkl1_parseGbeWriteRegisters() documentation.
 * \param  verifyWrite If this is set to TRUE then the register will be read back from to ensure that
 *           the data written corresponds to that read. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_gbeWrite(string ukl1Name, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, const dyn_string& data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Check that the register names and data arrays are the same size.
  //Save only the lenght of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): Number of register names (" + numRegs + ") is not equal to the amount of data (" + dynlen(data) + ").", "1");
    return;
  }

  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseGbeWriteRegisters(ukl1Name, registerNames, data, ddcData, masks, exceptionInfo);
  //Check that we were able to parse the data.
  if ( exceptSize == dynlen(exceptionInfo) ) {
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen) {
	      fullDdcData[index] = ddcData[element];
	    }
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe
    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//Check to see if we want to verify the write.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_gbeWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): A fatal error and an unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeWrite(): Failed to parse input data, not data was written.", "1");
  }

  //and we are done.
  return;
}

/*!
 * This reads from a given set of general GBE registers. The reads can either be from the hardware or they can be
 * done from a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will be zero padded to the nearest 32-bit word.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_gbeRead(string ukl1Name, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Make sure we have an empty data array.
  dynClear(data);
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".GBE.";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);

  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwhw_getRecipe())
    else
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbeRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");
    
  }//if(readRecipe)
  else {
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }

  //Parse the data we just read and let it deal with any errors that occurred.
  fwUkl1_parseGbeReadRegisters(ukl1Name, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //and we are done
  return;

}

/*!
 * This writes to a given set of registers on a GBE port. It can be used to write to either the
 * registers on the UKL1 board or to a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  port Output port on the GBE card that is to be set.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register. The accepted values are described in
 *           fwUkl1_parseGbeWriteRegisters();
 * \param  data By default a 32-bit value that is to be written to the register represented as a hex string, it need not be zero padded.
 *           Certain settings accept or require an alternative format as described for each register above.
 * \param  verifyWrite If this is set to TRUE then the register will be read back from to ensure that
 *           the data written corresponds to that read. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_gbePortWrite(string ukl1Name, unsigned port, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, dyn_string data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Check that the register names and data arrays are the same size.
  //Save only the lenght of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): On port " + port + ", the number of register names (" + numRegs + ") is not equal to the amount of data (" + dynlen(data) + ").", "1");
    return;
  }

  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseGbePortWriteRegisters(ukl1Name, port, registerNames, data, ddcData, masks, exceptionInfo);
  //Check that we were able to parse the data.
  if ( exceptSize == dynlen(exceptionInfo) ) {
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen)
	      fullDdcData[index] = ddcData[element];
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe
    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//Check to see if we want to verify the write.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_gbePortWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): A fatal error and an unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortWrite(): Failed to parse input data, not data was written.", "1");
  }

  //and we are done.
  return;
}

/*!
 * This reads from a given set of GBE port registers. The reads can either be from the hardware or they can be
 * done from a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  port Output port on the GBE card that is to be set.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will be zero padded to the nearest 32-bit word.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_gbePortRead(string ukl1Name, unsigned port, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Make sure we have an empty data array.
  dynClear(data);
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".GBE.PORT" + port + ".";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);

  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwhw_getRecipe())
    else
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_gbePortRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");

  }//if(readRecipe)
  else {
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }

  //Parse the data we just read and let it deal with any errors that occurred.
  fwUkl1_parseGbePortReadRegisters(ukl1Name, port, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //and we are done
  return;
}

// ========
//    I2C
// ========

/*!
 * This writes to a given set of registers on TTCrx. It can be used to write to either the
 * registers on the UKL1 board or to a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  writeRecipe If TRUE then the function will write the given register settings to the configurationDB recipe given by
 *           recipe name. FALSE it will write the register settings to the UKL1 board.
 * \param  recipeNameAndType Two element array containing:
 *   \li Element 1 - The name of the recipe to save the data to. It must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE
 *         is descriptive of the type of run the settings are saved for e.g. PHYSICS, CALIBRATION; FsmAction is the action of
 *         FSM state transition where the recipe should be loaded e.g. Configure, the first letter only should be capitalised.
 *   \li Element 2 - The second element contains the recipe type that it is to be saved to, this can be anything provided it already exists.
 *          This array is only required if writeRecipe=TRUE, otherwise it can be empty.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           written to, which corresponds to the appropriate hardware register. The accepted values are described in
 *           fwUkl1_parseGbeWriteRegisters();
 * \param  data By default an 8-bit value that is to be written to the register represented as a hex string, it need not be zero padded.
 *           Certain settings accept or require an alternative format as described for each register above.
 * \param  verifyWrite If this is set to TRUE then the register will be read back from to ensure that
 *           the data written corresponds to that read. Default: FALSE.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_ttcrxWrite(string ukl1Name, bool writeRecipe, const dyn_string& recipeNameAndType, dyn_string registerNames, dyn_string data, bool verifyWrite, dyn_string& exceptionInfo) {
  //Check that the register names and data arrays are the same size.
  //Save only the lenght of one of the arrays as the other must be the same size, in this case it is the registerNames.
  const unsigned numRegs = dynlen(registerNames);
  if ( numRegs != dynlen(data) ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): On port " + port + ", the number of register names (" + numRegs + ") is not equal to the amount of data (" + dynlen(data) + ").", "1");
    return;
  }

  //Holds the data in the form that the fwCcpc_write function expects.
  dyn_dyn_char ddcData;
  //Holds the mask for each element.
  dyn_dyn_char masks;

  //Converts the data given as strings in the calling arguments to byte arguments required by the fwCcpc_write.
  //Also it creates the masks required for each write and pads the register names, such that they refer to the appropriate
  //data point elements.
  const int exceptSize = dynlen(exceptionInfo);
  fwUkl1_parseTtcrxWriteRegisters(ukl1Name, registerNames, data, ddcData, masks, exceptionInfo);
  //Check that we were able to parse the data.
  if ( exceptSize == dynlen(exceptionInfo) ) {
    if (writeRecipe) {
      //Check that the recipeNameAndType is populated with two values as required.
      if ( 2 == dynlen(recipeNameAndType) ) {
	//For some reason we must first get all the data from the recipe and then write it all back otherwise it does not work properly.
	//When we get the recipe data it will give us all registers in the recipe and also data.
	//This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
	dyn_string fullRegisterNames;
	//Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
	dyn_dyn_char fullDdcData;
	if ( fwHw_getRecipe(ukl1Name, recipeNameAndType[1], fullRegisterNames, fullDdcData) ) {
	  //It is possible that we are writting to a new recipe, so check that we actually got something back from the recipe.
	  int fullLen = dynlen(fullDdcData);

	  //Loop over the registers that we want to update.
	  for (int element = 1; element <= numRegs; ++element) {
	    //In order to write to the recipe we must combine the masks and the data into a single byte array with the mask on the end.
	    //This must be done before we update the full list of data!
	    dynAppend(ddcData[element], masks[element]);
	    //This take the specific register name and returns the index of its occurance in the full list.
	    //It is this index that we can use to over write the appropriate data element.
	    int index = dynContains(fullRegisterNames, registerNames[element]);
	    //Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	    if (0 != fullLen)
	      fullDdcData[index] = ddcData[element];
	  }//for element

	  //Could be a new recipe, in this case just write what we are given.
	  if (0==fullLen) {
	    //So we create the appropriate number of register types.
	    fullLen = numRegs;
	    fullRegisterNames = registerNames;
	    fullDdcData = ddcData;
	  }

	  //The recipe needs to know the recipe type that we are using. Has to be for the full list.
	  dyn_int regTypes;
	  for (int i = 1; i <= fullLen; ++i) {
	    //All the registers are of the same type, LBUS, so just populate array with that value. Where is memset when you need it?
	    regTypes[i] = FWHW_LBUS;
	  }

	  //Now we can save the ddcData to the recipe. The empty array is for the type of register that will therefore be found by the fwHw function.
	  if ( !fwHw_saveRecipe(ukl1Name, recipeNameAndType[2], recipeNameAndType[1], fullRegisterNames, regTypes, fullDdcData) )
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): Failed to save the recipe " + recipeName + " for " + ukl1Name + ".", "1");
	}//if (fwHw_getRecipe())
	else {
	    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): Failed to read the recipe " + recipeName + " can't update the values.", "1");
	}//else (fwHw_getRecipe())
      }//if(2 == dynlen(recipeNameAndType))
      else {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): Recipe name and/or type not given. Cannot save to recipe.", "1");
      }//else(2 == dynlen(recipeNameAndType))
    }//writeToRecipe
    else {
      //The data and registers have been parsed, so let us write to the hardware.
      //We will wait for each element to be written before moving on.
      const bool waitFlag = TRUE;
      //Holds the status of the writes.
      dyn_int callStatusList;
      if ( fwCcpc_write(registerNames, ddcData, masks, callStatusList, waitFlag) ) {
	//Check to see if we want to verify the write.
	if (verifyWrite) {
	  //Call the verification function.
	  _fwUkl1_verifyWrite(registerNames, ddcData, masks, exceptionInfo);
	}
      } else {
	//Check where the errors occured.
	const unsigned statusLen = dynlen(callStatusList);
	if ( 0 != statusLen ) {
	  for (unsigned element = 1; element <= statusLen; ++element) {
	    if ( 0 != callStatusList[element] ) {
	      //Construct the error message.
	      string errMsg = "fwUkl1_ttcrxWrite(): Failed while writing to " + registerNames[element] + ", encountered ";
	      if (callStatusList[element] > 0) {
		errMsg += "a problem writing to the CCPC server.";
	      } else if (callStatusList[element] < 0) {
		errMsg += "a problem with PVSS.";
	      } else {
		errMsg += "an unknown error.";
	      }
	      fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	    }
	  }
	} else {
	  fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): A fatal error and an unknown number of writes were successful. L1 is in a highly inconsistent state.", "1");
	}
      }
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxWrite(): Failed to parse input data, not data was written.", "1");
  }

  //and we are done.
  return;
}

/*!
 * This reads from a given set of TTCrx registers. The reads can either be from the hardware or they can be
 * done from a configurationDB recipe.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  readRecipe If TRUE then the function will read the settings from the given recipe name and return the settings
 *           contained within. FALSE and the returned values will come directly from the hardware registers.
 * \param  recipeName Name of the recipe to save the configuration information to if writeDb=TRUE, not used if writeDb=FALSE.
 *           The recipe must be of the form "RUN_TYPE/FsmAction", where RUN_TYPE is type of run that the recipe is to be loaded
 *           e.g PHYSICS, CALIBRATION, can take any value, but should be descriptive of the type of run the recipe settings represent;
 *           FsmAction is the action in the FSM when these settings should be loaded e.g. Configure if the recipe is to be used to moved to
 *           the READY state, typically only the first letter is upper case.
 * \param  registerNames dyn_string that contains the name of the HwTypeCCPCUKL1 datapoint element to be
 *           read from, which corresponds to the appropriate hardware register.
 * \param  data dyn_string returned by reference that contains a string representing the value in the register.
 *           It will be zero padded to the nearest 32-bit word.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 */
void fwUkl1_ttcrxRead(string ukl1Name, bool readRecipe, const string& recipeName, dyn_string registerNames, dyn_string& data, dyn_string& exceptionInfo) {
  //Make sure we have an empty data array.
  dynClear(data);
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".TTCRX.";
  //Useful to know how many elements we may need to loop over.
  const unsigned numRegs = dynlen(registerNames);
  //Add prefix to each name and convert the data into a byte array.
  for (unsigned element = 1; element <= numRegs; ++element) {
    //Add the prefix.
    registerNames[element] = prefix + registerNames[element];
  }

  //Do the reads...
  //Holds the data from the reads.
  dyn_dyn_char ddcData;
  //Holds the status of each hardware read. Required by the parsing function,
  //but will be empty and hence ignored in the case of the recipe access.
  dyn_int callStatusList;
  dynClear(callStatusList);

  if (readRecipe) {
    //When we get the recipe data it will give us all registers in the recipe and also data.
    //This will hold the registers we are given and then we can check them against the list of requested registers to see what data to parse.
    dyn_string tmpRegisterNames;
    //Holds the data from the recipe, but it will be all the data stored in the recipe and will later be extracted.
    dyn_dyn_char tmpDdcData;
    if ( fwHw_getRecipe(ukl1Name, recipeName, tmpRegisterNames, tmpDdcData) ) {
      for (int element = 1; element <= numRegs; ++element) {
	//This take the specific register name and returns the index of its occurance in the full list.
	//Relies on the fact that the list is unique.
	//It is this index that we can use to over write the appropriate data element.
	int index = dynContains(tmpRegisterNames, registerNames[element]);
	//Now we know where the appropriate recipe data is copy it into the ddcData array converting from its string format.
	ddcData[element] = tmpDdcData[index];
      }
    }//if(fwhw_getRecipe())
    else
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_ttcrxRead(): Failed to retrieve the recipe " + recipeName + " for " + ukl1Name + ".", "1");
    
  }//if(readRecipe)
  else {
    fwCcpc_read(registerNames, ddcData, callStatusList);
  }

  //Parse the data we just read and let it deal with any errors that occurred.
  fwUkl1_parseTtcrxReadRegisters(ukl1Name, registerNames, ddcData, data, callStatusList, exceptionInfo);

  //and we are done
  return;
}

/*!
 * A function to read the serial number from a UKl1 board.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \return string Contain the 32-bit serial number as a single hex string. The string will return `Fail' if it does not succeed.
 */
string fwUkl1_getUkl1SerialNumber(string ukl1Name) {
  //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
  const string prefix = ukl1Name + ".PROM.";

  //Name of the register to read the ID from.
  dyn_string registerNames;
  registerNames[1] = prefix + "Id";

  //This will hold the serial number once read.
  string serialNumber = "";
  //Do the reads...
  //Holds the data in the form that the fwCcpc_read function returns.
  dyn_dyn_char ddcData;
  //Holds the status of the write.
  dyn_int callStatusList;
  fwCcpc_read(registerNames, ddcData, callStatusList);

  //Check the arrays are well formed.
  if ( 0 != dynlen(callStatusList) ) {
    //Put the data into a single string for return.
    if ( 0 != callStatusList[1] ) {
      //Construct the error message.
      string errMsg = "While reading UKL1 serial number the function encounterd ";
      if (callStatusList[1] > 0) {
	errMsg += "a problem writing to the CCPC server.";
      } else if (callStatusList[1] < 0) {
	errMsg += "a problem with PVSS.";
      } else {
	errMsg += "an unknown error.";
      }
      //Now log the error.
      _fwUkl1_LogFwCcpcError(errMsg);
      //Mark the error.
      serialNumber = "Fail";
    }
    else {
      //The ID number can just be converted straight to a hex string.
      dyn_string exceptionInfo;
      serialNumber = fwUkl1_convertByteToHexLittleEndian(ddcData[1], exceptionInfo);
    }
  }//if ( 0 != dynlen(callStatusList) )
  else {
    //The function didn't execute properly as the callStatusList size is zero.
    serialNumber = "Fail";
  }

  //Return the ID number.
  return serialNumber;

}

/*!
 * This function allows the temperature probes to be read.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return dyn_int Temperatures read from the temperature sensors around the UKL1 board. The order of the readings in the array
 *           is such that the sensor on the lowest register address is given in the lowest element number and the highest register
 *           address is in the last element. -1 will be returned if there is an error, which will be described in exceptionInfo.
 */
dyn_int fwUkl1_getUkl1Temperatures(string ukl1Name, dyn_string& exceptionInfo) {
  //Holds the temperature values.
  dyn_int temperatureReadings;

  //Loop over all the temperature sensors.
  for (int tempSens = 1; tempSens <= FWUKL1_TEMPERATURE_SENSORS; ++tempSens) {
    //Define the string to be added to each of the register names such that they are written to the appropriate datapoint element.
    const string prefix = ukl1Name + ".Temp " + tempSens + ".";

    //Name of the register to read the ID from.
    dyn_string registerNames;
    registerNames[1] = prefix + "TempRead";

    //Do the reads...
    //Holds the data in the form that the fwCcpc_read function returns.
    dyn_dyn_char ddcData;
    //Holds the status of the write.
    dyn_int callStatusList;
    fwCcpc_read(registerNames, ddcData, callStatusList);

    //Check the arrays are well formed.
    if ( 0 != dynlen(callStatusList) ) {
      //Put the data into a single string for return.
      if ( 0 != callStatusList[1] ) {
	//Construct the error message.
	string errMsg = "fwUkl1_getUkl1Temperatures(): Failed to read temperature from temperature sensor number " + tempSens + ". ";
	if (callStatusList[1] > 0) {
	  errMsg += "A problem writing to the CCPC server ";
	} else if (callStatusList[1] < 0) {
	  errMsg += "A problem with PVSS ";
	} else {
	  errMsg += "An unknown error ";
	}
	errMsg += "was encountered.";
	//Now raise an exception.
	fwException_raise(exceptionInfo, "ERROR", errMsg, "2");
	//Mark the error.
	temperatureReadings[tempSens] = -1;
      }
      else {
	//The ID number can just be converted straight to a hex string.
	temperatureReadins[tempSens] = fwCcpc_convertByteToDec(ddcData[1]);
      }
    }//if ( 0 != dynlen(callStatusList) )
    else {
      //The function didn't execute properly as the callStatusList size is zero.
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_getUkl1Temperatures(): Failed to read temperature from temperature sensor number " + tempSens + ". No data was returned.", "1");
    }

  }//for tempSens

  //Return the temperature readings.
  return temperatureReadings;

}


// ===========
//   SUPPORT
// ===========

/*!
 * In order to write to specific registers on the FPGAs we must first convert the address
 * to include the address of the FPGAs from the glue card perspective.
 *
 * \param  addr Address of the register to be written to.
 * \return unsigned Full address of the FPGA register.
 */
unsigned fwUkl1_FpgaizeAddress(unsigned addr) {
  // Based on code written by Steve Wotton, Cambridge.
  // The address sent to lb_{write|read}_hword should be above 0x20000
  // to be outside the GBE & gluecard ranges
  // The FPGA is sensitive only to bits 2-13 of the address
  //  (=> the FPGA is not sensitive to the 0x2???? that indicates
  //      that an address is outside the GBE/gluecard ranges)
  // So whenever we want to address the FPGA, set the 13th bit:
  //   the FPGA will only respond to addresses for which the 13th bit has been set
  //
  // So: << 2    translates a (32-bit) int address into (8-bit) byte address
  //     0x20000 gets us outside the address ranges of the GBE and gluecard
  //               (but will be ignored by the FPGA)
  //     0x2000  should stop the gluecard from reacting to the operation
  //               (but will be ignored by the FPGA)
  //
  // => (int) register addresses 0x0 to 0x7ff should be usable by the FPGA
  //

  return (0x08000000 + ( 0x800 | ( addr<<2 ) ) );
}

/*!
 * Reads from the registers on the L1 board to ensure that the data just written has been written successfully.
 *
 * \param  registerNames Names of the registers that were written to.
 * \param  writtenData The data that was written to the registers that the read back values should be compared against.
 * \param  masks Any masks that should be applied to the read data, such that only specific bits in a register are compared
 *           against the write bits.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void _fwUkl1_verifyWrite(dyn_string registerNames, dyn_dyn_char writtenData, dyn_dyn_char masks, dyn_string& exceptionInfo) {
  //This holds the read data.
  dyn_dyn_char ddcCheckData;
  //This holds the status from the register reads.
  dyn_int callStatusList;
  //Number of elements of data to loop through.
  const unsigned numRegs = dynlen(registerNames);
  //Read back from all registers.
  if ( fwCcpc_read(registerNames, ddcCheckData, callStatusList) ) {
    //Check the data.
    for (unsigned element = 1; element <= numRegs; ++element) {
      const int checkLen = dynlen(ddcCheckData[element]);
      const int dataLen  = dynlen(writtenData[element]);
      if (checkLen  != dataLen) {
	fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_verifyWrite(): For register " + registerNames[element] + " check data (" + checkLen + ") and written data (" + dataLen + ") lengths differ when verifying write.", "2");
      } else {
	//Only check the data if they are the same size.
	//Use to note if an element in the read array differ.
	bool diff = FALSE;
	//This will provide us with our loop index. The two arrays have been found to be the same size, so just set to either.
	for (unsigned index = 1; index <= dataLen; ++index) {
	  //Apply the masks that are given to the read data.
	  ddcCheckData[element][index] &= masks[element][index];
	  //Now check the data against the read values.
	  if ( ddcCheckData[element][index] != writtenData[element][index] ) {
	    diff = TRUE; 
	  }
	}
	//Will only raise the exception displaying the whole value and not for each differing element.
	if (diff)
	  fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_verifyWrite(): Register " + registerNames[element] + " checkData (" + fwCcpc_convertByteToHex(ddcCheckData[element]) + ") and written data (" + fwCcpc_convertByteToHex(writtenData[element]) + ") differ.", "2");
      }
    }
  } else {
    //Check where the errors occured.
    const unsigned statusLen = dynlen(callStatusList);
    if ( 0 != statusLen ) {
      for (unsigned element = 1; element <= statusLen; ++element) {
	if ( 0 != callStatusList[element] ) {
	  //Construct the error message.
	  string errMsg = "Failed to read the following register" + (1!=statusLen ? "s":"") + " during verification:\n";
	  //Construct the error message.
	  errMsg += registerNames[element] + " encountered ";
	  if (callStatusList[element] > 0) {
	    errMsg += "a problem writing to the CCPC server.";
	  } else if (callStatusList[element] < 0) {
	    errMsg += "a problem with PVSS.";
	  } else {
	    errMsg += "an unknown error.";
	  }
	  errMsg += "\n";
	  fwException_raise(exceptionInfo, "WARNING", errMsg, "2");
	}
      }
    } else {
      fwException_raise(exceptionInfo, "WARNING", "_fwUkl1_verifyWrite(): Failed to read from UKL1 to verify the write.", "2");
    }
  }

  //Done.
  return;
}

//@}


// =========================
//  STATE MACHINE FUNCTIONS
// =========================

/** @defgroup SectionState A group of functions that can be used to perform state machine actions
 *    for a UKL1 board. The following actions are defined:
 *    \li Configure - Loads of the configuration registers on the UKL1 board with the settings defined in the database.
 *          It will leave the board in the READY state. It is slow transition.
 *    \li Start - Moves the board into the RUNNING state. Enables data to flow through the UKL1 board. Fast transition.
 *    \li Stop  - Moves the board into the READY state. Prevents data flow through the UKL1 board. It will ignore TFC
 *          commands and cannot output over the GBE card.
 *    \li Reset  - Performs a reset of the UKL1. Currently the actions of this are ill defined.
 *    \li Reload - Reloads the FPGA (Ingress/FE and Egress/BE) firmware from the onboard memory.
 *    \li Recover - Currently does nothing.
 *
 *    None of the defined functions change the FSM state, this is left to the User to ensure the FSM is in the appropriate
 *    state after the function has been called.
 *  @{
 */

/*!
 * Gets the state of the current UKL1.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \return string State that the UKL1 board is currently in.
 */
string fwUkl1_GetUkl1State(string ukl1Name) {
  //Retrieve the state from the datapoint.
  string sState;
  dpGet(ukl1Name + ".status", sState);
  return sState;
}

/*!
 * Loads a specific recipe to a UKL1 board from the configuration database.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  recipe Name of the recipe that is to be applied to the UKL1 board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 board to a configured state, it is typically a long operation and
 * it is recommended that a fwUkl1_stop() is performed once completed. It will not update the
 * state of the UKL1, this must be done by the User.
 */
void fwUkl1_configure(string ukl1Name, string recipe, dyn_string& exceptionInfo) {
  //Holds the registers that can be set with this recipe.
  dyn_string regs;
  //Holds the recipe data.
  dyn_dyn_anytype data;
  
  //For the moment we will still configure the hardcoded settings of the GBE from value defined settings and not DB.
  //This performs some configuration routines for the GBE, setting the GBE into its running mode.
  const int exceptSize = dynlen(exceptionInfo);
  _fwUkl1_configureHardcodedSettings(ukl1Name, exceptionInfo);
  //Check for errors.
  if ( (exceptSize == dynlen(exceptionInfo)) || (-1 == exceptSize) ) {
    //Initialised GBE.
    //Get the recipe.
    if ( fwHw_getRecipe(ukl1Name, recipe, regs, data) ) {
      //Now apply the recipe.
      if ( !fwHw_applyRecipe(ukl1Name, regs, data) ) {
	fwException_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to apply recipe.", "1");
      }
    } else {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to read recipe.", "1");
    }
  } else {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_configure(): Failed to initialise GBE.", "1");
  }

  return;
}

/*!
 * Start does only two things and that is to enable the TFC decoding on the UKL1 and enable the link from
 * the BE FPGA to the GBE card for data output.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  recipe Name of the recipe that is to be applied to the UKL1 board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 to the start state where it is ready to take data. It will not update the state
 * once complete this is upto the User.
 */
void fwUkl1_start(string ukl1Name, string recipe, dyn_string& exceptionInfo) {
  //Holds the registers that can be set with this recipe.
  dyn_string regs;
  //Holds the recipe data.
  dyn_dyn_anytype data;
  
  //Get the recipe.
  int status = fwHw_getRecipe(ukl1Name, recipe, regs, data);
  if ( 1 != status ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_start(): Failed to read recipe.", "1");
  } else {
    //Now apply the recipe.
    status = fwHw_applyRecipe(ukl1Name, regs, data);
    if ( 1 != status ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_start(): Failed to apply recipe.", "1");
    }
  }

  return;

}


/*!
 * Stop does only two things and that is to disable the TFC decoding on the UKL1 and disable the link from
 * the BE FPGA to the GBE card for data output.
 *
 * \param  ukl1Name Name of the UKL1 board as defined in the DIM server.
 * \param  recipe Name of the recipe that is to be applied to the UKL1 board.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * This will move the UKL1 to the start state where it will not be able to see triggers sent from the TFC and also
 * is not capable of outputting data. It will not update the state once complete this is upto the User.
 */
void fwUkl1_stop(string ukl1Name, string recipe, dyn_string& exceptionInfo) {
  //Holds the registers that can be set with this recipe.
  dyn_string regs;
  //Holds the recipe data.
  dyn_dyn_anytype data;
  
  //Get the recipe.
  int status = fwHw_getRecipe(ukl1Name, recipe, regs, data);
  if ( 1 != status ) {
    fwException_raise(exceptionInfo, "ERROR", "fwUkl1_stop(): Failed to read recipe.", "1");
  } else {
    //Now apply the recipe.
    status = fwHw_applyRecipe(ukl1Name, regs, data);
    if ( 1 != status ) {
      fwException_raise(exceptionInfo, "ERROR", "fwUkl1_stop(): Failed to apply recipe.", "1");
    }
  }

  return;

}

/*!
 * Performs a reset of the FPGAs on the L1 board. Effects still under discussion.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_reset(string ukl1Name, dyn_string& exceptionInfo) {
  //The reset is performed by setting the GPIO line 5 low for a few 10s of milliseconds and then setting it back to high.
  //First it must be enabled for output.
  //Note: we could check that it is enabled for output, but it would always require at least 1 hardware
  //  access, so that 1 access might as well be the setting.
  int callStatus = fwCcpc_GPIOEnable(ukl1Name, FWCCPC_GPIO_LINE_5, FWCCPC_GPIO_OUTPUT);
  //Check for errors.
  if ( 0 == callStatus ) {
    //Only worth trying to do this if it succeeded.
    int callStatus = fwCcpc_GPIOSet(ukl1Name, FWCCPC_GPIO_LINE_5, FWCCPC_GPIO_LOW);
    //Check for errors.
    if ( 0 == callStatus ) {
      //We have driven it low, wait a few milli-seconds for the reset to take affect.
      delay(0,20);
      //Only worth trying to do this if it succeeded.
      int callStatus = fwCcpc_GPIOSet(ukl1Name, FWCCPC_GPIO_LINE_5, FWCCPC_GPIO_HIGH);
      //Check for errors.
      if ( 0 != callStatus ) {
	//Construct the error message.
	string errMsg = "fwUkl1_reset(): Failed to set the GPIO line 5 high during a reset, unable to desassert the reset. Encountered";
	if (callStatus > 0) {
	  errMsg += " a problem writing to the CCPC server.";
	} else if (callStatus < 0) {
	  errMsg += " a problem with PVSS.";
	} else {
	  errMsg += " an unknown error.";
	}
	//Now raise the exception.
	fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
      }
    } else {
      //Construct the error message.
      string errMsg = "fwUkl1_reset(): Failed to set the GPIO line 5 low during a reset, unable to reset. Encountered";
      if (callStatus > 0) {
	errMsg += " a problem writing to the CCPC server.";
      } else if (callStatus < 0) {
	errMsg += " a problem with PVSS.";
      } else {
	errMsg += " an unknown error.";
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
    }
  } else {
    //Construct the error message.
    string errMsg = "fwUkl1_reset(): Failed to enable GPIO line 5 for output during a reset, unable to reset. Encountered";
    if (callStatus > 0) {
      errMsg += " a problem writing to the CCPC server.";
    } else if (callStatus < 0) {
      errMsg += " a problem with PVSS.";
    } else {
      errMsg += " an unknown error.";
    }
    //Now raise the exception.
    fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
  }

  //Now return.
  return;
}

/*!
 * Performs a reload of the FPGA firmware on the L1 board. It will reset all values back to their defaults.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 */
void fwUkl1_reload(string ukl1Name, dyn_string& exceptionInfo) {
  //The reload is performed by setting the GPIO line 6 low for a few 10s of milliseconds and then setting it back to high.
  //First it must be enabled for output.
  //Note: we could check that it is enabled for output, but it would always require at least 1 hardware
  //  access, so that 1 access might as well be the setting.
  int callStatus = fwCcpc_GPIOEnable(ukl1Name, FWCCPC_GPIO_LINE_6, FWCCPC_GPIO_OUTPUT);
  //Check for errors.
  if ( 0 == callStatus ) {
    //Only worth trying to do this if it succeeded.
    int callStatus = fwCcpc_GPIOSet(ukl1Name, FWCCPC_GPIO_LINE_6, FWCCPC_GPIO_LOW);
    //Check for errors.
    if ( 0 == callStatus ) {
      //We have driven it low, wait a few milli-seconds for the reload to take affect.
      delay(0,20);
      //Only worth trying to do this if it succeeded.
      int callStatus = fwCcpc_GPIOSet(ukl1Name, FWCCPC_GPIO_LINE_6, FWCCPC_GPIO_HIGH);
      //Check for errors.
      if ( 0 != callStatus ) {
	//Construct the error message.
	string errMsg = "fwUkl1_reload(): Failed to set the GPIO line 6 high during a reload, unable to desassert the reload. Encountered";
	if (callStatus > 0) {
	  errMsg += " a problem writing to the CCPC server.";
	} else if (callStatus < 0) {
	  errMsg += " a problem with PVSS.";
	} else {
	  errMsg += " an unknown error.";
	}
	//Now raise the exception.
	fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
      }
    } else {
      //Construct the error message.
      string errMsg = "fwUkl1_reload(): Failed to set the GPIO line 6 low during a reload, unable to reload. Encountered";
      if (callStatus > 0) {
	errMsg += " a problem writing to the CCPC server.";
      } else if (callStatus < 0) {
	errMsg += " a problem with PVSS.";
      } else {
	errMsg += " an unknown error.";
      }
      //Now raise the exception.
      fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
    }
  } else {
    //Construct the error message.
    string errMsg = "fwUkl1_reload(): Failed to enable GPIO line 6 for output during a reload, unable to reload. Encountered";
    if (callStatus > 0) {
      errMsg += " a problem writing to the CCPC server.";
    } else if (callStatus < 0) {
      errMsg += " a problem with PVSS.";
    } else {
      errMsg += " an unknown error.";
    }
    //Now raise the exception.
    fwException_raise(exceptionInfo, "ERROR", errMsg, "1");
  }

  //Now return.
  return;
}

/*!
 * Recovers the UKL1 board from an error state. This will return the UKL1 to an unconfigured state and the values in
 * the registers cannot be guarantied to be the same as before recovering. The User should update the FSM state appropriately.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void
 *
 * The interface to this function is unlikely to change.
 */
void fwUkl1_recover(string ukl1Name, dyn_string& exceptionInfo) {
  //Do something to recover from an error...
  return;
}



//THIS IS STILL REQUIRED WHEN CONFIGURING!!!
/*!
 * Configures all the settings on the GBE that are never changed by the User and all values are hardcoded into this function.
 * It also setups the single TTCrx register is hardcoded to a none default value.
 *
 * \param  ukl1Name Name of the UKL1 board as it appears in the DIM server.
 * \param  exceptionInfo Error information. If dynlen is zero no errors occurred. The standard error codes (defined in library
 *           documentation) are used.
 * \return void.
 *
 * The values set here are defined by the LHCb technical note Quad Gigabit Ethernet plug-in card and are correct as of
 * issue 2.2 revision 0 of that document. They must be performed in a specific order.
 */
void _fwUkl1_configureHardcodedSettings(const string& ukl1Name, dyn_string& exceptionInfo) {
  //Sort this out and make it look like the rest of the functions.
  //Add TTCrx setting and update the gbeWrite calls.

  //This holds the list of registers that to be written
  dyn_string registerList;
  //and their corresponding data.
  dyn_string sData;
  //both MUST be cleared after each write.

  //Set up the TTCrx control register as desired.
  registerList[1] = "Control";
  sData[1] = "a1";
  //Don't write to a recipe and don't verify the write.
  //As we are not writing to a recipe we don't need to give it a recipe name and type.
  fwUkl1_ttcrxWrite(ukl1Name, FALSE, makeDynString(), registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);

  //Place MAC, SPI3 TX and RX, RX and TX FIFOs in reset.
  registerList[1] = "MACSoftReset";
  sData[1] = "0000000f";
  registerList[2] = "SPI3ConfigTrnmtGlobal";
  sData[2] = "000c000f";
  registerList[3] = "RxFifoPrtReset";
  sData[3] = "0000000f";
  registerList[4] = "TxFifoPrtReset";
  sData[4] = "0000000f";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  //Disable all ports.
  registerList[1] = "PortEnable";
  sData[1] = "Disable,Disable,Disable,Disable";
  registerList[2] = "Clock";
  sData[2] = "00000000";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  //Select copper mode
  registerList[1] = "Mode";
  sData[1] = "0000000f";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);
  
  for (unsigned port = 0; port < FWUKL1_GBE_PORTS; ++port) {
    // Select copper mode for each port.
    registerList[1] = "MAC";
    sData[1] = "00000002";
    //Perform the write.
    fwUkl1_gbePortWrite(ukl1Name, port, registerList, sData, registerList, sData, FALSE, exceptionInfo);
    //Clear the arrays for next use.
    dynClear(registerList);
    dynClear(sData);
    //Delaying for the shortest time possible.
    delay(0,1);

    //Set full duplex mode for each port.
    registerList[1] = "Duplex";
    sData[1] = "00000001";
    //Perform the write.
    fwUkl1_gbePortWrite(ukl1Name, port, registerList, sData, registerList, sData, FALSE, exceptionInfo);
    //Clear the arrays for next use.
    dynClear(registerList);
    dynClear(sData);
    //Delaying for the shortest time possible.
    delay(0,1);
  }

  //Enable clocks for all active channels
  registerList[1] = "Clock";
  sData[1] = "0000000f";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  //Remove resets
  registerList[1] = "MACSoftReset";
  sData[1] = "00000000";
  registerList[2] = "SPI3ConfigTrnmtGlobal";
  sData[2] = "0000000f";
  registerList[3] = "RxFifoPrtReset";
  sData[3] = "00000000";
  registerList[4] = "TxFifoPrtReset";
  sData[4] = "00000000";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  // Enable automatic padding and CRC generation for all active channels
  registerList[1] = "Config";
  // Port setting
  sData[1] = "000011cd";
  for (unsigned port = 0; port < FWUKL1_GBE_PORTS; ++port) {
    //Perform the write.
    fwUkl1_gbePortWrite(ukl1Name, port, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  }
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  // Setup Rx interface: configure SPI3 width 32 bit, RVAL pause and active channel
  registerList[1] = "SPI3ConfigRcv";
  sData[1] = "00ffff80";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  // Setup Rx interface: configure RX CRC stripping and check
  registerList[1] = "RxFifoEnable";
  sData[1] = "000000f0";
  //Perform the write.
  fwUkl1_gbeWrite(ukl1Name, registerList, sData, registerList, sData, FALSE, exceptionInfo);
  //Clear the arrays for next use.
  dynClear(registerList);
  dynClear(sData);
  //Delaying for the shortest time possible.
  delay(0,1);

  //Done.
  return;
}

//@}


// ======================================
//  ERROR FUNCTIONS
// ======================================

/** @defgroup SectionError Error functions.
 *  These functions are used for logging errors that occur when the various read/write
 *  operations are performed.
 *  @{
 */

/*!
 * Sends a message to the FSM message box and optionally the RichL1Log datapoint,
 * also appends 'STATUS:' to the beginning of all messages.
 *
 * \param  message Debug message that is to be sent to the datapoint.
 * \param  level This is the level of message that is to be logged. It will only be logged if it is
 *           greater or equal to the global message level. Should be a value specified by the
 *           FWUKL1_*_MESSAGE_LEVEL constants.
 *
 * It can also send a message to the Ukl1MessageLog datapoint, this depends on the library constant
 * FWUKL1_SEND_MESSAGE_TO_DATAPOINT. This will be set to false in versions released with the FSM.
 */
// void fwUkl1_StatusMessage(string message, unsigned level) {
//   //First check the level of the warning against the global level.
//   if (gufwUkl1_StatusMessageLevel >= level) {
//     //Append the appropriate text to indicate this is a warning.
//     message = "STATUS: " + message;
//     //Send the message to the FSM message box.
//     DebugTN(message);
//     if (bfwUkl1_SendMessageToDataPoint) {
//       //First we must get the existing debug messages.
//       dyn_string dbgMsg;
//       int dpStatus = dpGet("RichL1Log.Status", dbgMsg);
//       if (0 == dpStatus) {
// 	//Successfully accessed the datapoint.
// 	//Now append the new message to this.
// 	dynAppend(dbgMsg, message);
// 	//and write it back to the datapoint.
// 	dpStatus = dpSet("RichL1Log.Status", dbgMsg);
//       } else {
// 	DebugTN("Failed to access the status datapoint:");
// 	DebugTN("RichL1Log.Status");
// 	DebugTN("Status messages diverted to PVSS log, the following message was recieved:");
// 	DebugTN(message);
//       }
//     }
//   }
// }

/*!
 * Sends a message to the FSM message box and optionally the RichL1Log datapoint,
 * also appends 'WARNING:' to the beginning of all messages.
 *
 * \param  message Warning message that is to be logged.
 * \param  level This is the level of message that is to be logged. It will only be logged if it is
 *           greater or equal to the global message level. Should be a value specified by the
 *           FWUKL1_*_MESSAGE_LEVEL constants.
 *
 * It can also send a message to the Ukl1MessageLog datapoint, this depends on the library constant
 * FWUKL1_SEND_MESSAGE_TO_DATAPOINT. This will be set to false in versions released with the FSM.
 */
// void fwUkl1_WarningMessage(string message, unsigned level) {
//   //First check the level of the warning against the global level.
//   if (gufwUkl1_WarningMessageLevel >= level) {
//     //Append the appropriate text to indicate this is a warning.
//     message = "WARNING: " + message;
//     //Send the message to the FSM message box.
//     DebugTN(message);
//     if (bfwUkl1_SendMessageToDataPoint) {
//       //First we must get the existing debug messages.
//       dyn_string dbgMsg;
//       int dpStatus = dpGet("RichL1Log.Warning", dbgMsg);
//       if (0 == dpStatus) {
// 	//Successfully accessed the datapoint.
// 	//Now append the new message to this.
// 	dynAppend(dbgMsg, message);
// 	//and write it back to the datapoint.
// 	dpStatus = dpSet("RichL1Log.Warning", dbgMsg);
//       } else {
// 	DebugTN("Failed to access the status datapoint:");
// 	DebugTN("RichL1Log.Warning");
// 	DebugTN("Warning messages diverted to PVSS log, the following message was recieved:");
// 	DebugTN(message);
//       }
//     }
//   }
// }

/*!
 * Sends a message to the FSM message box and optionally the RichL1Log datapoint,
 * also appends 'ERROR:' to the beginning of all messages.
 *
 * \param  message Warning message that is to be logged.
 * \param  level This is the level of message that is to be logged. It will only be logged if it is
 *           greater or equal to the global message level. Should be a value specified by the
 *           FWUKL1_*_MESSAGE_LEVEL constants.
 *
 * It can also send a message to the Ukl1MessageLog datapoint, this depends on the library constant
 * FWUKL1_SEND_MESSAGE_TO_DATAPOINT. This will be set to false in versions released with the FSM.
 */
// void fwUkl1_ErrorMessage(string message, unsigned level) {
//   if (gufwUkl1_ErrorMessageLevel >= level) {
//     //Append the appropriate text to indicate this is a warning.
//     message = "ERROR: " + message;
//     //Send the message to the FSM message box.
//     DebugTN(message);
//     if (bfwUkl1_SendMessageToDataPoint) {
//       //First we must get the existing debug messages.
//       dyn_string dbgMsg;
//       int dpStatus = dpGet("RichL1Log.Error", dbgMsg);
//       if (0 == dpStatus) {
// 	//Successfully accessed the datapoint.
// 	//Now append the new message to this.
// 	dynAppend(dbgMsg, message);
// 	//and write it back to the datapoint.
// 	dpStatus = dpSet("RichL1Log.Error", dbgMsg);
//       } else {
// 	DebugTN("Failed to access the status datapoint:");
// 	DebugTN("RichL1Log.Error");
// 	DebugTN("Error messages diverted to PVSS log, the following message was recieved:");
// 	DebugTN(message);
//       }
//     }
//   }
// }

/**
 * Clears all the messages from the message log.
 *
 * \return int Returns an error code that indicates the execution status of the function. They are as follows:
 *   \li  0 Function completed successfully.
 *   \li -1 Execution failed and the board is in an unuseable state due to a PVSS problem.
 */
int fwUkl1_ClearAllLogMessages() {
  //Check the return status for errors.
  //Write to the appropriate log if we failed to clear it (it will default to the PVSS log if there is a datapoint issue).
  //Return the appropriate state.
  if ( 0 != fwUkl1_ClearStatusMessages() ) {
    fwUkl1_StatusMessage("Failed to clear status log, due to a PVSS error.", FWUKL1_INFO_MESSAGE_LEVEL);
    return -1;
  }
  if ( 0 != fwUkl1_ClearWarningMessages() ) {
    fwUkl1_WarningMessage("Failed to clear warning log, due to a PVSS error.", FWUKL1_INFO_MESSAGE_LEVEL);
    return -1;
  }
  if ( 0 == fwUkl1_ClearCriticalMessages() ) {
    return 0;
  } else {
    fwUkl1_ErrorMessage("Failed to clear error log, due to a PVSS error.", FWUKL1_INFO_MESSAGE_LEVEL);
    return -1;
  }
}

/**
 * Clears all the status messages from the message log.
 *
 * \return int Returns an error code that indicates the execution status of the function. They are as follows:
 *   \li  0 Function completed successfully.
 *   \li -1 Execution failed and the board is in an unuseable state due to a PVSS problem.
 */
int fwUkl1_ClearStatusMessages() {
  //Write a 1 element dyn_string to the datapoint. Prevents problems accessing the zeroth element.
  dyn_string clearedString = makeDynString("Cleared status messages.");
  if ( 0 == dpSet("RichL1Log.Status", clearedString) ) {
    return 0;
  } else {
    return 1;
  }
}

/**
 * Clears all the warning messages from the message log.
 *
 * \return int Returns an error code that indicates the execution status of the function. They are as follows:
 *   \li  0 Function completed successfully.
 *   \li -1 Execution failed and the board is in an unuseable state due to a PVSS problem.
 */
int fwUkl1_ClearWarningMessages() {
  //Write a 1 element dyn_string to the datapoint. Prevents problems accessing the zeroth element.
  dyn_string clearedString = makeDynString("Cleared warning messages.");
  if ( 0 == dpSet("RichL1Log.Warning", clearedString) ) {
    return 0;
  } else {
    return 1;
  }
}

/**
 * Clears all the error messages from the message log.
 *
 * \return int Returns an error code that indicates the execution status of the function. They are as follows:
 *   \li  0 Function completed successfully.
 *   \li -1 Execution failed and the board is in an unuseable state due to a PVSS problem.
 */
int fwUkl1_ClearErrorMessages() {
  //Write a 1 element dyn_string to the datapoint. Prevents problems accessing the zeroth element.
  dyn_string clearedString = makeDynString("Cleared error messages.");
  if ( 0 == dpSet("RichL1Log.Error", clearedString) ) {
    return 0;
  } else {
    return 1;
  }
}

/**
 * Logs an error that occurred with while using the framework library for writing/reading from the
 * CCPC.
 *
 * \param  messageLevel Level that the message is defined as.
 * \param  message Error message that is to be logged.
 */
// void _fwUkl1_LogFwCcpcError(string message) {
//   //Append a note specifying it occurred within the framework and then log it as a critical error.
//   string newMessage = "fwCcpc.ctl lib reported error: " + message;
//   //Just call the appropriate logging function return its state.
//   fwUkl1_ErrorMessage(newMessage, FWUKL1_DEBUG_MESSAGE_LEVEL);
// }

/**
 * Logs an error that occurred with while retrieving/setting a datapoint.
 *
 * \param  messageLevel Level that the message is defined as.
 * \param  message Error message that is to be logged.
 */
// void _fwUkl1_LogDpError(string message) {
//   //Append a note specifying it occurred within the framework and then log it as a critical error.
//   string newMessage = "datapoint error: " + message;
//   //Just call the appropriate logging function return its state.
//   fwUkl1_ErrorMessage(newMessage, FWUKL1_DEBUG_MESSAGE_LEVEL);
// }

//@}


// ======================================
//  SUPPORT FUNCTIONS
// ======================================

/** @defgroup SectionSupport Support funtions
 *  @{
 */

/*!
 * This takes an FSM state represented by an integer and defined by the library constants FWUKL1_*_STATE and converts
 * a string that represents that state. Any unrecognised states will be returned as "Not Ready".
 *
 * \param  state Integer state that is to be converted.
 * \return string State represented as text.
 */
string fwUkl1_ConvertStateNumberToText(int state) {
  //State that the boards are in.
  string sState;
  if (FWUKL1_NOT_READY_STATE == state) {
    sState = "Not ready";
  } else if (FWUKL1_CONFIGURING_STATE == state) {
    sState = "Configuring";
  }else if (FWUKL1_START_STATE == state) {
    sState = "Running";
  } else if (FWUKL1_STOP_STATE == state) {
    sState = "Stopped";
  } else {
    sState = "Not ready";
  }
  return sState;
}

/*!
 * Strips the system name from the given string. Will remove only the system name. If no system name is present
 * then the effects are undefined, unless there are no semi-colons present in the string.
 *
 * \param  dpName String containing the DP name that the system name is to be removed from.
 * \return string dpName minus the beginning system name.
 */
string fwUkl1_removeSystemName(string dpName) {
  //Will hold the stripped DP name, will rely on it being initialised to NULL.
  string strippedDpName = "";

  //First break up the string into its semi-colon parts.
  const dyn_string semiColonDelimited = strsplit(dpName, ':');
  //Determine how many parts we have.
  const int numSec = dynlen(semiColonDelimited);
  if ( 1 == numSec ) {
    //No semi-colons found just return the original string.
    strippedDpName = dpName;
  } else if ( 2 <= numSec ) {
    //Start at the second element, bypassing the system name and then append the other elements.
    for (int sec = 2; sec <= numSec; ++sec) {
      strippedDpName += semiColonDelimited[sec];
    }
  } else {
    //Must be no elements, just return the original string which is probably NULL :S
    strippedDpName = dpName;
  }

  //Return!
  return strippedDpName;
  
}


//@}
